/*
===============================================================================

  FILE:  SVreader.h
  
  CONTENTS:
  
    Streaming Volume Mesh Reader
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    12 August 2008 -- added hexahedrons at SIGGRAPH but thinking of Henna
    20 February 2005 -- created after streaming points turned out too tricky
  
===============================================================================
*/
#ifndef SVREADER_H
#define SVREADER_H

#include "mydefs.h"

// types
typedef enum {
  SV_UNDEF = -1,
  SV_TET = 0,
  SV_HEX = 1,
  SV_MIXED = 2
} SVtype;

// events 
typedef enum {
  SV_ERROR = -1,
  SV_EOF = 0,
  SV_VERTEX = 1,
  SV_TETRAHEDRON = 2,
  SV_HEXAHEDRON = 3,
  SV_FINALIZED = 4
} SVevent;

class SVreader
{
public:

  // vertex variables

  I32 v_idx;
  F32* v_prop_f;

  // cell variables

  I32    c_nverts;
  I32*   c_idx;
  bool*  c_final;
  F32* c_prop_f;

  // explicit finalization variables

  I32 final_idx;

  // mesh variables

  SVtype type;

  I32 ncomments;
  char** comments;

  I32 v_pnum_f;
  F32* v_pmin_f;
  F32* v_pmax_f;

  I32 c_pnum_f;
  F32* c_pmin_f;
  F32* c_pmax_f;

  I32 nverts;
  I32 ncells;

  int v_count;
  int c_count;

  bool post_order;

  // functions

  virtual bool compute_bounding_box();

  virtual SVevent read_element()=0;
  virtual SVevent read_event()=0;

  virtual void close(bool close_file=true);

  SVreader();
  virtual ~SVreader();
};

#endif
