/*
===============================================================================

  FILE:  SVreader_silo.h

  CONTENTS:

    Reads a Streaming Volume Mesh in SILO format.
    Handles Tet and hex mixed meshes with arbitrary number of vertex and cell properties.

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009 clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    6 November 2009 -- Created

===============================================================================
*/
#ifndef SVREADER_SILO_H
#define SVREADER_SILO_H

#include "svreader.h"

#include <stdio.h>

class SVreader_silo : public SVreader
{
	public:

		// SVreader interface function implementations
#ifdef BUILD_SILO_READER
		SVevent read_element();
		SVevent read_event();

		void close ( bool close_file=true );

		// SVreader_silo functions

		bool open ( FILE* fp );

		SVreader_silo();
		~SVreader_silo();

	private:
		DBFile* file;
		int have_finalized, next_finalized;
		I32* finalized_vertices;

		void read_header();
		void read_buffer();

#else //BUILD_SILO_READER
		SVreader_silo()
		{
			fprintf(stderr, "ERROR: Your version of SV_API has been built without SILO support. You can enable SILO support with -DBUILD_SILO_READER.\n\n");
		}
		
		SVevent read_element(){return SV_ERROR;}
		SVevent read_event(){return SV_ERROR;}
		void close ( bool close_file=true ){}
		bool open ( FILE* fp ){return false;}
#endif //BUILD_SILO_READER

	
};

#endif
