/*
===============================================================================

  FILE:  SVreader_vtk.h
  
  CONTENTS:
  
    Reads a mesh from the VTK format and presents it as a (poor) streaming mesh. 
    Only the vertex positions and the tetrahedral/hexahedral connectivity are extracted from the VTK file. 
  
  PROGRAMMERS:
  
    clement.courbet@ecp.fr
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2009 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    09 November 2009 -- (clement )extended to handle arbitrary number of cell properties
    06 November 2009 -- (clement )fixed the bug where the reader could not read
				(valid) VTK files generated by VisIt. 
				These files had several vertices per line.
    13 August 2008 -- extended to hex meshes during the morning paper session
    17 September 2007 -- created the day i had to choose between tcp1 and tcp2

===============================================================================
*/
#ifndef SVREADER_VTK_H
#define SVREADER_VTK_H

#include "svreader.h"

#include <stdio.h>

class SVreader_vtk : public SVreader
{
public:

  // SVreader interface function implementations

  SVevent read_element();
  SVevent read_event();

  void close(bool close_file=true);

  // SVreader_vtk functions

  bool open(FILE* file);

  SVreader_vtk();
  ~SVreader_vtk();

private:
  int skipped_lines;
  FILE* file;
  char* line;
  char* curLinePos;

  long cellDataCurFilePos;

};

#endif
