/*
===============================================================================

  FILE:  SVwriter.h
  
  CONTENTS:
  
    Streaming Volume Mesh Writer Interface Class
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    21 February 2005 -- created after booking Henna's ticket to APA conference
  
===============================================================================
*/
#ifndef SVWRITER_H
#define SVWRITER_H

#include "mydefs.h"

class SVwriter
{
public:

  // mesh variables

  I32 ncomments;
  char** comments;

  I32 nverts;
  I32 ncells;

  I32 v_count;
  I32 c_count;

  I32 v_pnum_f;
  float* v_pmin_f;
  float* v_pmax_f;

  I32 c_pnum_f;
  float* c_pmin_f;
  float* c_pmax_f;

  // functions

  virtual void add_comment(const char* comment);

  virtual void set_nverts(I32 nverts);
  virtual void set_ncells(I32 ncells);
  virtual void set_v_pnum_f(I32 v_pnum_f);
  virtual void set_v_pbox_f(const float* v_pmin_f, const float* v_pmax_f);
  virtual void set_c_pnum_f(int c_pnum_f);
  virtual void set_c_pbox_f(const float* c_pmin_f, const float* c_pmax_f);

  virtual void write_vertex(const float* v_data_f)=0;

  virtual void write_tetrahedron(const I32* c_idx, const bool* c_final=0)=0;
  virtual void write_tetrahedron(const I32* c_idx, const float* c_data_f, const bool* c_final=0)=0;

  virtual void write_hexahedron(const I32* c_idx, const bool* c_final=0)=0;
  virtual void write_hexahedron(const I32* c_idx, const float* c_data_f, const bool* c_final=0)=0;

  virtual void write_finalized(I32 final_idx)=0;

  virtual void close(bool close_file=true, bool update_header=true);

  SVwriter();
  virtual ~SVwriter();
};

#endif
