/*
===============================================================================

  FILE:  SVwriter_hzip.h

  CONTENTS:

    Writes a Streaming Volume Mesh using Peter Lindstrom's HexZip library.
    The library has to be dowmloaded from LLNL's LOCAL project page.
    To enable this writer, use BUILD_HZIP.

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009 clement.courbet@ecp.fr

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    4 December 2009 -- adapted from HZIP's examples to support streaming meshes

===============================================================================
*/
#ifndef SVWRITER_HZIP_H
#define SVWRITER_HZIP_H

#include "svwriter.h"

#include <hzip.h>

#include <stdio.h>

//note: dont define this, it does not work
//#define MULTIPLEX

class SVwriter_hzip : public SVwriter
{

	public:
		// SVwriter interface function implementations

		void write_vertex ( const float* v_prop_f );

		void write_tetrahedron ( const int32_t* c_idx, const bool* c_final=0 );
		void write_tetrahedron ( const int32_t* c_idx, const float* c_prop_f, const bool* c_final=0 );

		void write_hexahedron ( const int32_t* c_idx, const bool* c_final=0 );
		void write_hexahedron ( const int32_t* c_idx, const float* c_data_f, const bool* c_final=0 );

		void write_finalized ( int32_t final_idx );

		void close ( bool close_file=true, bool update_header=true );

		// SVwriter_hzip functions

		bool open ( FILE* file );

		SVwriter_hzip();
		~SVwriter_hzip();

	private:
		FILE* file;
		HZMstream *conn_stream;
		HZNstream **geo_stream; //the HZIP streams
		HZMCODECzlib param;
#ifndef MULTIPLEX
		FILE *conn_file;
		FILE **geo_file;
#endif

		void write_header();

};

#endif
