/*
===============================================================================

  FILE:  SVwriter_nil.h
  
  CONTENTS:
  
    A Streaming Volume Mesh writer that is a dummy. It does not do anything.
    
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2006  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    03 January 2006 -- created after eating the gummybears henna got in eugene  

===============================================================================
*/
#ifndef SVWRITER_NIL_H
#define SVWRITER_NIL_H

#include "svwriter.h"

class SVwriter_nil : public SVwriter
{
public:

  // svwriter interface function implementations

  void add_comment(const char* comment) {};

  void set_nverts(I32 nverts);
  void set_ncells(I32 ncells);

  void set_v_pnum_f(I32 v_pnum_f) {};
  void set_v_pbox_f(const float* v_pmin_f, const float* v_pmax_f) {};
  void set_c_pnum_f(int c_pnum_f) {};
  void set_c_pbox_f(const float* c_pmin_f, const float* c_pmax_f) {};

  void write_vertex(const float* v_prop_f);

  void write_tetrahedron(const I32* c_idx, const bool* c_final=0);
  void write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final=0);

  void write_hexahedron(const I32* c_idx, const bool* c_final=0);
  void write_hexahedron(const I32* c_idx, const float* c_data_f, const bool* c_final=0);

  void write_finalized(I32 final_idx) {};

  void close(bool close_file=true, bool update_header=true);

  // svwriter_nil functions

  bool open(FILE* file);

  SVwriter_nil() {nverts = -1; ncells = -1; v_count = 0; c_count = 0;};
  ~SVwriter_nil() {};

  FILE* file;
};

void SVwriter_nil::set_nverts(I32 nverts)
{
  this->nverts = nverts;
}
void SVwriter_nil::set_ncells(I32 ncells)
{
  this->ncells = ncells;
}
void SVwriter_nil::write_vertex(const float* v_prop_f)
{
  v_count++;
}
void SVwriter_nil::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  c_count++;
}
void SVwriter_nil::write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  c_count++;
}
void SVwriter_nil::write_hexahedron(const I32* c_idx, const bool* c_final)
{
  c_count++;
}
void SVwriter_nil::write_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  c_count++;
}
void SVwriter_nil::close(bool close_file, bool update_header)
{
  // close of SVwriter interface
  if (nverts != -1) if (nverts != v_count) fprintf(stderr,"WARNING (SVwriter_nil): set nverts %d but v_count %d\n",nverts,v_count);
  if (ncells != -1) if (ncells != c_count) fprintf(stderr,"WARNING (SVwriter_nil): set ncells %d but c_count %d\n",ncells,c_count);
  nverts = v_count;
  ncells = c_count;
  v_count = -1;
  c_count = -1;
  if (close_file && file && file != stdout) fclose(file);
  file = 0;
}
bool SVwriter_nil::open(FILE* file)
{
  this->file = file;
  return true;
}
#endif
