/*
===============================================================================

  FILE:  SVwriter_off.h
  
  CONTENTS:
  
    Writes a Streaming Volume Mesh in a simple ASCII format.
    
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005  martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    26 June 2005 -- created after a yummy "Kartoffelauflauf" made by Muran

===============================================================================
*/
#ifndef SVWRITER_OFF_H
#define SVWRITER_OFF_H

#include "svwriter.h"

#include <stdio.h>

class SVwriter_off : public SVwriter
{
public:

  // SVwriter interface function implementations

  void write_vertex(const float* v_prop_f);

  void write_tetrahedron(const I32* c_idx, const bool* c_final=0);
  void write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final=0);

  void write_hexahedron(const I32* c_idx, const bool* c_final=0);
  void write_hexahedron(const I32* c_idx, const float* c_data_f, const bool* c_final=0);

  void write_finalized(I32 final_idx);

  void close(bool close_file=true, bool update_header=true);

  // SVwriter_off functions

  bool open(FILE* file);

  SVwriter_off();
  ~SVwriter_off();

private:
  FILE* file;
  int vertex_buffer_alloc;
  float* vertex_buffer;
  int cell_buffer_alloc;
  int* cell_buffer;
};

#endif
