/*
===============================================================================

  FILE:  SVwriter_sva.h
  
  CONTENTS:
  
    Writes a Streaming Volume Mesh in a simple ASCII format.
    
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005-08 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    15 August 2008 -- added hexahedra support after missing Henna all night long
    17 October 2007 -- changed way header is written so it can be "updated"
    21 Februar 2005 -- created after finding out how little I will earn at MPI

===============================================================================
*/
#ifndef SVWRITER_SVA_H
#define SVWRITER_SVA_H

#include "svwriter.h"

#include <stdio.h>

class SVwriter_sva : public SVwriter
{
public:

  // SVwriter interface function implementations

  void write_vertex(const float* v_prop_f);

  void write_tetrahedron(const I32* c_idx, const bool* c_final=0);
  void write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final=0);

  void write_hexahedron(const I32* c_idx, const bool* c_final=0);
  void write_hexahedron(const I32* c_idx, const float* c_data_f, const bool* c_final=0);

  void write_finalized(I32 final_idx);

  void close(bool close_file=true, bool update_header=true);

  // SVwriter_sva functions

  bool open(FILE* file);

  SVwriter_sva();
  ~SVwriter_sva();

private:
  FILE* file;
  void write_header();
  void update_header();
};

#endif
