/*
===============================================================================

  FILE:  SVwriter_svg.h

  CONTENTS:

    Writes a Streaming Volume Mesh to a binary format.
    Handles Tet and hex mixed meshes with arbitrary number of vertex and cell properties.
    see SVreader_svg.h for a description of the format


  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009 clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    6 November 2009 -- Created

===============================================================================
*/

#ifndef SVWRITER_SVG_H
#define SVWRITER_SVG_H

#include "svwriter.h"

#include <stdio.h>

class SVwriter_svg : public SVwriter
{
	public:

		// SVwriter interface function implementations

		void write_vertex ( const float* v_prop_f );

		void write_tetrahedron ( const I32* c_idx, const bool* c_final=0 );
		void write_tetrahedron ( const I32* c_idx, const float* c_prop_f, const bool* c_final=0 );

		void write_hexahedron ( const I32* c_idx, const bool* c_final=0 );
		void write_hexahedron ( const I32* c_idx, const float* c_prop_f, const bool* c_final=0 );

		void write_finalized ( I32 final_idx );

		void close ( bool close_file=true, bool update_header=true );

		// SVwriter_svg functions

		bool open ( FILE* file );

		SVwriter_svg();
		~SVwriter_svg();

	private:
		FILE* file;

		void write_header();
		void update_header();

		inline void writeInt32(const I32 *i);
		inline void writeFloat(const float *f);
		inline void writeNFloats(const float *f, int n);
		inline void writeUChar(const unsigned char *c);

		U32 update_seek_pos;
};

#endif
