/*
===============================================================================

  FILE:  SVwriter_vtk.h
  
  CONTENTS:
  
    Writes a Streaming Volume Mesh in the non-streaming VTK format.
    
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2005-08 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    10 November 2009 -- (clement) added support for cell properties
    20 October 2009 -- (clement) fixed the bug with the CELLS label not having the right <size> for hex meshes
    15 August 2008 -- added hexahedra support after missing Henna all night long
    17 September 2007 -- created after Henna skipped Pinnacles because of 'work'

===============================================================================
*/
#ifndef SVWRITER_VTK_H
#define SVWRITER_VTK_H

#include "svwriter.h"

#include <stdio.h>

class SVwriter_vtk : public SVwriter
{
public:

  // SVwriter interface function implementations

  void write_vertex(const float* v_prop_f);

  void write_tetrahedron(const I32* c_idx, const bool* c_final=0);
  void write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final=0);

  void write_hexahedron(const I32* c_idx, const bool* c_final=0);
  void write_hexahedron(const I32* c_idx, const float* c_data_f, const bool* c_final=0);

  void write_finalized(I32 final_idx);

  void close(bool close_file=true, bool update_header=true);

  // svwriter_vtk functions

  bool open(FILE* file);

  SVwriter_vtk();
  ~SVwriter_vtk();

private:
  FILE* file;
  int vertex_buffer_alloc;
  float* vertex_buffer;
  int cell_buffer_alloc;
  int* cell_buffer;
  float* cellData_buffer;
};

#endif
