/*
===============================================================================

  FILE:  SVreader_svc.cpp
  
  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2005  martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svreader_svc.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define PRINT_CONTROL_OUTPUT
#undef PRINT_CONTROL_OUTPUT

#define PRINT_DEBUG_OUTPUT
#undef PRINT_DEBUG_OUTPUT

//code that is common to svreader_svh.cpp and svwriter_svh.cpp
#include "svcommon_svc.cpp"

// rangedecoder

#include "rangedecoder.h"

static RangeDecoder* rd_conn;
static RangeDecoder* rd_conn_op;
static RangeDecoder* rd_conn_cache;
static RangeDecoder* rd_conn_index;
static RangeDecoder* rd_conn_candi;
static RangeDecoder* rd_conn_final;

static RangeDecoder* rd_geom;

// statistics

#ifdef PRINT_CONTROL_OUTPUT
static int predict_none = 0;
static int predict_last = 0;
static int predict_inside = 0;
static int predict_across = 0;

static int op_start = 0;
static int op_start_x[] = {0,0,0,0,0};
static int op_add = 0;
static int op_join = 0;
static int op_fill_end = 0;
static int op_fill = 0;
static int op_end = 0;

static int used_index = 0;
static int used_cache = 0;

static int add_miss = 0;
static int add_hit[] = {0,0,0,0,0,0};

static int fill_miss = 0;
static int fill_hit[] = {0,0,0,0,0,0,0,0,0};

static int prediction_none = 0;
static int prediction_last = 0;
static int prediction_across = 0;

static int vertex_buffer_maxsize = 0;
static int edge_buffer_maxsize = 0;
#endif

static void initDecoder(FILE* file)
{
  rd_conn = new RangeDecoder(file);
  rd_conn_op = rd_conn;
  rd_conn_cache = rd_conn;
  rd_conn_index = rd_conn;
  rd_conn_candi = rd_conn;
  rd_conn_final = rd_conn;
  rd_geom = rd_conn;
}

static void finishDecoder()
{
  rd_conn->done();
  delete rd_conn;
}

#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
static void decompressVertexPropertiesLocal(float* p0, float* p1, float* p2, float* p3)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_inside++;
#endif

  float P[3];
  float P1[3];
  float P2[3];
  float Pp[3];

  VecSubtract3fv(P1, p2, p1);
  VecSubtract3fv(P2, p0, p1);

  VecCrossProd3fv(P, P1, P2);

  VecSelfNormalize3fv(P1);
  VecSelfNormalize3fv(P);

  VecCrossProd3fv(P2, P1, P);

  VecSelfNormalize3fv(P2);

#ifdef SVC_ADD_NORMAL_COMPONENT
  VecAdd3fv(Pp, p0, p1, p2);
  VecSelfScalarDiv3fv(Pp, 3.0f);
  float normal_f[3];
  float scale = sqrt(VecCcwNormNormalReturnLength3fv(normal_f, p0, p1, p2));
  VecSelfAddScalarMult3fv(Pp, normal_f, SVC_NORMAL_SCALE*scale);
  v_pq->Clamp(Pp);
#else // SVC_ADD_NORMAL_COMPONENT
  VecZero3fv(Pp);
  VecSelfAddScalarMult3fv(Pp, p0, 1.0f/3.0f);
  VecSelfAddScalarMult3fv(Pp, p1, 1.0f/3.0f);
  VecSelfAddScalarMult3fv(Pp, p2, 1.0f/3.0f);
#endif // SVC_ADD_NORMAL_COMPONENT

  I32 a,b,c;

#ifdef SVC_ADD_NORMAL_COMPONENT
  a = local_ic[0]->Decompress(maxAbso, 2) - maxAbso;
  b = local_ic[1]->Decompress(maxAbso, 2) - maxAbso;
  c = local_ic[2]->Decompress(maxAbso, 2) - maxAbso;
#else
  a = local_ic[0]->Decompress(0, 2);
  b = local_ic[1]->Decompress(maxAbso, 2) - maxAbso;
  c = local_ic[2]->Decompress(maxAbso, 2) - maxAbso;
#endif

  float diff[3];

  diff[2] = dequantize(a);
  diff[0] = dequantize(b);
  diff[1] = dequantize(c);
  
  VecSelfAddScalarMult3fv(Pp, P1, diff[0]);
  VecSelfAddScalarMult3fv(Pp, P2, diff[1]);
  VecSelfAddScalarMult3fv(Pp, P, diff[2]);

  VecCopy3fv(p3, Pp);
}
#endif 

static void decompressVertexProperties(float* real_prop_f, I32 num_prop_f)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_none++;
#endif

  if (num_prop_f == 3)
  {
    if (v_pq)
    {
      I32* qn = (I32*)real_prop_f;
      for (int i = 0; i < 3; i++)
      {
        qn[i] = global_ic[i]->Decompress(0, 0);
      }
#ifdef SVC_STORE_QUANTIZED_COORDINATES
      I32 mist[3];
      fread(mist, sizeof(I32), 3, quantized_coordinates_file);
      if (!VecEqual3iv(mist,qn))
      {
        fprintf(stderr, "problem at v %d %d %d none\n", qn[0], qn[1], qn[2]);
      }
#endif

#ifdef PRINT_DEBUG_OUTPUT
      printf("v %d %d %d none\n", qn[0], qn[1], qn[2]);
#endif
    }
    else
    {
      for (int i = 0; i < num_prop_f; i++)
      {
        real_prop_f[i] = v_fc[i]->Decompress(0.0f, 0);
      }
    }
  }
  else
  {
    assert(false);
  }
}

static void decompressVertexProperties(float* pred_prop_f, float* real_prop_f, I32 num_prop_f)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_last++;
#endif

  if (num_prop_f == 3)
  {
    if (v_pq)
    {
      I32* ql = (I32*)pred_prop_f;
      I32* qn = (I32*)real_prop_f;
      for (int i = 0; i < 3; i++)
      {
        qn[i] = global_ic[i]->Decompress(ql[i], 1);
      }
#ifdef SVC_STORE_QUANTIZED_COORDINATES
      I32 mist[3];
      fread(mist, sizeof(I32), 3, quantized_coordinates_file);
      if (!VecEqual3iv(mist,qn))
      {
        fprintf(stderr, "problem at v %d %d %d none\n", qn[0], qn[1], qn[2]);
      }
#endif
#ifdef PRINT_DEBUG_OUTPUT
      printf("v %d %d %d (%d,%d,%d) l\n", qn[0], qn[1], qn[2], ql[0], ql[1], ql[2]);
#endif
    }
    else
    {
      for (int i = 0; i < num_prop_f; i++)
      {
        real_prop_f[i] = v_fc[i]->Decompress(pred_prop_f[i], 1);
      }
    }
  }
  else
  {
    assert(false);
  }
}

static void decompressVertexPropertiesGlobal(float* pred_prop_f0, float* pred_prop_f1, float* pred_prop_f2, float* real_prop_f, I32 num_prop_f)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_inside++;
#endif

  assert(num_prop_f == 3);

  if (v_pq)
  {
    I32 pred[3];

#ifdef SVC_ADD_NORMAL_COMPONENT
    if (SVC_NORMAL_SCALE == -8.0f)
    {
      VecAdd3iv(pred, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfScalarDiv3iv(pred, 3);
    }
    else if (SVC_NORMAL_SCALE == -4.0f)
    {
      VecAdd3iv(pred, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfAdd3iv(pred, 1);
      VecSelfScalarDiv3iv(pred, 3);
    }
    else if (SVC_NORMAL_SCALE == -2.0f)
    {
      float pred_f[3];
      VecAdd3fv(pred_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfScalarDiv3fv(pred_f, 3.0f);
      v_pq->Clamp(pred_f, pred);
    }
    else
    {
      float pred_f[3];
      VecAdd3fv(pred_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
      VecSelfScalarDiv3fv(pred_f, 3.0f);
      float normal_f[3];
      float scale = sqrt(VecCcwNormNormalReturnLength3fv(normal_f, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2));
      VecSelfAddScalarMult3fv(pred_f, normal_f, SVC_NORMAL_SCALE*scale);
      v_pq->Clamp(pred_f, pred);
    }
#else // SVC_ADD_NORMAL_COMPONENT
    VecAdd3iv(pred, (I32*)pred_prop_f0, (I32*)pred_prop_f1, (I32*)pred_prop_f2);
    VecSelfScalarDiv3iv(pred, 3);
#endif // SVC_ADD_NORMAL_COMPONENT

    I32* qn = (I32*)real_prop_f;
    for (int i = 0; i < 3; i++)
    {
      qn[i] = global_ic[i]->Decompress(((I32*)pred)[i], 2);
    }
#ifdef SVC_STORE_QUANTIZED_COORDINATES
    I32 mist[3];
    fread(mist, sizeof(I32), 3, quantized_coordinates_file);
    if (!VecEqual3iv(mist,(I32*)pred_prop_f0))
    {
      fprintf(stderr, "problem at v %d %d %d none\n", pred_prop_f0[0], pred_prop_f0[1], pred_prop_f0[2]);
    }
    fread(mist, sizeof(I32), 3, quantized_coordinates_file);
    if (!VecEqual3iv(mist,(I32*)pred_prop_f1))
    {
      fprintf(stderr, "problem at v %d %d %d none\n", pred_prop_f1[0], pred_prop_f1[1], pred_prop_f1[2]);
    }
    fread(mist, sizeof(I32), 3, quantized_coordinates_file);
    if (!VecEqual3iv(mist,(I32*)pred_prop_f2))
    {
      fprintf(stderr, "problem at v %d %d %d none\n", pred_prop_f2[0], pred_prop_f2[1], pred_prop_f2[2]);
    }
    fread(mist, sizeof(I32), 3, quantized_coordinates_file);
    if (!VecEqual3iv(mist,(I32*)pred))
    {
      fprintf(stderr, "problem at v %d %d %d none\n", pred[0], pred[1], pred[2]);
    }
    fread(mist, sizeof(I32), 3, quantized_coordinates_file);
    if (!VecEqual3iv(mist,qn))
    {
      fprintf(stderr, "problem at v %d %d %d none\n", qn[0], qn[1], qn[2]);
    }
#endif
#ifdef PRINT_DEBUG_OUTPUT
    printf("v %d %d %d (%d,%d,%d) a\n", qn[0], qn[1], qn[2], pred[0], pred[1], pred[2]);
#endif
  }
  else
  {
    float pred[3];
    VecAdd3fv(pred, pred_prop_f0, pred_prop_f1, pred_prop_f2);
    VecSelfScalarDiv3fv(pred, 3.0f);
    for (int i = 0; i < num_prop_f; i++)
    {
      real_prop_f[i] = v_fc[i]->Decompress(pred[i], 2);
    }
  }
}

bool SVreader_svc::open(FILE* file)
{
  if (file == 0)
  {
    return false;
  }

  this->file = file;

#ifdef _WIN32
  if (file == stdin)
  {
    if(_setmode( _fileno( stdin ), _O_BINARY ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdin to binary (untranslated) mode\n");
    }
  }
#endif

  // read version
  I32 version = fgetc(file);

  // compiled version
  I32 compiled_version;

#ifdef SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    compiled_version = 3;
  #else
    compiled_version = 2;
  #endif
#elif SVC_DO_FLIP_PREDICTION
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    compiled_version = 5;
  #else
    compiled_version = 4;
  #endif
#elif SVC_ADD_NORMAL_COMPONENT
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    compiled_version = 7;
  #else
    compiled_version = 6;
  #endif
#else // simple mid-point
  #ifdef SVC_USE_LOCAL_COORDINATE_SYSTEM
    compiled_version = 1;
  #else
    compiled_version = 0;
  #endif
#endif

  if (version != compiled_version)
  {
    fprintf(stderr,"ERROR: wrong SVreader (data is %d but compiled reader is %d)\n",version, compiled_version);
    if (version == EOF) fprintf(stderr,"       check ... maybe the file (or the stream) is empty\n");
    exit(0);
  }

  initDecoder(file);
  initModels(0);

  // read precision
  nbits = rd_conn->readRange(25);

  read_header();

  last_op = 0;
  last_fin = 0;

  initVertexBuffer(1024, v_pnum_f);
  initEdgeBuffer(1024);

  dv = new DynamicVector();
  lc = new LittleCache(4);
  lc_f = new LittleCache(4);

  v_count = 0;
  c_count = 0;

  if (v_pnum_f) v_prop_f = new float[v_pnum_f];
  if (c_pnum_f) c_prop_f = new float[c_pnum_f];

  // currently only TET meshes are supported
  type = SV_TET;
  c_nverts = 4;
  c_idx = new I32[4];
  c_final = new bool[4];
  new_vertices = new I32[4]; 
  finalized_vertices = new I32[4]; 

#ifdef SVC_STORE_QUANTIZED_COORDINATES
  quantized_coordinates_file = fopen("temp_file.bin","rb");
#endif

  return true;
}

static SVvertex* vertices[4];

I32 SVreader_svc::decompress_tetrahedron()
{
  int i,j;
  int op;
  int dv_index;
  int lc_pos;
  int candidate;
  int candidate_count;
  int degree_one;
  int use_count;

  SVedge* f_visited[4]={NULL,NULL,NULL,NULL};
 
  if (edge_buffer_size)
  {
//    op = rd_conn_op->decode(rmOp[last_op]);
    op = rd_conn_op->decode(rmOpX[last_fin][last_op]);
  }
  else
  {
    if (rd_conn_op->decode(rmDone))
    {
      return 0;
    }
    op = SVC_START;
  }

  if (op == SVC_START)
  {
    // specify how many old (non-manifold) vertices there are
    int num_v_visited = rd_conn_op->decode(rmS_Old);

    if (num_v_visited == 5)
    {
      return 0; // end of streaming mesh (despite non-finalized vertices)
    }

    // decompress the vertices

    for (int i = 0; i < 4; i++)
    {
      if (num_v_visited)
      {
        // decode a reference to v[i] (-> improve with cache of with prediction)
        lc_pos = rd_conn_cache->decode(rmS_Cache[i]);
        if (lc_pos == 4) //get it from the dynamic vector
        {
          dv_index = rd_conn_index->readRange(dv->size());        
          vertices[i] = (SVvertex*)dv->getElementWithRelativeIndex(dv_index); 
        }
        else //get it from the little cache
        {
          vertices[i] = (SVvertex*)lc->get(lc_pos);
        }
        // decrease counter of visited vertices
        num_v_visited--;
      }
      else
      {
        // create a new vertex 
        vertices[i] = allocVertex();
        // give it its index
        vertices[i]->index = v_count + have_new;
        // decompress its position
        if (i == 3)
        {
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
          decompressVertexPropertiesLocal(vertices[0]->prop_f, vertices[1]->prop_f, vertices[2]->prop_f, vertices[3]->prop_f);
#else
          decompressVertexPropertiesGlobal(vertices[0]->prop_f, vertices[1]->prop_f, vertices[2]->prop_f, vertices[3]->prop_f, v_pnum_f);
#endif
        }
        else
        {
          if (i > 0)
          {
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
            if (v_pq)
            {
              I32 temp[3];
              v_pq->EnQuantize(vertices[i-1]->prop_f, temp);
              decompressVertexProperties((float*)temp, vertices[i]->prop_f, v_pnum_f);
              v_pq->DeQuantize((I32*)(vertices[i]->prop_f), vertices[i]->prop_f);
            }
            else
            {
              decompressVertexProperties(vertices[i-1]->prop_f, vertices[i]->prop_f, v_pnum_f);
            }
#else
            decompressVertexProperties(vertices[i-1]->prop_f, vertices[i]->prop_f, v_pnum_f);
#endif
          }
          else
          {
            decompressVertexProperties(vertices[i]->prop_f, v_pnum_f);
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
            if (v_pq)
            {
              v_pq->DeQuantize((I32*)(vertices[i]->prop_f), vertices[i]->prop_f);
            }
#endif
          }
        }
        // insert it into the indexable data structure
        dv->addElement(vertices[i]);
        // enlist it for being output as a new vertex
        new_vertices[have_new] = i;
        have_new++;
      }
    }
    f_visited[0] = f_visited[1] = f_visited[2] = f_visited[3] = 0;
  }
  else if (op == SVC_ADD || op == SVC_JOIN)
  {
    // decode face 0

//    lc_pos = rd_conn_cache->decode(rmAJ_FaceCache[last_op]);
    lc_pos = rd_conn_cache->decode(rmAJ_FaceCacheX[last_fin][last_op]);

    if (lc_pos < 4) //f0 is one of the faces of the previously added tet, so it is in little face/halfedge cache
    {
      f_visited[0] = (SVedge*)lc_f->get(lc_pos);
      vertices[0] = f_visited[0]->target;
      vertices[1] = f_visited[0]->next->next->target;
      vertices[2] = f_visited[0]->next->target;
    }
    else
    {
      lc_pos = rd_conn_cache->decode(rmAJ_EdgeCache[last_op]);
      if (lc_pos != 4) //f0 shares an edge with the previous tet, lc_pos is the source vertex of this shared halfedge.
      {
      	// (lc_pos+1) has 4 bits (14 possible values = choose two vertices in four, order counting), the two MSB give the position of v1, the two LSB the position of v0. 
        vertices[1] = (SVvertex*)lc->get((lc_pos+1)/4);
        vertices[0] = (SVvertex*)lc->get((lc_pos+1)%4);

        // decode which of the faces incident to v1 and v0 is f0
        //    first ... how many potential candidates are around the vertex with the shorter face list
        candidate_count = 0;
        if (vertices[0]->list_size < vertices[1]->list_size)
        {
          for (i = 0; i < vertices[0]->list_size; i++)
          {
            if (vertices[0]->list[i]->next->target == vertices[1])
            {
              f_visited[0] = vertices[0]->list[i]->next->next;
              candidate_count++;
            }
          }
          assert(candidate_count);
          if (candidate_count > 1)
          {
            if (candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if (candidate != candidate_count - 1)
            {
              for (i = 0; i < vertices[0]->list_size; i++)
              {
                if (vertices[0]->list[i]->next->target == vertices[1])
                {
                  if (candidate == 0)
                  {
                    f_visited[0] = vertices[0]->list[i]->next->next;
                    break;
                  }
                  else
                  {
                    candidate--;
                  }
                }
              }
            }
          }
        }
        else
        {          
          for (i = 0; i < vertices[1]->list_size; i++)
          {
            if (vertices[1]->list[i]->target == vertices[0])
            {
              f_visited[0] = vertices[1]->list[i];
              candidate_count++;
            }
          }
          assert(candidate_count);
          if (candidate_count > 1)
          {
            if (candidate_count == 2)
            {
              candidate = rd_conn_candi->decode(rm_FaceCandi);
            }
            else
            {
              candidate = rd_conn_candi->readRange(candidate_count);
            }
            if (candidate != candidate_count - 1)
            {
              for (i = 0; i < vertices[1]->list_size; i++)
              {
                if (vertices[1]->list[i]->target == vertices[0])
                {
                  if (candidate == 0)
                  {
                    f_visited[0] = vertices[1]->list[i];
                    break;
                  }
                  else
                  {
                    candidate--;
                  }
                }
              }
            }
          }
        }
        vertices[2] = f_visited[0]->next->target;
      }
      else //f0 is not a face of the previous tet and does not share an edge with the previous tet
      {
        lc_pos = rd_conn_cache->decode(rmAJ_VertexCache[last_op]);
        if (lc_pos < 4) //f0 shares a vertex with the previous tet
        {
          vertices[1] = (SVvertex*)lc->get(lc_pos);
        }
        else //f0 does not share a vertex with the previous tet, read the explicit index.
        {
          dv_index = rd_conn_index->readRange(dv->size());
          vertices[1] = (SVvertex*)dv->getElementWithRelativeIndex(dv_index);
        }

        // decode which of the faces incident to v1 is f0
        if (vertices[1]->list_size > 1) 
        {
          if (vertices[1]->list_size < 14) //NOTE:why 14 ?
          {
            i = rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]);
          }
          else
          {
            i = rd_conn->readRange(vertices[1]->list_size);
          }
        }
        else
        {
          i = 0;
        }
        f_visited[0] = vertices[1]->list[i];
        vertices[0] = f_visited[0]->target;
        vertices[2] = f_visited[0]->next->target;
      }
    }

    // decode v3
    if (op == SVC_JOIN)
    {
      degree_one = rd_conn->decode(rmAJ_LocalJoin[last_op]);

      if (degree_one < 6)
      {
        vertices[3] = vertices[degree_one/2]; // vertices[degree_one/2] is the vertex at which begins the halfedge that is shared
        candidate = rd_conn->readRange(vertices[3]->list_size); //now decode which of the halfedges that begin at vertices[degree_one/2] that is.
        if (degree_one%2)
        {
          vertices[3] = vertices[3]->list[candidate]->next->target;
        }
        else
        {
          vertices[3] = vertices[3]->list[candidate]->target;
        }
      }
      else //explicitely read the dynamic index.
      {
        // old position
        dv_index = rd_conn_index->readRange(dv->size());
        vertices[3] = (SVvertex*)dv->getElementWithRelativeIndex(dv_index);
      }
    }
    else // (op == SVC_ADD)
    {
      // create a new vertex 
      vertices[3] = allocVertex();
      // give it its index
      vertices[3]->index = v_count + have_new;
      // decompress its position
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
      decompressVertexPropertiesLocal(vertices[0]->prop_f, vertices[1]->prop_f, vertices[2]->prop_f, vertices[3]->prop_f);
#else
      decompressVertexPropertiesGlobal(vertices[0]->prop_f, vertices[1]->prop_f, vertices[2]->prop_f, vertices[3]->prop_f, v_pnum_f);
#endif
      // insert it into the indexable data structure
      dv->addElement(vertices[3]);
      // enlist it for being output as a new vertex
      new_vertices[have_new] = 3;
      have_new++;
    }
    f_visited[1] = f_visited[2] = f_visited[3] = 0;
  }
  else if (op == SVC_FLIP)
  {
    // decode face 0
//    lc_pos = rd_conn_cache->decode(rmFE_FaceCache[last_op]);
    lc_pos = rd_conn_cache->decode(rmFE_FaceCacheX[last_fin][last_op]);
    // maybe face 0 is in the face cache
    if (lc_pos < 12) //f0 can be any of the three halfedges of any of the 4 faces of the previous tet
    {
      if (lc_pos < 4)
      {
        f_visited[0] = (SVedge*)lc_f->get(lc_pos);
      }
      else if (lc_pos < 8)
      {
        f_visited[0] = (SVedge*)lc_f->get(lc_pos-4);
        f_visited[0] = f_visited[0]->next->next;
      }
      else
      {
        f_visited[0] = (SVedge*)lc_f->get(lc_pos-8);
        f_visited[0] = f_visited[0]->next;
      }
      vertices[0] = f_visited[0]->target;
      vertices[1] = f_visited[0]->next->next->target;
      vertices[2] = f_visited[0]->next->target;
    }
    else
    {
      // maybe two vertices (an edge) are in the vertex cache
      //2 possible edges in any 6 possible positions, in any 2 possible directions, is 24 possibilities (NOTE:martin counted the possibilities where the vertices are the same, hence his 28 possibilities )
      lc_pos = rd_conn_cache->decode(rmFE_EdgeCache[last_op]);

      if (lc_pos != 4)
      {
        candidate_count = 0;

        if (lc_pos < 14)
        {
          // v1 and v2 are in the vertex cache 
	  // (lc_pos+1) has 4 bits (14 possible values = choose two vertices in four, order counting), the two MSB give the position of v1, the two LSB the position of v2. 
          vertices[1] = (SVvertex*)lc->get((lc_pos+1)/4);
          vertices[2] = (SVvertex*)lc->get((lc_pos+1)%4);

          // get face 0 that is adjacent to v1 and v2
          if (vertices[1]->list_size < vertices[2]->list_size)
          {
            for (i = 0; i < vertices[1]->list_size; i++)
            {
              if (vertices[1]->list[i]->next->target == vertices[2])
              {
                f_visited[0] = vertices[1]->list[i];
                candidate_count++;
              }
            }
            assert(candidate_count);
            if (candidate_count > 1) 
            {
              if (candidate_count == 2)
              {
                candidate = rd_conn_candi->decode(rm_FaceCandi);
              }
              else
              {
                candidate = rd_conn_candi->readRange(candidate_count);
              }
              if (candidate != candidate_count-1)
              {
                for (i = 0; i < vertices[1]->list_size; i++)
                {
                  if (vertices[1]->list[i]->next->target == vertices[2])
                  {
                    if (candidate == 0)
                    {
                      assert(f_visited[0] && f_visited[0] != vertices[1]->list[i]);
                      f_visited[0] = vertices[1]->list[i];
                      break;
                    }
                    else
                    {
                      candidate--;
                    }
                  }
                }
              }
            }
          }
          else
          {
            for (i = 0; i < vertices[2]->list_size; i++)
            {
              if (vertices[2]->list[i]->target == vertices[1])
              {
                f_visited[0] = vertices[2]->list[i]->next;
                candidate_count++;
              }
            }
            assert(candidate_count);
            if (candidate_count > 1)
            {
              if (candidate_count == 2)
              {
                candidate = rd_conn_candi->decode(rm_FaceCandi);
              }
              else
              {
                candidate = rd_conn_candi->readRange(candidate_count);
              }
              if (candidate != candidate_count - 1)
              {
                for (i = 0; i < vertices[2]->list_size; i++)
                {
                  if (vertices[2]->list[i]->target == vertices[1])
                  {
                    if (candidate == 0)
                    {
                      assert(f_visited[0] && f_visited[0] != vertices[2]->list[i]->next);
                      f_visited[0] = vertices[2]->list[i]->next;
                      break;
                    }
                    else
                    {
                      candidate--;
                    }
                  }
                }
              }
            }
          }
          vertices[0] = f_visited[0]->target;
        }
        else
        {
	  // v0 and v2 are in the vertex cache (same remark as above)
          vertices[0] = (SVvertex*)lc->get((lc_pos+1-14)/4);
          vertices[2] = (SVvertex*)lc->get((lc_pos+1-14)%4);

          // get face 0 that is adjacent to v0 and v2
          if (vertices[0]->list_size < vertices[2]->list_size)
          {
            for (i = 0; i < vertices[0]->list_size; i++)
            {
              if (vertices[0]->list[i]->target == vertices[2])
              {
                f_visited[0] = vertices[0]->list[i]->next->next;
                candidate_count++;
              }
            }
            assert(candidate_count);
            if (candidate_count > 1) 
            {
              if (candidate_count == 2)
              {
                candidate = rd_conn_candi->decode(rm_FaceCandi);
              }
              else
              {
                candidate = rd_conn_candi->readRange(candidate_count);
              }
              if (candidate != candidate_count-1)
              {
                for (i = 0; i < vertices[0]->list_size; i++)
                {
                  if (vertices[0]->list[i]->target == vertices[2])
                  {
                    if (candidate == 0)
                    {
                      assert(f_visited[0] && f_visited[0] != vertices[0]->list[i]->next->next);
                      f_visited[0] = vertices[0]->list[i]->next->next;
                      break;
                    }
                    else
                    {
                      candidate--;
                    }
                  }
                }
              }
            }
          }
          else
          {
            for (i = 0; i < vertices[2]->list_size; i++)
            {
              if (vertices[2]->list[i]->next->target == vertices[0])
              {
                f_visited[0] = vertices[2]->list[i]->next;
                candidate_count++;
              }
            }
            assert(candidate_count);
            if (candidate_count > 1)
            {
              if (candidate_count == 2)
              {
                candidate = rd_conn_candi->decode(rm_FaceCandi);
              }
              else
              {
                candidate = rd_conn_candi->readRange(candidate_count);
              }
              if (candidate != candidate_count - 1)
              {
                for (i = 0; i < vertices[2]->list_size; i++)
                {
                  if (vertices[2]->list[i]->next->target == vertices[0])
                  {
                    if (candidate == 0)
                    {
                      assert(f_visited[0] && f_visited[0] != vertices[2]->list[i]->next);
                      f_visited[0] = vertices[2]->list[i]->next;
                      break;
                    }
                    else
                    {
                      candidate--;
                    }
                  }
                }
              }
            }
          }
          vertices[1] = f_visited[0]->next->next->target;
        }
      }
      else
      {
        // maybe vertex 1 or vertex 2 are in the vertex cache
        lc_pos = rd_conn_cache->decode(rmFE_VertexCache[last_op]);
        if (lc_pos < 8) //2 possible verts in any 4 possible positions is 8 possibilities
        {
          if (lc_pos < 4)
          {
            // decode vertex 1
            vertices[1] = (SVvertex*)lc->get(lc_pos);
          }
          else
          {
            // decode vertex 2
            vertices[2] = (SVvertex*)lc->get(lc_pos-4);
          }
        }
        else
        {
          // decode vertex 1
          dv_index = rd_conn_index->readRange(dv->size());
          vertices[1] = (SVvertex*)dv->getElementWithRelativeIndex(dv_index);
          lc_pos = -1;
        }

        if (lc_pos < 4) //lc_pos was <4 and v1 was in cache, or lc_pos was 8 and v1 was explicitely given
        {
          if (vertices[1]->list_size > 1)
          {
            if (vertices[1]->list_size < 14)
            {
              i = rd_conn->decode(rm_VertexCandi[vertices[1]->list_size]);
            }
            else
            {
              i = rd_conn->readRange(vertices[1]->list_size);
            }
          }
          else
          {
            i = 0;
          }
          f_visited[0] = vertices[1]->list[i];
          vertices[0] = f_visited[0]->target;
          vertices[2] = f_visited[0]->next->target;
        }
        else //lc_pos was in [4,7] and so we know v2 
        {
          if (vertices[2]->list_size > 1)
          {
            if (vertices[2]->list_size < 14)
            {
              i = rd_conn->decode(rm_VertexCandi[vertices[2]->list_size]);
            }
            else
            {
              i = rd_conn->readRange(vertices[2]->list_size);
            }
          }
          else
          {
            i = 0;
          }
          f_visited[0] = vertices[2]->list[i]->next;
          vertices[0] = f_visited[0]->target;
          vertices[1] = f_visited[0]->next->next->target;
        }
      }
    }

    // decode which of the faces incident to v0 and v1 is f1
    candidate_count = 0;
    candidate = 0;
    vertices[3] = 0;
    f_visited[1] = 0;
    // how many potential candidates are around the vertex with the shorter face list
    if (vertices[0]->list_size < vertices[1]->list_size)
    {
      for (i = 0; i < vertices[0]->list_size; i++)
      {
        if (vertices[0]->list[i]->target == vertices[1])
        {
          f_visited[1] = vertices[0]->list[i];
          candidate_count++;
        }
      }
      assert(candidate_count);
      if (candidate_count > 1) 
      {
        if (candidate_count == 2)
        {
          candidate = rd_conn_candi->decode(rm_FaceCandi);
        }
        else
        {
          candidate = rd_conn_candi->readRange(candidate_count);
        }
        if (candidate != candidate_count-1)
        {
          for (i = 0; i < vertices[0]->list_size; i++)
          {
            if (vertices[0]->list[i]->target == vertices[1])
            {
              if (candidate == 0)
              {
                assert(f_visited[1] != vertices[0]->list[i]);
                f_visited[1] = vertices[0]->list[i];
                break;
              }
              else
              {
                candidate--;
              }
            }
          }
        }
      }
      vertices[3] = f_visited[1]->next->target;
    }
    else
    {
      for (i = 0; i < vertices[1]->list_size; i++)
      {
        if (vertices[1]->list[i]->next->target == vertices[0])
        {
          f_visited[1] = vertices[1]->list[i]->next->next;
          candidate_count++;
        }
      }
      assert(candidate_count);
      if (candidate_count > 1)
      {
        if (candidate_count == 2)
        {
          candidate = rd_conn_candi->decode(rm_FaceCandi);
        }
        else
        {
          candidate = rd_conn_candi->readRange(candidate_count);
        }
        if (candidate != candidate_count - 1)
        {
          for (i = 0; i < vertices[1]->list_size; i++)
          {
            if (vertices[1]->list[i]->next->target == vertices[0])
            {
              if (candidate == 0)
              {
                assert(f_visited[1] != vertices[1]->list[i]->next->next);
                f_visited[1] = vertices[1]->list[i]->next->next;
                break;
              }
              else
              {
                candidate--;
              }
            }
          }
        }
      }
      vertices[3] = f_visited[1]->next->target;
    }
    assert(vertices[3] && vertices[3] != vertices[2] && vertices[3] != vertices[1] && vertices[3] != vertices[0]);

    // figure out if this was maybe a SVC_FILL or even a SVC_END operation

    // check for existance of f2
    f_visited[2] = 0;
    for (i = 0; i < vertices[3]->list_size; i++)
    {
      if (vertices[3]->list[i]->next->target == vertices[2])
      {
        if (vertices[3]->list[i]->target == vertices[1])
        {
          op = SVC_FILL;
          f_visited[2] = vertices[3]->list[i]->next;
          break;
        }
      }
    }
    // check for existance of f3
    f_visited[3] = 0;
    for (i = 0; i < vertices[3]->list_size; i++)
    {
      if (vertices[3]->list[i]->target == vertices[2])
      {
        if (vertices[3]->list[i]->next->target == vertices[0])
        {
          op = SVC_END;
          assert(f_visited[2]);
          f_visited[3] = vertices[3]->list[i]->next;
          break;
        }
      }
    }
  }
  else
  {
    assert(0);
  }

  // statistic

//  fprintf(stderr,"op %d %d\n",op,c_count); 

//  fprintf(stderr,"%d %d %d %d\n",vertices[0]->index, vertices[1]->index, vertices[2]->index, vertices[3]->index);

  // update data structures

  if (f_visited[0])
  {
    removeEdgeFromVertex(f_visited[0]->next->next, vertices[2]);
    removeEdgeFromVertex(f_visited[0]->next, vertices[0]);
    removeEdgeFromVertex(f_visited[0], vertices[1]);

    deallocEdge( f_visited[0]->next->next);
    deallocEdge( f_visited[0]->next);
    deallocEdge( f_visited[0]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[0]->target->index, f_visited[0]->next->target->index, f_visited[0]->next->next->target->index);
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e0;

    e0->target = vertices[1];
    e1->target = vertices[2];
    e2->target = vertices[0];

    addEdgeToVertex(e0, vertices[0]);
    addEdgeToVertex(e1, vertices[1]);
    addEdgeToVertex(e2, vertices[2]);

    f_visited[0] = e0;
//    fprintf(stderr,"add %d %d %d\n",e0->target->index, e0->next->target->index, e0->next->next->target->index);
  }

  if (f_visited[1])
  {
    removeEdgeFromVertex(f_visited[1]->next->next, vertices[3]);
    removeEdgeFromVertex(f_visited[1]->next, vertices[1]);
    removeEdgeFromVertex(f_visited[1], vertices[0]);

    deallocEdge(f_visited[1]->next->next);
    deallocEdge(f_visited[1]->next);
    deallocEdge(f_visited[1]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[1]->target->index, f_visited[1]->next->target->index, f_visited[1]->next->next->target->index);
  }
  else
  {
    SVedge* e00 = allocEdge();
    SVedge* e01 = allocEdge();
    SVedge* e02 = allocEdge();

    e00->next = e01;
    e01->next = e02;
    e02->next = e00;

    e00->target = vertices[0];
    e01->target = vertices[3];
    e02->target = vertices[1];

    addEdgeToVertex(e00, vertices[1]);
    addEdgeToVertex(e01, vertices[0]);
    addEdgeToVertex(e02, vertices[3]);

    f_visited[1] = e00;
//    fprintf(stderr,"add %d %d %d\n",e00->target->index, e00->next->target->index, e00->next->next->target->index);
  }

  if (f_visited[2])
  {
    removeEdgeFromVertex(f_visited[2]->next->next, vertices[3]);
    removeEdgeFromVertex(f_visited[2]->next, vertices[2]);
    removeEdgeFromVertex(f_visited[2], vertices[1]);

    deallocEdge(f_visited[2]->next->next);
    deallocEdge(f_visited[2]->next);
    deallocEdge(f_visited[2]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[2]->target->index, f_visited[2]->next->target->index, f_visited[2]->next->next->target->index);
  }
  else
  {
    SVedge* e10 = allocEdge();
    SVedge* e11 = allocEdge();
    SVedge* e12 = allocEdge();

    e10->next = e11;
    e11->next = e12;
    e12->next = e10;

    e10->target = vertices[1];
    e11->target = vertices[3];
    e12->target = vertices[2];

    addEdgeToVertex(e10, vertices[2]);
    addEdgeToVertex(e11, vertices[1]);
    addEdgeToVertex(e12, vertices[3]);

    f_visited[2] = e10;
//    fprintf(stderr,"add %d %d %d\n",e10->target->index, e10->next->target->index, e10->next->next->target->index);
  }

  if (f_visited[3])
  {
    removeEdgeFromVertex(f_visited[3]->next->next, vertices[3]);
    removeEdgeFromVertex(f_visited[3]->next, vertices[0]);
    removeEdgeFromVertex(f_visited[3], vertices[2]);

    deallocEdge(f_visited[3]->next->next);
    deallocEdge(f_visited[3]->next);
    deallocEdge(f_visited[3]);

//    fprintf(stderr,"remove %d %d %d\n",f_visited[3]->target->index, f_visited[3]->next->target->index, f_visited[3]->next->next->target->index);
  }
  else
  {
    SVedge* e20 = allocEdge();
    SVedge* e21 = allocEdge();
    SVedge* e22 = allocEdge();

    e20->next = e21;
    e21->next = e22;
    e22->next = e20;

    e20->target = vertices[2];
    e21->target = vertices[3];
    e22->target = vertices[0];

    addEdgeToVertex(e20, vertices[0]);
    addEdgeToVertex(e21, vertices[2]);
    addEdgeToVertex(e22, vertices[3]);

    f_visited[3] = e20;
//    fprintf(stderr,"add %d %d %d\n",e20->target->index, e20->next->target->index, e20->next->next->target->index);
  }

  // update cache

  lc->put(vertices[0], vertices[1], vertices[2], vertices[3]);
  lc_f->put(f_visited[0], f_visited[1], f_visited[2], f_visited[3]);

//  if (c_count < 22)
//  {
//    fprintf(stderr,"%d %d %d %d\n",c_count,last_op,last_fin,op);
//  }

  // update last_op

  last_op = op;

  // increment vertex use_counts and copy tetrahedra indices into API

  for (i = 0; i < 4; i++)
  {
    vertices[i]->use_count++;
    c_idx[i] = vertices[i]->index;
  }

  last_fin = 0;

  // decode finalization info and finalize vertices
  for (i = 0; i < 4; i++)
  {
    degree_one = vertices[i]->list_size;
    if (degree_one >= SVC_MAX_DEGREE_ONE)
    {
      degree_one = SVC_MAX_DEGREE_ONE-1;
    }
    use_count = vertices[i]->use_count;
    if (use_count >= SVC_MAX_USE_COUNT)
    {
      use_count = SVC_MAX_USE_COUNT-1;
    }
    c_final[i] = (rd_conn_final->decode(rmFinalized[degree_one][use_count]) == 1);

//    fprintf(stderr, "fin %d %d %d %d\n", degree_one, use_count, c_final[i], last_op);

    if (c_final[i])
    {
      SVvertex* vertex = vertices[i];

      if (i < 2) last_fin = last_fin | (1<<i);
//      if (c_count < 22) fprintf(stderr,"finalize %d %d %d\n",i,vertices[i]->index,last_fin);

//      fprintf(stderr,"remove vertex %d\n",vertex->index);
      dv->removeElement(vertex);
      for (j = 0; j < vertex->list_size; j++)
      {
        SVedge* edge = vertex->list[j];
        removeEdgeFromVertex(edge->next->next, edge->next->target);
        removeEdgeFromVertex(edge->next, edge->target);
//        removeEdgeFromVertex(edge, edge->next->next->target);
        deallocEdge(edge->next->next);
        deallocEdge(edge->next);
        deallocEdge(edge);
      }
      deallocVertex(vertex);
      finalized_vertices[have_finalized] = i;
      have_finalized++;
    }
  }

//  for (i = 0; i < 4; i++) if (vertices[i]->index == 699) fprintf(stderr,"%d %d %d %d\n",c_count, vertices[i]->use_count,vertices[i]->list_size);

  have_tetrahedron = 1;
  return 1;
}

void SVreader_svc::close(bool close_file)
{
  // close of SVreader_svc interface
  finishDecoder();
  if (v_pq)
  {
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"predict_none %d predict_last %d predict_inside %d predict_across %d\n", predict_none, predict_last, predict_inside, predict_across);
#endif
    global_ic[0]->FinishDecompressor();
    global_ic[1]->FinishDecompressor();
    global_ic[2]->FinishDecompressor();
    delete global_ic[0];
    delete global_ic[1];
    delete global_ic[2];
    delete v_pq;
  }
  else
  {
    v_fc[0]->FinishDecompressor();
    v_fc[1]->FinishDecompressor();
    v_fc[2]->FinishDecompressor();
    delete v_fc[0];
    delete v_fc[1];
    delete v_fc[2];
  }
  finishModels();

  destroyVertexBuffer();
  destroyEdgeBuffer();

  if (dv->size()) fprintf(stderr,"WARNING: there are %d unfinalized vertices\n",dv->size());
  delete dv;
  delete lc;
  delete lc_f;
  nbits = -1;

  if (new_vertices) delete [] new_vertices;
  new_vertices = 0;
  if (finalized_vertices) delete [] finalized_vertices;
  finalized_vertices = 0;

#ifdef SVC_STORE_QUANTIZED_COORDINATES
  fclose(quantized_coordinates_file);
#endif

  // close of SVreader interface
  v_count = -1;
  c_count = -1;
  if (v_prop_f) delete [] v_prop_f;
  v_prop_f = 0;
  if (c_idx) delete [] c_idx;
  c_idx = 0;
  if (c_final) delete [] c_final;
  c_final = 0; 
  if (close_file && file && file != stdin) fclose(file);
}

void SVreader_svc::read_header()
{
  int i;

  // read nverts
  if (rd_conn->readBit())
  {
    nverts = rd_conn->readInt();
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"nverts: %d\n",nverts);
#endif
  }

  // read ncells
  if (rd_conn->readBit())
  {
    ncells = rd_conn->readInt();
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"ncells: %d\n",ncells);
#endif
  }

  // read which properties are there
  
  int prop = rd_geom->readByte();
#ifdef PRINT_CONTROL_OUTPUT
  printf("property mapping is %d\n",prop);
#endif

  if (prop & 1)
  {
    v_pnum_f = rd_geom->readByte();
#ifdef PRINT_CONTROL_OUTPUT
    printf("have %d float vertex attributes\n",v_pnum_f);
#endif
    if (rd_geom->readBit())
    {
      if (v_pmin_f) delete [] v_pmin_f;
      v_pmin_f = new float[v_pnum_f];
      if (v_pmax_f) delete [] v_pmax_f;
      v_pmax_f = new float[v_pnum_f];
      for (i=0; i < v_pnum_f; i++)
      {
        v_pmin_f[i] = rd_geom->readFloat();
        v_pmax_f[i] = rd_geom->readFloat();
      }
    }
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no float vertex attributes\n");
#endif
  }

  if (prop & 2)
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have integer vertex attributes\n");
#endif
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no integer vertex attributes\n");
#endif
  }

  if (prop & 4)
  {
    c_pnum_f = rd_geom->readByte();
    if (rd_geom->readBit())
    {
      if (c_pmin_f) delete [] c_pmin_f;
      c_pmin_f = new float[c_pnum_f];
      if (c_pmax_f) delete [] c_pmax_f;
      c_pmax_f = new float[c_pnum_f];
      for (i=0; i < c_pnum_f; i++)
      {
        c_pmin_f[i] = rd_geom->readFloat();
        c_pmax_f[i] = rd_geom->readFloat();
      }
    }
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no float cell attributes\n");
#endif
  }

  if (prop & 8)
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have integer cell attributes\n");
#endif
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no integer cell attributes\n");
#endif
  }

  if (rd_conn->readBit()) // if we have comments
  {
    /* yet to be implemented */
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no comments\n");
#endif
  }

  // setup float quantizers and integer compressors

  int high_bits = rd_conn->readRange(17);

#ifdef PRINT_CONTROL_OUTPUT
  printf("high_bits %d\n", high_bits);
#endif

  int cbm_precision = 0;

  if (v_pmin_f && v_pmax_f && nbits)
  {
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"dequantizing v_prop_f with %d bits\n",nbits);
#endif
    v_pq = new PositionQuantizerNew();
    v_pq->SetMinMax(v_pmin_f, v_pmax_f);
    v_pq->SetPrecision(nbits);

    cbm_precision = rd_conn->readBit();

    if (cbm_precision)
    {
/* simulating CBM precision */
      float diagRange = (float)sqrt((v_pmax_f[0]-v_pmin_f[0])*(v_pmax_f[0]-v_pmin_f[0])+(v_pmax_f[1]-v_pmin_f[1])*(v_pmax_f[1]-v_pmin_f[1])+(v_pmax_f[2]-v_pmin_f[2])*(v_pmax_f[2]-v_pmin_f[2]));
      float tmpRanges[3];
      VecSubtract3fv(tmpRanges, v_pmax_f, v_pmin_f);
      float axisRange = VecMax3fv(tmpRanges);
      int cbmPrecisionRange = (int)(axisRange/diagRange * (1<<(nbits-1)));
      v_pq->SetRange(cbmPrecisionRange);
        fprintf(stderr,"CBM precision simulation on ... using partial range of %d instead of %d\n",cbmPrecisionRange,1<<nbits);
/* simulating CBM precision */
    }
    else
    {
        fprintf(stderr,"CBM precision simulation off ... using full range of %d\n",1<<nbits);
    }
    v_pq->SetupQuantizer();

#ifdef PRINT_CONTROL_OUTPUT
    printf("aiQuantRange %d %d %d across\n", v_pq->m_aiQuantRange[0], v_pq->m_aiQuantRange[1], v_pq->m_aiQuantRange[2]);
#endif

    global_ic[0] = new RangeIntegerCompressor();
    global_ic[0]->SetRange(v_pq->m_aiQuantRange[0]);
    global_ic[0]->SetupDecompressor(rd_geom, 3, high_bits);

    global_ic[1] = new RangeIntegerCompressor();
    global_ic[1]->SetRange(v_pq->m_aiQuantRange[1]);
    global_ic[1]->SetupDecompressor(rd_geom, 3, high_bits);

    global_ic[2] = new RangeIntegerCompressor();
    global_ic[2]->SetRange(v_pq->m_aiQuantRange[2]);
    global_ic[2]->SetupDecompressor(rd_geom, 3, high_bits);
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    if (nbits)
    {
      fprintf(stderr,"no bounding box. dequantizing v_prop_f adaptive to %d bits\n",nbits);
    }
    else
    {
      fprintf(stderr,"no quantization. lossless compression of floating-point numbers.\n");
    }
#endif
    v_pq = 0;

    v_fc[0] = new RangeFloatCompressor();
    v_fc[0]->SetPrecision(nbits);
    v_fc[0]->SetupDecompressor(rd_geom, 3);

    v_fc[1] = new RangeFloatCompressor();
    v_fc[1]->SetPrecision(nbits);
    v_fc[1]->SetupDecompressor(rd_geom, 3);

    v_fc[2] = new RangeFloatCompressor();
    v_fc[2]->SetPrecision(nbits);
    v_fc[2]->SetupDecompressor(rd_geom, 3);
  }

#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)

  fprintf(stderr,"prediction correctors expressed in local coordinates\n");

  if (v_pmin_f == 0 || v_pmax_f == 0)
  {
    fprintf(stderr,"ERROR: needs advance knowledge of bounding box for decompressing with local coordinates\n");
    exit(0);
  }

  if (cbm_precision)
  {
    maxAbso = (1 << (nbits-1)) - 1;
    maxDist = rd_geom->readFloat();

    local_ic[0] = new RangeIntegerCompressor();
#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
    fprintf(stderr,"performing mid-point + height prediction\n");
    local_ic[0]->SetPrecision(nbits);
#elif defined(SVC_DO_FLIP_PREDICTION)
    fprintf(stderr,"performing flip across through mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits);
#else
    fprintf(stderr,"performing pure mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits-1);
#endif
    local_ic[0]->SetupDecompressor(rd_geom, 3, high_bits);

    local_ic[1] = new RangeIntegerCompressor();
    local_ic[1]->SetPrecision(nbits);
    local_ic[1]->SetupDecompressor(rd_geom, 3, high_bits);

    local_ic[2] = new RangeIntegerCompressor();
    local_ic[2]->SetPrecision(nbits);
    local_ic[2]->SetupDecompressor(rd_geom, 3, high_bits);
  }
  else
  {
    maxAbso = (1 << nbits) - 1;
    maxDist = rd_geom->readFloat();

    local_ic[0] = new RangeIntegerCompressor();
#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
    fprintf(stderr,"performing mid-point + height prediction\n");
    local_ic[0]->SetPrecision(nbits+1);
#elif defined(SVC_DO_FLIP_PREDICTION)
    fprintf(stderr,"performing flip across through mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits+1);
#elif defined(SVC_ADD_NORMAL_COMPONENT)
    fprintf(stderr,"performing add %g normal to mid-point prediction\n",SVC_NORMAL_SCALE);
    local_ic[0]->SetPrecision(nbits+1);
#else
    fprintf(stderr,"performing pure mid-point prediction\n");
    local_ic[0]->SetPrecision(nbits);
#endif
    local_ic[0]->SetupDecompressor(rd_geom, 3, high_bits);

    local_ic[1] = new RangeIntegerCompressor();
    local_ic[1]->SetPrecision(nbits+1);
    local_ic[1]->SetupDecompressor(rd_geom, 3, high_bits);

    local_ic[2] = new RangeIntegerCompressor();
    local_ic[2]->SetPrecision(nbits+1);
    local_ic[2]->SetupDecompressor(rd_geom, 3, high_bits);
  }

#else

  fprintf(stderr,"prediction correctors expressed in global coordinates\n");

#if defined(SVC_DO_TETRAHEDRAL_ACROSS_PREDICTION)
    fprintf(stderr,"performing mid-point + height prediction\n");
#elif defined(SVC_DO_FLIP_PREDICTION)
    fprintf(stderr,"performing flip across through mid-point prediction\n");
#elif defined(SVC_ADD_NORMAL_COMPONENT)
    fprintf(stderr,"performing add %g normal to mid-point prediction\n",SVC_NORMAL_SCALE);
#else
    fprintf(stderr,"performing pure mid-point prediction\n");
#endif

#endif
}

SVevent SVreader_svc::read_element()
{
  if (have_tetrahedron == 0)
  {
    assert(have_new == 0);
    next_new = 0;
    have_finalized = 0;
    if (decompress_tetrahedron() == 0)
    {
      if (nverts != -1 && v_count != nverts)
      {
        fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
      }
      nverts = v_count;
      if (ncells != -1 && c_count != ncells)
      {
        fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
      }
      ncells = c_count;
      have_tetrahedron = -1;
      return SV_EOF;
    }
  }
  if (have_new)
  {
    v_idx = c_idx[new_vertices[next_new]];
#if defined(SVC_USE_LOCAL_COORDINATE_SYSTEM)
      VecCopy3fv(v_prop_f, vertices[new_vertices[next_new]]->prop_f);
#else
    if (v_pq)
    {
      v_pq->DeQuantize((I32*)(vertices[new_vertices[next_new]]->prop_f), v_prop_f);
    }
    else
    {
      VecCopy3fv(v_prop_f, vertices[new_vertices[next_new]]->prop_f);
    }
#endif
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else if (have_tetrahedron > 0)
  {
    have_tetrahedron = 0;
    c_count++;
    return SV_TETRAHEDRON;
  }
  else if (have_tetrahedron < 0)
  {
    return SV_EOF;
  }
  return SV_ERROR;
}

SVevent SVreader_svc::read_event()
{
  if (have_tetrahedron == 0 && have_finalized == 0)
  {
    assert(have_new == 0);
    next_new = 0;
    next_finalized = 0;
    if (decompress_tetrahedron() == 0)
    {
      if (nverts != -1 && v_count != nverts)
      {
        fprintf(stderr,"WARNING: wrong vertex count: v_count (%d) != nverts (%d)\n", v_count, nverts);
      }
      nverts = v_count;
      if (ncells != -1 && c_count != ncells)
      {
        fprintf(stderr,"WARNING: wrong cell count: c_count (%d) != ncells (%d)\n", c_count, ncells);
      }
      ncells = c_count;
      have_tetrahedron = -1;
      return SV_EOF;
    }
  }
  if (have_new)
  {
//    v_idx = c_idx[new_vertices[next_new]];
    if (v_pq)
    {
      v_pq->DeQuantize((I32*)(vertices[new_vertices[next_new]]->prop_f), v_prop_f);
    }
    else
    {
      VecCopy3fv(v_prop_f, vertices[new_vertices[next_new]]->prop_f);
    }
    have_new--; next_new++;
    v_count++;
    return SV_VERTEX;
  }
  else if (have_tetrahedron > 0)
  {
    have_tetrahedron = 0;
    c_count++;
    return SV_TETRAHEDRON;
  }
  else if (have_finalized)
  {
    final_idx = c_idx[finalized_vertices[next_finalized]];
    have_finalized--; next_finalized++;
    return SV_FINALIZED;
  }
  else if (have_tetrahedron < 0)
  {
    return SV_EOF;
  }
  return SV_ERROR;
}

SVreader_svc::SVreader_svc()
{
  // init of SVreader_svc
  have_new = 0; next_new = 0; new_vertices = 0; 
  have_tetrahedron = 0;
  have_finalized = 0; next_finalized = 0; finalized_vertices = 0;

  nbits = -1;
}

SVreader_svc::~SVreader_svc()
{
  // clean-up for SVreader_svc
}
