/*
===============================================================================

  FILE:  svwriter_sva.cpp
  
  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    martin isenburg@cs.unc.edu

  COPYRIGHT:

    copyright (C) 2005  martin isenburg@cs.unc.edu

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_sva.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "vecnfv.h"
#include "vecniv.h"

bool SVwriter_sva::open(FILE* file)
{
  if (file == 0)
  {
    fprintf(stderr, "ERROR: zero file pointer not supported by SVwriter_sva\n");
    exit(0);
  }

#ifdef _WIN32
  if (file == stdout)
  {
    if(_setmode( _fileno( stdout ), _O_TEXT ) == -1 )
    {
      fprintf(stderr, "ERROR: cannot set stdout to text (translated) mode\n");
      exit(0);
    }
  }
#endif

  this->file = file;
  v_count = 0;
  c_count = 0;
  return true;
}

void SVwriter_sva::write_vertex(const float* v_prop_f)
{
  if (v_count + c_count == 0) write_header();

  if (v_pnum_f == 3)
  {
    fprintf(file, "v %.8g %.8g %.8g\012",v_prop_f[0],v_prop_f[1],v_prop_f[2]);
  }
  else if (v_pnum_f == 4)
  {
    fprintf(file, "v %.8g %.8g %.8g %.8g\012",v_prop_f[0],v_prop_f[1],v_prop_f[2], v_prop_f[3]);
  }
  else if (v_pnum_f == 5)
  {
    fprintf(file, "v %.8g %.8g %.8g %.8g %.8g\012",v_prop_f[0],v_prop_f[1],v_prop_f[2], v_prop_f[3], v_prop_f[4]);
  }
  v_count++;
}

void SVwriter_sva::write_tetrahedron(const I32* c_idx, const bool* c_final)
{
  if (v_count + c_count == 0) write_header();

  if (c_final)
  {
    fprintf(file, "t %d %d %d %d\012",(c_final[0] ? c_idx[0]-v_count : c_idx[0]+1),(c_final[1] ? c_idx[1]-v_count : c_idx[1]+1), (c_final[2] ? c_idx[2]-v_count : c_idx[2]+1), (c_final[3] ? c_idx[3]-v_count : c_idx[3]+1));
  }
  else
  {
    fprintf(file, "t %d %d %d %d ",c_idx[0]+1,c_idx[1]+1,c_idx[2]+1,c_idx[3]+1);
  }
  c_count++;
}

void SVwriter_sva::write_tetrahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  if (v_count + c_count == 0) write_header();

  if (c_final)
  {
    fprintf(file, "t %d %d %d %d",(c_final[0] ? c_idx[0]-v_count : c_idx[0]+1),(c_final[1] ? c_idx[1]-v_count : c_idx[1]+1), (c_final[2] ? c_idx[2]-v_count : c_idx[2]+1), (c_final[3] ? c_idx[3]-v_count : c_idx[3]+1));
  }
  else
  {
    fprintf(file, "t %d %d %d %d",c_idx[0]+1,c_idx[1]+1,c_idx[2]+1,c_idx[3]+1);
  }
  if (c_pnum_f)
  {
    for (int i = 0; i < c_pnum_f; i++)
    {
      fprintf(file, " %.8g",c_prop_f[i]);
    }
  }
  fprintf(file, "\012");
  c_count++;
}

void SVwriter_sva::write_hexahedron(const I32* c_idx, const bool* c_final)
{
  if (v_count + c_count == 0) write_header();

  if (c_final)
  {
    fprintf(file, "h %d %d %d %d %d %d %d %d\012",(c_final[0] ? c_idx[0]-v_count : c_idx[0]+1),(c_final[1] ? c_idx[1]-v_count : c_idx[1]+1), (c_final[2] ? c_idx[2]-v_count : c_idx[2]+1), (c_final[3] ? c_idx[3]-v_count : c_idx[3]+1),       (c_final[4] ? c_idx[4]-v_count : c_idx[4]+1),(c_final[5] ? c_idx[5]-v_count : c_idx[5]+1), (c_final[6] ? c_idx[6]-v_count : c_idx[6]+1), (c_final[7] ? c_idx[7]-v_count : c_idx[7]+1));
  }
  else
  {
    fprintf(file, "h %d %d %d %d %d %d %d %d",c_idx[0]+1,c_idx[1]+1,c_idx[2]+1,c_idx[3]+1,c_idx[4]+1,c_idx[5]+1,c_idx[6]+1,c_idx[7]+1);
  }
  c_count++;
}

void SVwriter_sva::write_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  if (v_count + c_count == 0) write_header();

  if (c_final)
  {
    fprintf(file, "h %d %d %d %d %d %d %d %d",(c_final[0] ? c_idx[0]-v_count : c_idx[0]+1),(c_final[1] ? c_idx[1]-v_count : c_idx[1]+1), (c_final[2] ? c_idx[2]-v_count : c_idx[2]+1), (c_final[3] ? c_idx[3]-v_count : c_idx[3]+1),       (c_final[4] ? c_idx[4]-v_count : c_idx[4]+1),(c_final[5] ? c_idx[5]-v_count : c_idx[5]+1), (c_final[6] ? c_idx[6]-v_count : c_idx[6]+1), (c_final[7] ? c_idx[7]-v_count : c_idx[7]+1));
  }
  else
  {
    fprintf(file, "h %d %d %d %d %d %d %d %d",c_idx[0]+1,c_idx[1]+1,c_idx[2]+1,c_idx[3]+1,c_idx[4]+1,c_idx[5]+1,c_idx[6]+1,c_idx[7]+1);
  }
  if (c_pnum_f)
  {
    for (int i = 0; i < c_pnum_f; i++)
    {
      fprintf(file, " %.8g",c_prop_f[i]);
    }
  }
  fprintf(file, "\012");
  c_count++;
}

void SVwriter_sva::write_finalized(I32 final_idx)
{
  if (final_idx < 0)
  {
    fprintf(file, "x %d\012",final_idx);
  }
  else
  {
    fprintf(file, "x %d\012",final_idx+1);
  }
}

void SVwriter_sva::close(bool close_file, bool update)
{
  // close of SVwriter_sva interface
  if (nverts != v_count || ncells != c_count)
  {
    if (update && file != stdout) update_header();
  }
  if (close_file && file && file != stdout) fclose(file);
  file = 0;

  // close of SVwriter interface
  if (nverts != -1) if (nverts != v_count)  fprintf(stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count);
  if (ncells != -1) if (ncells != c_count)  fprintf(stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count);
  nverts = v_count;
  ncells = c_count;
  v_count = -1;
  c_count = -1;
}

void SVwriter_sva::write_header()
{
  int i;
  if (nverts != -1 || file != stdout) fprintf(file, "# nverts %-12d\012",nverts);
  if (ncells != -1 || file != stdout) fprintf(file, "# ncells %-12d\012",ncells);
  if (v_pnum_f) fprintf(file, "# v_pnum_f %d\012",v_pnum_f);
  if (v_pmin_f) {fprintf(file, "# v_pmin_f"); for(i = 0; i < v_pnum_f; i++) fprintf(file, " %.8g", v_pmin_f[i]); fprintf(file, "\012");}
  if (v_pmax_f) {fprintf(file, "# v_pmax_f"); for(i = 0; i < v_pnum_f; i++) fprintf(file, " %.8g", v_pmax_f[i]); fprintf(file, "\012");}
  if (c_pnum_f) fprintf(file, "# c_pnum_f %d\012",c_pnum_f);
  if (c_pmin_f) {fprintf(file, "# c_pmin_f"); for(i = 0; i < c_pnum_f; i++) fprintf(file, " %.8g", c_pmin_f[i]); fprintf(file, "\012");}
  if (c_pmax_f) {fprintf(file, "# c_pmax_f"); for(i = 0; i < c_pnum_f; i++) fprintf(file, " %.8g", c_pmax_f[i]); fprintf(file, "\012");}
  if (comments)
  {
    for (i = 0; i < ncomments; i++)
    {
      fprintf(file, "# %s\012",comments[i]);
    }
  }
}

void SVwriter_sva::update_header()
{
  if (file == stdout)
  {
    fprintf(stderr,"WARNING: cannot update header when writing to stdout\n");
    return;
  }
  if (fseek(file, 0L, SEEK_SET) != 0)
  {
    fprintf(stderr,"WARNING: cannot update header because fseek failed\n");
    return;
  }
  fprintf(file, "# nverts %-12d\012",v_count);
  fprintf(file, "# ncells %-12d\012",c_count);
  if (fseek(file, 0L, SEEK_END) != 0)
  {
    fprintf(stderr,"WARNING: header was updated but final fseek failed\n");
  }
}

SVwriter_sva::SVwriter_sva()
{
  // init of SVwriter_sva
  file = 0;
}

SVwriter_sva::~SVwriter_sva()
{
  // clean-up for SVwriter_sva
}
