/*
===============================================================================

  FILE:  SVwriter_svh.cpp

  CONTENTS:

    see corresponding header file

  PROGRAMMERS:

    clement.courbet@ecp.fr

  COPYRIGHT:

    copyright (C) 2009  clement courbet

    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

  CHANGE HISTORY:

    see corresponding header file

===============================================================================
*/
#include "svwriter_svh.h"

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <assert.h>

//this is for compatibility with the new g++
#ifdef __GNUC__
  #include <ext/hash_map>
  namespace std { using namespace __gnu_cxx; }
  using namespace std;
#else
  #include <hash_map.h>
#endif

#define PRINT_CONTROL_OUTPUT
#undef PRINT_CONTROL_OUTPUT

#define PRINT_DEBUG_OUTPUT
#undef PRINT_DEBUG_OUTPUT

#define PRINT_GEOM_DEBUG_OUTPUT
#undef PRINT_GEOM_DEBUG_OUTPUT

#define PRINT_CELLDATA_DEBUG_OUTPUT
#undef PRINT_CELLDATA_DEBUG_OUTPUT

#define CALCULATE_QUANTIZATION_ERROR
#undef CALCULATE_QUANTIZATION_ERROR

#define CALCULATE_PREDICTION_ERROR
#undef CALCULATE_PREDICTION_ERROR

#define OUTPUT_MACHINE_READABLE_STATS
#undef OUTPUT_MACHINE_READABLE_STATS

//code that is common to svreader_svh.cpp and svwriter_svh.cpp
#include "svcommon_svh.cpp"

#include "hexahedronrotations.h"

typedef hash_map<I32, SVvertex*> my_vertex_hash;
static my_vertex_hash* vertex_hash;

// rangeencoder 

#include "rangeencoder.h"

static RangeEncoder* re_conn;
static RangeEncoder* re_conn_op;
static RangeEncoder* re_conn_cache;
static RangeEncoder* re_conn_index;
static RangeEncoder* re_conn_candi;
static RangeEncoder* re_conn_final;
static RangeEncoder* re_conn_join;

static RangeEncoder* re_geom;

#ifdef SVH_COMPRESS_CELL_DATA
static RangeEncoder* re_celldata;
#endif

// statistics

#ifdef PRINT_CONTROL_OUTPUT

static int op_start = 0;
static int op_hut = 0;
static int op_step = 0;
static int op_roof = 0;
static int op_corner = 0;
static int op_bridge = 0;
static int op_tunnel = 0;
static int op_gap = 0;
static int op_pit = 0;
static int op_den = 0;

static int hut_join = 0;
static int step_join = 0;
static int corner_join = 0;

static int used_index = 0;
static int used_cache = 0;

static int start_hit_vertex = 0;
static int start_hit_edge_possible = 0;
static int start_miss = 0;

static int hut_hit_face = 0;
static int hut_hit_edge = 0;
static int hut_hit_vertex = 0;
static int hut_miss = 0;

static int roof_hit_face = 0;
static int roof_hit_edge = 0;
static int roof_hit_vertex = 0;
static int roof_miss = 0;

static int step_hit_face = 0;
static int step_hit_edge = 0;
static int step_hit_vertex = 0;
static int step_miss = 0;

static int corner_hit_face = 0;
static int corner_hit_edge = 0;
static int corner_hit_vertex = 0;
static int corner_miss = 0;

static int bridge_hit_face = 0;
static int bridge_hit_edge = 0;
static int bridge_hit_vertex = 0;
static int bridge_miss = 0;

static int tunnel_hit_face = 0;
static int tunnel_hit_edge = 0;
static int tunnel_hit_vertex = 0;
static int tunnel_miss = 0;

static int gap_hit_face = 0;
static int gap_hit_edge = 0;
static int gap_hit_vertex = 0;
static int gap_miss = 0;

static int pit_hit_face = 0;
static int pit_hit_edge = 0;
static int pit_hit_vertex = 0;
static int pit_miss = 0;

static int den_hit_face = 0;
static int den_hit_edge = 0;
static int den_hit_vertex = 0;
static int den_miss = 0;

static int confim_yes = 0;
static int confim_no = 0;

static int predict_lorenzo = 0;
static int predict_parallelogram = 0;
static int predict_shoot = 0;
static int predict_last = 0;
#endif

#if !defined(SVH_COMPRESS_GEOMETRY)
FILE *geofile;
#endif

static void initEncoder(FILE* file)
{
  if(file)
  {
    re_conn = new RangeEncoder(file);
    re_conn_op = re_conn;
    re_conn_cache = re_conn;
    re_conn_index = re_conn;
    re_conn_candi = re_conn;
    re_conn_final = re_conn;
    re_conn_join = re_conn;
    re_geom = re_conn;
#ifdef SVH_COMPRESS_CELL_DATA
    re_celldata = re_conn;
#endif
  }
  else
  {
    re_conn = new RangeEncoder(0,false);
    re_conn_op = new RangeEncoder(0,false);
    re_conn_cache = new RangeEncoder(0,false);
    re_conn_index = new RangeEncoder(0,false);
    re_conn_candi = new RangeEncoder(0,false);
    re_conn_final = new RangeEncoder(0,false);
    re_conn_join = new RangeEncoder(0,false);
    re_geom = new RangeEncoder(0,false);
#ifdef SVH_COMPRESS_CELL_DATA
    re_celldata = new RangeEncoder(0,false);
#endif
  }
}

static void finishEncoder(I32 nverts, I32 ncells)
{
  if (re_conn != re_conn_op)
  {
    re_conn_op->done();
    re_conn_cache->done();
    re_conn_index->done();
    re_conn_candi->done();
    re_conn_final->done();
    re_conn_join->done();
    re_conn->done();
    re_geom->done();
    
    fprintf(stderr,"\n");
    fprintf(stderr,"*** compression rate details ***\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"op:   \t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_conn_op->getNumberBytes(),  8.0f/ncells*re_conn_op->getNumberBytes());
    fprintf(stderr,"cache:\t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_conn_cache->getNumberBytes(), 8.0f/ncells*re_conn_cache->getNumberBytes());
    fprintf(stderr,"index:\t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_conn_index->getNumberBytes(), 8.0f/ncells*re_conn_index->getNumberBytes());
    fprintf(stderr,"candi:\t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_conn_candi->getNumberBytes(), 8.0f/ncells*re_conn_candi->getNumberBytes());
    fprintf(stderr,"final:\t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_conn_final->getNumberBytes(), 8.0f/ncells*re_conn_final->getNumberBytes());
    fprintf(stderr,"join: \t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_conn_join->getNumberBytes(), 8.0f/ncells*re_conn_join->getNumberBytes());
    fprintf(stderr,"rest: \t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_conn->getNumberBytes(), 8.0f/ncells*re_conn->getNumberBytes());
    fprintf(stderr,"--\ntotal: \t%6.3f bpv \t%6.3f bph\n\n", 8.0f/nverts*(re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn_join->getNumberBytes()+re_conn->getNumberBytes()), 8.0f/ncells*(re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn_join->getNumberBytes()+re_conn->getNumberBytes()));

#ifdef PRINT_CONTROL_OUTPUT
    fprintf (stderr, "Statistics:       number     face hit    edges hit   vertex hit         miss         join\n");
    fprintf (stderr, "SVH_START:  % 12i            X            X % 12i % 12i\n", op_start, start_hit_vertex, start_miss);
    fprintf (stderr, "SVH_HUT:    % 12i % 12i % 12i % 12i % 12i % 12i\n", op_hut, hut_hit_face, hut_hit_edge, hut_hit_vertex, hut_miss, hut_join);
    fprintf (stderr, "SVH_ROOF:   % 12i % 12i % 12i % 12i % 12i\n", op_roof, roof_hit_face, roof_hit_edge, roof_hit_vertex, roof_miss);
    fprintf (stderr, "SVH_STEP:   % 12i % 12i % 12i % 12i % 12i % 12i\n", op_step, step_hit_face, step_hit_edge, step_hit_vertex, step_miss, step_join);
    fprintf (stderr, "SVH_BRIDGE: % 12i % 12i % 12i % 12i % 12i\n", op_bridge, bridge_hit_face, bridge_hit_edge, bridge_hit_vertex, bridge_miss);
    fprintf (stderr, "SVH_CORNER: % 12i % 12i % 12i % 12i % 12i % 12i\n", op_corner, corner_hit_face, corner_hit_edge, corner_hit_vertex, corner_miss, corner_join);
    fprintf (stderr, "SVH_TUNNEL: % 12i % 12i % 12i % 12i % 12i\n", op_tunnel, tunnel_hit_face, tunnel_hit_edge, tunnel_hit_vertex, tunnel_miss);
    fprintf (stderr, "SVH_GAP:    % 12i % 12i % 12i % 12i % 12i\n", op_gap, gap_hit_face, gap_hit_edge, gap_hit_vertex, gap_miss);
    fprintf (stderr, "SVH_PIT:    % 12i % 12i % 12i % 12i % 12i\n", op_pit, pit_hit_face, pit_hit_edge, pit_hit_vertex, pit_miss);
    fprintf (stderr, "SVH_DEN:    % 12i % 12i % 12i % 12i % 12i\n", op_den, den_hit_face, den_hit_edge, den_hit_vertex, den_miss);
    fprintf (stderr, "confirmed/unconfirmed face merges: %i/%i\n", confim_yes, confim_no);
    
    fprintf(stderr,"predict_last %d, predict_shoot %d, predict_parallelogram %d, predict_lorenzo %d\n", predict_last, predict_shoot, predict_parallelogram, predict_lorenzo);
#endif
    fprintf(stderr,"geometry bitrate: \t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_geom->getNumberBytes(), 8.0f/ncells*re_geom->getNumberBytes());

#ifdef SVH_COMPRESS_CELL_DATA
    fprintf(stderr,"cell data bitrate: \t%6.3f bpv \t%6.3f bph\n", 8.0f/nverts*re_celldata->getNumberBytes(), 8.0f/ncells*re_celldata->getNumberBytes());
#endif

#ifdef OUTPUT_MACHINE_READABLE_STATS
    printf("MRS%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f",
        8.0f/ncells*re_conn_op->getNumberBytes(),
        8.0f/ncells*re_conn_cache->getNumberBytes(),
        8.0f/ncells*re_conn_index->getNumberBytes(),
        8.0f/ncells*re_conn_candi->getNumberBytes(),
        8.0f/ncells*re_conn_final->getNumberBytes(),
        8.0f/ncells*re_conn_join->getNumberBytes(),
        8.0f/ncells*re_conn->getNumberBytes(),
        8.0f/ncells*(re_conn_op->getNumberBytes()+re_conn_cache->getNumberBytes()+re_conn_index->getNumberBytes()+re_conn_candi->getNumberBytes()+re_conn_final->getNumberBytes()+re_conn_join->getNumberBytes()+re_conn->getNumberBytes()));
    printf("\t%f", 8.0f/nverts*re_geom->getNumberBytes());
    printf("\t%i\t0\t0\t0\t%i\t%i", 
        op_start, start_hit_vertex, start_miss);
    printf("\t%i\t%i\t%i\t%i\t%i\t%i", 
        op_hut, hut_hit_face, hut_hit_edge, hut_hit_vertex, hut_miss, hut_join);
    printf("\t%i\t%i\t%i\t%i\t%i\t0",
        op_roof, roof_hit_face, roof_hit_edge, roof_hit_vertex, roof_miss);
    printf("\t%i\t%i\t%i\t%i\t%i\t%i",
        op_step, step_hit_face, step_hit_edge, step_hit_vertex, step_miss, step_join);
    printf("\t%i\t%i\t%i\t%i\t%i\t0",
        op_bridge, bridge_hit_face, bridge_hit_edge, bridge_hit_vertex, bridge_miss);
    printf("\t%i\t%i\t%i\t%i\t%i\t%i",
        op_corner, corner_hit_face, corner_hit_edge, corner_hit_vertex, corner_miss, corner_join);
    printf("\t%i\t%i\t%i\t%i\t%i\t0",
        op_tunnel, tunnel_hit_face, tunnel_hit_edge, tunnel_hit_vertex, tunnel_miss);
    printf("\t%i\t%i\t%i\t%i\t%i\t0",
        op_gap, gap_hit_face, gap_hit_edge, gap_hit_vertex, gap_miss);
    printf("\t%i\t%i\t%i\t%i\t%i\t0",
        op_pit, pit_hit_face, pit_hit_edge, pit_hit_vertex, pit_miss);
    printf("\t%i\t%i\t%i\t%i\t%i\t0",
        op_den, den_hit_face, den_hit_edge, den_hit_vertex, den_miss);
    printf("\n");
#endif

    delete re_conn_op;
    delete re_conn_cache;
    delete re_conn_index;
    delete re_conn_candi;
    delete re_conn_final;
    delete re_conn_join;
    delete re_conn;
    delete re_geom;
#ifdef SVH_COMPRESS_CELL_DATA
    delete re_celldata;
#endif
  }
  else
  {
    re_conn->done();
  
#ifdef PRINT_CONTROL_OUTPUT
    fprintf (stderr, "Statistics:       number     face hit    edges hit   vertex hit         miss         join\n");
    fprintf (stderr, "SVH_START:  % 12i            X            X % 12i % 12i\n", op_start, start_hit_vertex, start_miss);
    fprintf (stderr, "SVH_HUT:    % 12i % 12i % 12i % 12i % 12i % 12i\n", op_hut, hut_hit_face, hut_hit_edge, hut_hit_vertex, hut_miss, hut_join);
    fprintf (stderr, "SVH_ROOF:   % 12i % 12i % 12i % 12i % 12i\n", op_roof, roof_hit_face, roof_hit_edge, roof_hit_vertex, roof_miss);
    fprintf (stderr, "SVH_STEP:   % 12i % 12i % 12i % 12i % 12i % 12i\n", op_step, step_hit_face, step_hit_edge, step_hit_vertex, step_miss, step_join);
    fprintf (stderr, "SVH_BRIDGE: % 12i % 12i % 12i % 12i % 12i\n", op_bridge, bridge_hit_face, bridge_hit_edge, bridge_hit_vertex, bridge_miss);
    fprintf (stderr, "SVH_CORNER: % 12i % 12i % 12i % 12i % 12i % 12i\n", op_corner, corner_hit_face, corner_hit_edge, corner_hit_vertex, corner_miss, corner_join);
    fprintf (stderr, "SVH_TUNNEL: % 12i % 12i % 12i % 12i % 12i\n", op_tunnel, tunnel_hit_face, tunnel_hit_edge, tunnel_hit_vertex, tunnel_miss);
    fprintf (stderr, "SVH_GAP:    % 12i % 12i % 12i % 12i % 12i\n", op_gap, gap_hit_face, gap_hit_edge, gap_hit_vertex, gap_miss);
    fprintf (stderr, "SVH_PIT:    % 12i % 12i % 12i % 12i % 12i\n", op_pit, pit_hit_face, pit_hit_edge, pit_hit_vertex, pit_miss);
    fprintf (stderr, "SVH_DEN:    % 12i % 12i % 12i % 12i % 12i\n", op_den, den_hit_face, den_hit_edge, den_hit_vertex, den_miss);
    fprintf (stderr, "confirmed/unconfirmed face merges: %i/%i\n", confim_yes, confim_no);
    fprintf(stderr,"total: bytes %d (%f bpv / %f bph)\n", re_conn->getNumberBytes(),(float) re_conn->getNumberBytes()*8 /nverts,(float) re_conn->getNumberBytes()*8 /ncells);
    
    fprintf(stderr,"predict_last %d, predict_shoot %d, predict_parallelogram %d, predict_lorenzo %d\n", predict_last, predict_shoot, predict_parallelogram, predict_lorenzo);
#endif
  
    delete re_conn;
  }
}

bool SVwriter_svh::open(FILE* file, int nbits, int high_bits, int nCellDataBits, int nCellDataHighBits)
{
  // store precision
  this->file = file;
  this->nbits = nbits;
  this->high_bits = high_bits;
  this->cellDataBits = nCellDataBits;
  this->cellDataHighBits = nCellDataHighBits;

  if(nbits < 0 || nbits > 24)
  {
    fprintf(stderr, "ERROR: nbits is %d. it should be in the range [0,24]\n", nbits);
    exit(0);
  }

  if(high_bits < 0 || high_bits > 16)
  {
    fprintf(stderr, "ERROR: high_bits is %d. it should be in the range [0,16]\n", high_bits);
    exit(0);
  }
  
  if(nCellDataBits < 0 || nCellDataBits > 24)
  {
    fprintf(stderr, "ERROR: nCellDataBits is %d. it should be in the range [0,24]\n", nCellDataBits);
    exit(0);
  }

  if(cellDataHighBits < 0 || cellDataHighBits > 16)
  {
    fprintf(stderr, "ERROR: cellDataHighBits is %d. it should be in the range [0,16]\n", cellDataHighBits);
    exit(0);
  }

#ifdef _WIN32
  if(file == stdout)
  {
    if(_setmode(_fileno(stdout), _O_BINARY) == -1)
    {
      fprintf(stderr, "ERROR: cannot set stdout to binary (untranslated) mode\n");
      exit(0);
    }
  }
#endif

  // write version
  if(file)
  {
    unsigned char compiled_version = 0;
#ifdef SVH_BETTER_SHOOT_PREDICTION
    compiled_version += 1;
#endif
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
    compiled_version += 2;
#endif
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
    compiled_version += 4;
#endif
#ifdef SVH_INITIALIZE_VISVERTICES
    compiled_version += 8;
#endif
#ifdef SVH_DUMB_JOIN_CODING
    compiled_version += 16;
#endif
#ifdef SVH_IMPLICIT_FINALIZATION
    compiled_version += 32;
#endif
#ifdef SVH_COMPRESS_CELL_DATA
    compiled_version += 64;
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
    compiled_version += 128;
#endif
    fputc(compiled_version, file);
  }

  vertex_hash = new my_vertex_hash;
  dv = new DynamicVector();
  lc = new LittleCache(8);
  lc_f = new LittleCache(6);

  initEncoder(file);
  initModels(1);

  last_op = 0;
  last_fin = 0;

  // write precision
  re_conn->writeRange(25, nbits);

  v_count = 0;
  c_count = 0;

#if !defined(SVH_COMPRESS_GEOMETRY)
  geofile = fopen("geofile","w");
#endif

  return true;
}

void SVwriter_svh::set_v_pnum_f(I32 v_pnum_f)
{
  if(v_pnum_f != 3)
  {
    fprintf(stderr,"WARNING: v_pnum_f = %d not implemented. changed to v_pnum_f = 3\n", v_pnum_f);
    v_pnum_f = 3;
  }
  this->v_pnum_f = v_pnum_f;
}

void SVwriter_svh::write_vertex(const float* v_prop_f)
{
  if(v_count + c_count == 0) write_header();

  SVvertex* vertex = allocVertex();
  vertex_hash->insert(my_vertex_hash::value_type(v_count, vertex));
  vertex->index = v_count;
  VecCopyNfv(vertex->v_prop_f, v_prop_f, v_pnum_f);
  v_count++;
}

#ifdef SVH_COMPRESS_CELL_DATA

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataStart(float* cd, float *residual)
#else
static inline void compressCellDataStart(float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data actual      : %i\n", ((I32*)cd)[i]);
#endif
      c_ic[i]->Compress(0, ((I32*)cd)[i], SVH_START);
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*)residual)[i] = 0;
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = 0;
#endif
      cd[i] = c_fc[i]->Compress(0.0f, cd[i], SVH_START);
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data actual      : %f\n", cd[i]);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataHut( const float *d0, float* cd, const float *r0, float *residual)
#else
static inline void compressCellDataHut( const float *d0, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", ((I32*)d0)[i], ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - ((I32*)d0)[i];
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", ((I32*)r0)[i]);
  #endif
      c_ic[i]->Compress(((I32*)r0)[i], ((I32*) residual)[i], SVH_HUT);
#else
      c_ic[i]->Compress(((I32*)d0)[i], ((I32*)cd)[i], SVH_HUT);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", d0[i], cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - d0[i];
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", r0[i]);
  #endif
      residual[i] = c_fc[i]->Compress(r0[i], residual[i], SVH_HUT);
#else
      cd[i] = c_fc[i]->Compress(d0[i], cd[i], SVH_HUT);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataRoof( const float *d0, const float *d5, float* cd, const float *r0, const float *r5,  float *residual)
#else
static inline void compressCellDataRoof( const float *d0, const float *d5, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d5)[i])/2;
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r5)[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_ROOF);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_ROOF);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d5[i])/2;
  
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r0[i] + r5[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_ROOF);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_ROOF);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataStep( const float *d0, const float *d1, float* cd, const float *r0, const float *r1, float *residual)
#else
static inline void compressCellDataStep( const float *d0, const float *d1, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i])/2;
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_STEP);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_STEP);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d1[i])/2;
  
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r0[i] + r1[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_STEP);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_STEP);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataCorner( const float *d0, const float *d1, const float *d2, float* cd, const float *r0, const float *r1, const float *r2, float *residual)
#else
static inline void compressCellDataCorner( const float *d0, const float *d1, const float *d2, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i] + ((I32*)d2)[i])/3;
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i] + ((I32*)r2)[i])/3;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_CORNER);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_CORNER);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d1[i] + d2[i])/3; 
  
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r0[i] + r1[i] + r2[i])/3; 
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_CORNER);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_CORNER);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataBridge( const float *d0, const float *d1, const float *d3, float* cd, const float *r0, const float *r1, const float *r3, float *residual)
#else
static inline void compressCellDataBridge( const float *d0, const float *d1, const float *d3, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d1)[i] + ((I32*)d3)[i])/2;
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r1)[i] + ((I32*)r3)[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_BRIDGE);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_BRIDGE);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d1[i] + d3[i])/2; 
  
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r1[i] + r3[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_BRIDGE);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_BRIDGE);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataTunnel( const float *d0, const float *d1, const float *d3, const float *d5, float* cd, const float *r0, const float *r1, const float *r3, const float *r5, float *residual)
#else
static inline void compressCellDataTunnel( const float *d0, const float *d1, const float *d3, const float *d5, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i] + ((I32*)d3)[i] + ((I32*)d5)[i])/4;
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i] + ((I32*)r3)[i] + ((I32*)r5)[i])/4;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_TUNNEL);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_TUNNEL);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d0[i] + d1[i] + d3[i] + d5[i])/4; 
  
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r0[i] + r1[i] + r3[i] + r5[i])/4; 
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_TUNNEL);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_TUNNEL);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataGap( const float *d0, const float *d1, const float *d2, const float *d3, float* cd, const float *r0, const float *r1, const float *r2, const float *r3, float *residual)
#else
static inline void compressCellDataGap( const float *d0, const float *d1, const float *d2, const float *d3, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d1)[i] + ((I32*)d3)[i])/2; 
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r1)[i] + ((I32*)r3)[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_GAP);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_GAP);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d1[i] + d3[i])/2; 
  
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r1[i] + r3[i])/2;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_GAP);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_GAP);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataPit( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, float* cd, const float *r0, const float *r1, const float *r2, const float *r3, const float *r4, float *residual)
#else
static inline void compressCellDataPit( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d1)[i] + ((I32*)d2)[i] + ((I32*)d3)[i] + ((I32*)d4)[i])/4; 
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r1)[i] + ((I32*)r2)[i] + ((I32*)r3)[i] + ((I32*)r4)[i])/4;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_PIT);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_PIT);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred;
      pred = (d1[i] + d2[i] + d3[i] + d4[i])/4; 
  
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r1[i] + r2[i] + r3[i] + r4[i])/4; 
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_PIT);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_PIT);
#endif
    }
  }
}

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
static inline void compressCellDataDen( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, const float *d5, float* cd, const float *r0, const float *r1, const float *r2, const float *r3, const float *r4, const float *r5, float *residual)
#else
static inline void compressCellDataDen( const float *d0, const float *d1, const float *d2, const float *d3, const float *d4, const float *d5, float* cd)
#endif
{
  if(c_quantizer)
  {
    c_quantizer->EnQuantize(cd, (I32*)cd);
    for (int i = 0; i < cell_pnum_f; i++)
    {
      I32 pred = (((I32*)d0)[i] + ((I32*)d1)[i] + ((I32*)d2)[i] + ((I32*)d3)[i] + ((I32*)d4)[i]+ ((I32*)d5)[i])/6;
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %i / %i\n", pred, ((I32*)cd)[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      ((I32*) residual)[i] = ((I32*)cd)[i] - pred;
      I32 residualPred = (((I32*)r0)[i] + ((I32*)r1)[i] + ((I32*)r2)[i] + ((I32*)r3)[i] + ((I32*)r4)[i] + ((I32*)r5)[i])/6;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %i\n", residualPred);
  #endif
      c_ic[i]->Compress(residualPred, ((I32*) residual)[i], SVH_DEN);
#else
      c_ic[i]->Compress(pred, ((I32*)cd)[i], SVH_DEN);
#endif
    }
  }
  else
  {
    for(int i=0; i<cell_pnum_f; ++i)
    {
      float pred = (d0[i] + d1[i] + d2[i] + d3[i] + d4[i]+ d5[i])/6;
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted/actual      : %f / %f\n", pred, cd[i]);
#endif
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      residual[i] = cd[i] - pred;
      float residualPred = (r0[i] + r1[i] + r2[i] + r3[i] + r4[i]+ r5[i])/6;
  #ifdef PRINT_CELLDATA_DEBUG_OUTPUT
      fprintf(stderr, "cell data predicted residual: %f\n", residualPred);
  #endif
      residual[i] = c_fc[i]->Compress(residualPred, residual[i], SVH_DEN);
#else
      cd[i] = c_fc[i]->Compress(pred, cd[i], SVH_DEN);
#endif
    }
  }
}

#endif //SVH_COMPRESS_CELL_DATA

#ifdef SVH_COMPRESS_GEOMETRY
/**
 * Compress the vertex properties of v6 using the lorenzo predictor
 * @param p0,p1,p2,p3,p4,p5,p7: vertex properties known to the decoder
 * @param p6: vertex properties unknown to the decoder
 */
static void compressVertexGeometryLorenzo(const float* p0, const float* p1, const float* p2, const float* p3, const float* p4, const float* p5, float* p6, const float* p7)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_lorenzo++;
#endif
  //TODO: extend to more than coordinates (i.e. properties)
  if(v_pq)
  {
    // compute quantized coordinates
    v_pq->EnQuantize(p6, (I32*)p6);
    
    //\bar{p6} = p0 - p1 + p2 - p3 - p4 + p5 + p7
    static I32 pred[3];
    pred[0] = ((I32*)p0)[0] - ((I32*)p1)[0] + ((I32*)p2)[0] - ((I32*)p3)[0] - ((I32*)p4)[0] + ((I32*)p5)[0] + ((I32*)p7)[0];
    pred[1] = ((I32*)p0)[1] - ((I32*)p1)[1] + ((I32*)p2)[1] - ((I32*)p3)[1] - ((I32*)p4)[1] + ((I32*)p5)[1] + ((I32*)p7)[1];
    pred[2] = ((I32*)p0)[2] - ((I32*)p1)[2] + ((I32*)p2)[2] - ((I32*)p3)[2] - ((I32*)p4)[2] + ((I32*)p5)[2] + ((I32*)p7)[2];
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(L): %i\t%i\t%i\nactual: %i\t%i\t%i\n", pred[0], pred[1], pred[2], ((I32*)p6)[0], ((I32*)p6)[1], ((I32*)p6)[2]);
#endif
    for (int i = 0; i < 3; i++)
    {
      v_ic[i]->Compress(pred[i],((I32*)p6)[i], SVH_PREDICT_LORENZO);
    }
  }
  else
  {
    static float pred[3];
    pred[0] = p0[0] - p1[0] + p2[0] - p3[0] - p4[0] + p5[0] + p7[0];
    pred[1] = p0[1] - p1[1] + p2[1] - p3[1] - p4[1] + p5[1] + p7[1];
    pred[2] = p0[2] - p1[2] + p2[2] - p3[2] - p4[2] + p5[2] + p7[2];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(L): %f\t%f\t%f\nactual: %f\t%f\t%f\n", pred[0], pred[1], pred[2], p6[0], p6[1], p6[2]);
#endif

    for (int i = 0; i < 3; i++)
    {
      p6[i] = v_fc[i]->Compress(pred[i], p6[i], SVH_PREDICT_LORENZO);
    }  
  }
}

/**
 * Compress the vertex properties of vb using the 'last decoded vertex' predictor
 * @param pa: vertex to predict
 */
static void compressVertexGeometryLast(float* pa)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_last++;
#endif
  if(v_pq)
  {
    // compute quantized coordinates
    v_pq->EnQuantize(pa, (I32*)pa);
    
    for (int i = 0; i < 3; i++)
    {
      v_ic[i]->Compress(0, ((I32*)pa)[i], SVH_PREDICT_NONE);
    }
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "actual(N): %i\t%i\t%i\n", ((I32*)pa)[0], ((I32*)pa)[1], ((I32*)pa)[2]);
#endif
  }
  else
  {
    for (int i = 0; i < 3; i++)
    {
      pa[i] = v_fc[i]->Compress(0.0f, pa[i], SVH_PREDICT_NONE);
    }  
  }
}

#ifdef SVH_BETTER_SHOOT_PREDICTION
/**
 * Compress the vertex properties of vb using the better 'shoot' predictor
 * @param p4: vertex v4 (to predict)
 * @param p0: vertex v0
 * @param p1: vertex v1
 * @param p3: vertex v3
 */
static void compressVertexGeometryShoot(float* p4, const float* p0, const float *p1, const float *p3)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_shoot++;
#endif
  if(v_pq)
  {
    // compute quantized coordinates
    v_pq->EnQuantize(p4, (I32*)p4);

    I32* ip0 = (I32*)p0; I32* ip1 = (I32*)p1; I32* ip3 = (I32*)p3; I32* ip4 = (I32*)p4;

    static I32 pred[3]; 
    I32 meanNorm = (I32)(0.5* (sqrt( (ip1[0]-ip0[0])*(ip1[0]-ip0[0]) + (ip1[1]-ip0[1])*(ip1[1]-ip0[1]) + (ip1[2]-ip0[2])*(ip1[2]-ip0[2])) + sqrt( (ip3[0]-ip0[0])*(ip3[0]-ip0[0]) + (ip3[1]-ip0[1])*(ip3[1]-ip0[1]) + (ip3[2]-ip0[2])*(ip3[2]-ip0[2]))));
    pred[0] = ip0[0] + ((ip1[1]-ip0[1])*(ip3[2]-ip0[2]) - (ip1[2]-ip0[2])*(ip3[1]-ip0[1]))/meanNorm;
    pred[1] = ip0[1] + ((ip1[2]-ip0[2])*(ip3[0]-ip0[0]) - (ip1[0]-ip0[0])*(ip3[2]-ip0[2]))/meanNorm;
    pred[2] = ip0[2] + ((ip1[0]-ip0[0])*(ip3[1]-ip0[1]) - (ip1[1]-ip0[1])*(ip3[0]-ip0[0]))/meanNorm;

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(S): \t%i\t%i\t%i\nold: \t\t%i\t%i\t%i\nactual: \t%i\t%i\t%i\n", pred[0], pred[1], pred[2], ip0[0], ip0[1], ip0[2], ip4[0], ip4[1], ip4[2]);
#endif
    for (int i = 0; i < 3; i++)
    {
      v_ic[i]->Compress(pred[i], ip4[i], SVH_PREDICT_SHOOT);
    }
  }
  else
  {
    static float pred[3]; 
    float meanNorm = 0.5* (sqrt( (p1[0]-p0[0])*(p1[0]-p0[0]) + (p1[1]-p0[1])*(p1[1]-p0[1]) + (p1[2]-p0[2])*(p1[2]-p0[2])) + sqrt( (p3[0]-p0[0])*(p3[0]-p0[0]) + (p3[1]-p0[1])*(p3[1]-p0[1]) + (p3[2]-p0[2])*(p3[2]-p0[2])));
    pred[0] = p0[0] + (p1[1]-p0[1])*(p3[2]-p0[2]) - (p1[2]-p0[2])*(p3[1]-p0[1])/meanNorm;
    pred[1] = p0[1] + (p1[2]-p0[2])*(p3[0]-p0[0]) - (p1[0]-p0[0])*(p3[2]-p0[2])/meanNorm;
    pred[2] = p0[2] + (p1[0]-p0[0])*(p3[1]-p0[1]) - (p1[1]-p0[1])*(p3[0]-p0[0])/meanNorm;

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(S): %f\t%f\t%f\nold: %f\t%f\t%f\nactual: %f\t%f\t%f\n", pred[0], pred[1], pred[2], p0[0], p0[1], p0[2], p4[0], p4[1], p4[2]);
#endif
    for (int i = 0; i < 3; i++)
    {
      p4[i] = v_fc[i]->Compress(pred[i], p4[i], SVH_PREDICT_SHOOT);
    }  
  }
}
#endif

/**
 * Compress the vertex properties of vb using the 'shoot' predictor
 * @param pb: vertex to predict
 * @param pa: vertex below it
 */
static void compressVertexGeometryShoot(float* pb, const float* pa)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_shoot++;
#endif
  if(v_pq)
  {
    // compute quantized coordinates
    v_pq->EnQuantize(pb, (I32*)pb);
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(S): %i\t%i\t%i\nactual: %i\t%i\t%i\n", ((I32*)pa)[0], ((I32*)pa)[1], ((I32*)pa)[2], ((I32*)pb)[0], ((I32*)pb)[1], ((I32*)pb)[2]);
#endif
    for (int i = 0; i < 3; i++)
    {
      v_ic[i]->Compress(((I32*)pa)[i],((I32*)pb)[i], SVH_PREDICT_SHOOT);
    }
  }
  else
  {
#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(S): %f\t%f\t%f\nactual: %f\t%f\t%f\n", pa[0], pa[1], pa[2], pb[0], pb[1], pb[2]);
#endif
    for (int i = 0; i < 3; i++)
    {
      pb[i] = v_fc[i]->Compress(pa[i],pb[i],SVH_PREDICT_SHOOT);
    }  
  }
}


/**
 * Compress the vertex properties of vd using the parallelogram predictor
 * @param pb: vertex before pa when enumerating the vertices of the face
 * @param pa: opposite vertex
 * @param pc: vertex after pa when enumerating the vertices of the face
 * @param pd: vertex properties unknown to the decoder
 */
static void compressVertexGeometryParallelogram(const float* pb, const float* pa, const float* pc, float* pd)
{
#ifdef PRINT_CONTROL_OUTPUT
  predict_parallelogram++;
#endif
  if(v_pq)
  {
    // compute quantized coordinates
    v_pq->EnQuantize(pd, (I32*)pd);
    
    //\bar{pd} = pb + pc - pa
    static I32 pred[3];
    pred[0] = ((I32*)pb)[0] - ((I32*)pa)[0] + ((I32*)pc)[0];
    pred[1] = ((I32*)pb)[1] - ((I32*)pa)[1] + ((I32*)pc)[1];
    pred[2] = ((I32*)pb)[2] - ((I32*)pa)[2] + ((I32*)pc)[2];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(P): %i\t%i\t%i\nactual: %i\t%i\t%i\n", pred[0], pred[1], pred[2], ((I32*)pd)[0], ((I32*)pd)[1], ((I32*)pd)[2]);
#endif

    for (int i = 0; i < 3; i++)
    {
      v_ic[i]->Compress(((I32*)pred)[i],((I32*)pd)[i], SVH_PREDICT_PARALLELOGRAM);
    }
  }
  else
  {
    static float pred[3];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "pb: %f\t%f\t%f\npa: %f\t%f\t%f\npc: %f\t%f\t%f\n", pb[0], pb[1], pb[2], pa[0], pa[1], pa[2], pc[0], pc[1], pc[2]);
#endif

    pred[0] = pb[0] - pa[0] + pc[0];
    pred[1] = pb[1] - pa[1] + pc[1];
    pred[2] = pb[2] - pa[2] + pc[2];

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "predicted(P): %f\t%f\t%f\nactual: %f\t%f\t%f\n", pred[0], pred[1], pred[2], pd[0], pd[1], pd[2]);
#endif

    for (int i = 0; i < 3; i++)
    {
      pd[i] = v_fc[i]->Compress(pred[i],pd[i],SVH_PREDICT_PARALLELOGRAM);
    }  
  }
}
#endif //SVH_COMPRESS_GEOMETRY

/**
 * helper function to determine if a face is visited.
 * @param a,b,c,d: the four vertices of the face in INWARDS-pointing screwing order (i.e. 1,0,3,2 for CUBIT ordering of face 0)
 * @return NULL if unvisited, the halfedge that represents the face else.
 */
inline static SVedge* isFaceVisited(const SVvertex *a, const SVvertex *b, const SVvertex *c, const SVvertex *d)
{
  if(a && b && c && d)
  {
    for(int i = 0; i < b->list_size; i++)
    {
      if(( b->list[i]->target == a) &&(b->list[i]->next->target == d) &&(b->list[i]->next->next->target == c))
        return b->list[i];
    }
  }
  return NULL;
}

static void addAndCompressV0(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
  compressVertexGeometryLast(vertices[rot[0]]->v_prop_f);
#else
  fwrite(vertices[rot[0]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[0]]);
}

static void addAndCompressV1(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
  compressVertexGeometryShoot(vertices[rot[1]]->v_prop_f, vertices[rot[0]]->v_prop_f);
#else
  fwrite(vertices[rot[1]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[1]]);
}

static void addAndCompressV2(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
  compressVertexGeometryShoot(vertices[rot[2]]->v_prop_f, vertices[rot[1]]->v_prop_f);
#else
  fwrite(vertices[rot[2]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[2]]);
}

static void addAndCompressV3(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
  compressVertexGeometryParallelogram(vertices[rot[2]]->v_prop_f, vertices[rot[1]]->v_prop_f, vertices[rot[0]]->v_prop_f, vertices[rot[3]]->v_prop_f);
#else
  fwrite(vertices[rot[3]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[3]]);
}

static void addAndCompressV4(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
#ifdef SVH_BETTER_SHOOT_PREDICTION
  compressVertexGeometryShoot(vertices[rot[4]]->v_prop_f, vertices[rot[0]]->v_prop_f, vertices[rot[1]]->v_prop_f, vertices[rot[3]]->v_prop_f);
#else
  compressVertexGeometryShoot(vertices[rot[4]]->v_prop_f, vertices[rot[0]]->v_prop_f);
#endif
#else
  fwrite(vertices[rot[4]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[4]]);
}

static void addAndCompressV7(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
  compressVertexGeometryParallelogram(vertices[rot[3]]->v_prop_f, vertices[rot[0]]->v_prop_f, vertices[rot[4]]->v_prop_f, vertices[rot[7]]->v_prop_f);
#else
  fwrite(vertices[rot[7]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[7]]);
}

static void addAndCompressV5(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
  compressVertexGeometryParallelogram(vertices[rot[4]]->v_prop_f, vertices[rot[0]]->v_prop_f, vertices[rot[1]]->v_prop_f, vertices[rot[5]]->v_prop_f);
#else
  fwrite(vertices[rot[5]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[5]]);
}

static void addAndCompressV6(void *p_vertices, const int *rot)
{
  SVvertex **vertices = (SVvertex **)p_vertices;
#ifdef SVH_COMPRESS_GEOMETRY
  compressVertexGeometryLorenzo(vertices[rot[0]]->v_prop_f, vertices[rot[1]]->v_prop_f, vertices[rot[2]]->v_prop_f, vertices[rot[3]]->v_prop_f, vertices[rot[4]]->v_prop_f, vertices[rot[5]]->v_prop_f, vertices[rot[6]]->v_prop_f, vertices[rot[7]]->v_prop_f);
#else
  fwrite(vertices[rot[6]]->v_prop_f, 1, sizeof(float)*vertex_pnum_f, geofile);
#endif
  // insert it into the indexable data structure:
  dv->addElement(vertices[rot[6]]);
}

/**
 * helper function to get the edge number from two vertex local indices
  - edge 0 is 0-1
  - edge 1 is 1-2
  - edge 2 is 2-3
  - edge 3 is 3-0
  - edge 4 is 4-5
  - edge 5 is 5-6
  - edge 6 is 6-7
  - edge 7 is 7-4
  - edge 8 is 0-4
  - edge 9 is 1-5
  - edge 10 is 2-6
  - edge 11 is 3-7
 * @param i0,i1: vertex indices
 * @return the edge number
 */
static int getEdgeNumberFromLocalIndices(int i1, int i0)
{
  if(i1==1 && i0==0)
    return 0;
  else if(i1==0 && i0==3)
    return 1;
  else if(i1==3 && i0==2)
    return 2;
  else if(i1==2 && i0==1)
    return 3;
  else if(i1==3 && i0==0)
    return 4;
  else if(i1==0 && i0==4)
    return 5;
  else if(i1==4 && i0==7)
    return 6;
  else if(i1==7 && i0==3)
    return 7;
  else if(i1==4 && i0==0)
    return 8;
  else if(i1==0 && i0==1)
    return 9;
  else if(i1==1 && i0==5)
    return 10;
  else if(i1==5 && i0==4)
    return 11;
  else if(i1==5 && i0==1)
    return 12;
  else if(i1==1 && i0==2)
    return 13;
  else if(i1==2 && i0==6)
    return 14;
  else if(i1==6 && i0==5)
    return 15;
  else if(i1==6 && i0==2)
    return 16;
  else if(i1==2 && i0==3)
    return 17;
  else if(i1==3 && i0==7)
    return 18;
  else if(i1==7 && i0==6)
    return 19;
  else if(i1==7 && i0==4)
    return 20;
  else if(i1==4 && i0==5)
    return 21;
  else if(i1==5 && i0==6)
    return 22;
  else if(i1==6 && i0==7)
    return 23;
  else
  {
    assert(false);
    return -1;
  }
}

void SVwriter_svh::compress_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "\nBEGIN (con/de)compress_hexahedron()\n");
#endif
  int i;
  int op=-1;
  int lc_pos_f = -1;
  int lc_pos_e = -1;
  int lc_pos_v = -1;
  int rot[] = {0,1,2,3,4,5,6,7};


  my_vertex_hash::iterator hash_elements[8];
  SVvertex* vertices[8];

  //cell data
#ifdef SVH_COMPRESS_CELL_DATA
  if(cell_pnum_f)
  {
#ifdef PRINT_CELLDATA_DEBUG_OUTPUT
//     fprintf(stderr, "cell data for this hex: ");
//     for(int ncd = 0; ncd<cell_pnum_f; ++ncd)
//       fprintf(stderr, "%f ", c_prop_f[ncd]);
//     fprintf(stderr, "\n");
#endif
    memcpy(cellDataTmp, c_prop_f, cell_pnum_f*sizeof(float));
  }
#endif

  // get vertices from hash
  for(i = 0; i < 8; i++)
  {
    hash_elements[i] = vertex_hash->find(c_idx[i]);
    if(hash_elements[i] == vertex_hash->end())
    {
      fprintf(stderr,"FATAL ERROR: vertex %d with index %d not in hash\n", i, c_idx[i]);
      exit(0);
    }
    vertices[i] =(*hash_elements[i]).second;
    vertices[i]->finalize = c_final[i];
  }

  // renumber the vertices of the hexahedron so that it looks like this (CUBIT ordering):
  // see hexahedronrotations.h

#ifdef INPUT_USES_GAMBIT_ORDERING
  fprintf(stderr, "ERR... NOT IMPLEMENTED yet.\n");
  exit(1);
#else

#endif

#ifdef PRINT_DEBUG_OUTPUT
  //fprintf(stderr, "hexa is %d %d %d %d %d %d %d %d\n", c_idx[0], c_idx[1], c_idx[2], c_idx[3], c_idx[4], c_idx[5], c_idx[6], c_idx[7]);
  //fprintf(stderr, "that is %d %d %d %d %d %d %d %d\n", vertices[0]->index, vertices[1]->index, vertices[2]->index, vertices[3]->index, vertices[4]->index, vertices[5]->index, vertices[6]->index, vertices[7]->index);
#endif

  // check which vertices are already visited
  SVvertex* v_visited[8];
  int num_v_visited = 0;
  int last_v_visited;
  int num_v_cached = 0;

  for(i = 0; i < 8; i++)
  {
    if(vertices[i]->use_count)
    {
      num_v_visited++;
      last_v_visited = i;
      v_visited[i] = vertices[i];
      if(vertices[i]->in_cache != -1)
      {
        num_v_cached++;
      }
      //printf("%i-th vertex (%i) is visited\n", i, c_idx[i]);
    }
    else
    {
      v_visited[i] = 0;
    }
  }
  assert(num_v_cached < 8);

  // check which faces are already visited.
  // note: since the already visited faces point outwards from current advancing front, the face vertices should be given in reverse order (i.e. pointing inside hexahedron).
  SVedge* f_visited[] = {NULL,NULL,NULL,NULL,NULL,NULL};
  int num_f_visited = 0;
  if(num_v_visited)
  {
    // check face 0 (0,1,2,3) halfedge 1->0
    if(( f_visited[0] = isFaceVisited(v_visited[0], v_visited[1], v_visited[2], v_visited[3])))
      num_f_visited++;

    // check face 1 (0,3,7,4) halfedge 3->0
    if(( f_visited[1] = isFaceVisited(v_visited[0], v_visited[3], v_visited[7], v_visited[4])))
      num_f_visited++;

    // check face 2 (0,4,5,1) halfedge 4->0
    if(( f_visited[2] = isFaceVisited(v_visited[0], v_visited[4], v_visited[5], v_visited[1])))
      num_f_visited++;

    // check face 3 (1,5,6,2) halfedge 5->1
    if(( f_visited[3] = isFaceVisited(v_visited[1], v_visited[5], v_visited[6], v_visited[2])))
      num_f_visited++;

    // check face 4 (2,6,7,3) halfedge 6->2
    if(( f_visited[4] = isFaceVisited(v_visited[2], v_visited[6], v_visited[7], v_visited[3])))
      num_f_visited++;

    // check face 5 (4,7,6,5) halfedge 7->4
    if(( f_visited[5] = isFaceVisited(v_visited[4], v_visited[7], v_visited[6], v_visited[5])))
      num_f_visited++;
  }

#ifdef PRINT_DEBUG_OUTPUT
  //fprintf(stderr, "visited faces are %d %d %d %d %d %d\n", f_visited[0], f_visited[1], f_visited[2], f_visited[3], f_visited[4], f_visited[5]);
#endif

#ifdef PRINT_GEOM_DEBUG_OUTPUT
    fprintf(stderr, "p0: %f\t%f\t%f\n", vertices[rot[0]]->v_prop_f[0], vertices[rot[0]]->v_prop_f[1], vertices[rot[0]]->v_prop_f[2]);
    fprintf(stderr, "p1: %f\t%f\t%f\n", vertices[rot[1]]->v_prop_f[0], vertices[rot[1]]->v_prop_f[1], vertices[rot[1]]->v_prop_f[2]);
    fprintf(stderr, "p2: %f\t%f\t%f\n", vertices[rot[2]]->v_prop_f[0], vertices[rot[2]]->v_prop_f[1], vertices[rot[2]]->v_prop_f[2]);
    fprintf(stderr, "p3: %f\t%f\t%f\n", vertices[rot[3]]->v_prop_f[0], vertices[rot[3]]->v_prop_f[1], vertices[rot[3]]->v_prop_f[2]);
    fprintf(stderr, "p4: %f\t%f\t%f\n", vertices[rot[4]]->v_prop_f[0], vertices[rot[4]]->v_prop_f[1], vertices[rot[4]]->v_prop_f[2]);
    fprintf(stderr, "p5: %f\t%f\t%f\n", vertices[rot[5]]->v_prop_f[0], vertices[rot[5]]->v_prop_f[1], vertices[rot[5]]->v_prop_f[2]);
    fprintf(stderr, "p6: %f\t%f\t%f\n", vertices[rot[6]]->v_prop_f[0], vertices[rot[6]]->v_prop_f[1], vertices[rot[6]]->v_prop_f[2]);
    fprintf(stderr, "p7: %f\t%f\t%f\n", vertices[rot[7]]->v_prop_f[0], vertices[rot[7]]->v_prop_f[1], vertices[rot[7]]->v_prop_f[2]);
#endif


  if(num_f_visited == 0) // SVH_START
  {
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "SVH_START\n");
#endif
#ifdef PRINT_CONTROL_OUTPUT
    op_start++;
    if(num_v_cached >= 2) start_hit_edge_possible++;
    if(num_v_cached) start_hit_vertex++;
    else  start_miss++;
#endif

    if(edge_buffer_size)
    {
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
      re_conn_op->encode(rmOpX[last_fin][last_op], SVH_START);
#else
      re_conn_op->encode(rmOp[last_op], SVH_START);
#endif
    }
    else
      re_conn_op->encode(rmDone, 0); // more to come
    op = SVH_START;

    //tell the decoder which vertices are visited and which are not.
    unsigned int joinedVertices = 0; 
    for(i = 7; i >= 0; --i)
    {
      joinedVertices <<= 1;
      if( vertices[rot[i]]->use_count) ++joinedVertices;
    }
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "joinedVertices is %i\n", joinedVertices);
#endif
    //joinedVertices LSB tells if vertices[rot[0]] is visited
    //joinedVertices MSB tells if vertices[rot[7]] is visited
    re_conn_join->encode(rmStart_VisitedVertices, joinedVertices);

    // compress vertices
    if(vertices[rot[0]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[0]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[0],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[0],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[0]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV0(vertices, rot);
    }
    if(vertices[rot[1]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[1]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[1],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[1],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV1(vertices, rot);
    }
    if(vertices[rot[2]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[2]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[2],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[2],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[2]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV2(vertices, rot);
    }
    if(vertices[rot[3]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[3]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[3],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[3],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[3]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV3(vertices, rot);
    }
    if(vertices[rot[4]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[4]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[4],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[4],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[4]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV4(vertices, rot);
    }
    if(vertices[rot[7]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[7]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[7],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[7],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[7]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV7(vertices, rot);
    }
    if(vertices[rot[5]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[5]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[5],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[5],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[5]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV5(vertices, rot);
    }
    if(vertices[rot[6]]->use_count)
    {
      //then code it
      if(( lc_pos_v = vertices[rot[6]]->in_cache) != -1)
      {
        //cool, the vertex is in the cache:
        re_conn_join->encode(rmS_Cache[6],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
        used_cache++;
#endif
      }
      else
      {
        re_conn_join->encode(rmS_Cache[6],8);
        // encode its index among all active vertices
        int dv_index = dv->getRelativeIndex(vertices[rot[6]]);
        re_conn_join->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
        used_index++;
#endif
      }
    }
    else
    {
      addAndCompressV6(vertices, rot);
    }

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      compressCellDataStart(cellDataTmp, tmpCellDataResidual);
#else
      compressCellDataStart(cellDataTmp);
#endif
    }
#endif  
  }
  else if(num_f_visited == 1) // SVH_HUT
  {
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
    re_conn_op->encode(rmOpX[last_fin][last_op], SVH_HUT);
#else
    re_conn_op->encode(rmOp[last_op], SVH_HUT);
#endif
    op = SVH_HUT;
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "SVH_HUT\n");
#endif
#ifdef PRINT_CONTROL_OUTPUT
    op_hut++;
#endif
    if(f_visited[0])
      ;//nothing to do
    else if(f_visited[1])
      rotateHexXIQ(rot, f_visited);
    else if(f_visited[3])
      rotateHexXQ(rot, f_visited);
    else if(f_visited[2])
      rotateHexYQ(rot, f_visited);
    else if(f_visited[4])
      rotateHexYIQ(rot, f_visited);
    else if(f_visited[5])
      rotateHexYH(rot, f_visited);
    //and check
    assert(f_visited[0]);

    // check whether an edge of the known face is in the face cache
    // if yes, then rotate the face appropriately.
    if(num_v_cached>=4)
    {
      if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
      {
        // great, the first edge is in the face cache, nothing to do!!!
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
      {
        // great, the second edge is in the face cache, rotate 90 degrees!!!
        rotateHexZQ(rot, f_visited);
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
      {
        // great, the third edge is in the face cache, rotate 180 degrees!!!
        rotateHexZH(rot, f_visited);
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
      {
        // great, the fourth edge is in the face cache, rotate 270 degrees!!!
        rotateHexZIQ(rot, f_visited);
      }
    }

#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 6);
#endif
    if(lc_pos_f != -1)
    {
      // yes ... we can use a face from the cache
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
      re_conn_cache->encode(rmHut_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
      re_conn_cache->encode(rmHut_FaceCache[last_op],lc_pos_f);
#endif
#ifdef PRINT_CONTROL_OUTPUT
      hut_hit_face++;
#endif
      assert(lc_f->pos(f_visited[0]) == lc_pos_f);
    }
    else
    {
      // no ... we cannot use a face from the cache
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
      re_conn_cache->encode(rmHut_FaceCacheX[last_fin][last_op],6);
#else
      re_conn_cache->encode(rmHut_FaceCache[last_op],6);
#endif

      int lc_pos_e = -1;
      // is there an edge of f0 in the cache (i.e. two consecutive vertices of f0 in the cache)?
      //NOTE: we do not use the same scheme as for tet meshes here since not all combinations of two vertices is an edge. We explicitely give the number of the (non-oriented) edge instead:
      if(num_v_cached >= 2)
      {
#ifdef PRINT_CONTROL_OUTPUT
        hut_hit_edge++;
#endif
        if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
        {
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
        {
          //ok, second edge is in the cache.
          rotateHexZQ(rot, f_visited);
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
        {
          //ok, third edge is in the cache.
          rotateHexZH(rot, f_visited);
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
        {
          //ok, fourth edge is in the cache.
          rotateHexZIQ(rot, f_visited);
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 24);
#endif
      if(lc_pos_e != -1)
      {
        //yes, we can use an edge from the cache.
        re_conn_cache->encode(rmHut_EdgeCache[last_op],lc_pos_e);

        // encode which of the faces incident to v1 and v0 is f0
        int candidate=-1, candidate_count=0;
        for(i = 0; i < vertices[rot[1]]->list_size; i++)
        {
          if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
          {
            if(vertices[rot[1]]->list[i] == f_visited[0])
            {
              assert(candidate == -1);
              candidate = candidate_count;
            }
            candidate_count++;
          }
        }
        assert(candidate != -1);
        if(candidate_count > 1)
        {
          if(candidate_count == 2)
            re_conn_candi->encode(rm_FaceCandi,candidate);
          else
            re_conn_candi->writeRange(candidate_count,candidate);
        }
      }
      else
      {
        // no .... we cannot use an edge from the cache
        re_conn_cache->encode(rmHut_EdgeCache[last_op],24);

        // check whether at least one vertex of f0 is in the vertex
        // cache ... if yes, then rotate the face appropriately
        if(num_v_cached)
        {
          if(vertices[rot[1]]->in_cache != -1)
          {
            // great
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[0]]->in_cache != -1) // this is vertex f_visited[0]->target
          {
            rotateHexZQ(rot, f_visited);
            // great
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[3]]->in_cache != -1) // this is vertex f_visited[0]->next->target
          {
            rotateHexZH(rot, f_visited);
            // great
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[2]]->in_cache != -1) // this is vertex f_visited[0]->next->next->target
          {
            rotateHexZIQ(rot, f_visited);
            // great
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 8);
#endif
        if(lc_pos_v != -1)
        {
          // yes ... we can use a vertex from the cache
          re_conn_cache->encode(rmHut_VertexCache[last_op],lc_pos_v);
#ifdef PRINT_CONTROL_OUTPUT
          hut_hit_vertex++;
#endif
          assert(vertices[rot[1]]->in_cache == lc_pos_v);
        }
        else
        {
          // no ... we cannot use a vertex from the cache
          re_conn_cache->encode(rmHut_VertexCache[last_op],8);
          int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
          re_conn_index->writeRange(dv->size(),dv_index);
#ifdef PRINT_CONTROL_OUTPUT
          hut_miss++;
#endif
        }

        // encode which of the faces incident to v1 is f0
        // how many potential candidates are around that vertex?
        for(i = 0; i < vertices[rot[1]]->list_size; i++)
        {
          if(vertices[rot[1]]->list[i] == f_visited[0])
          {
            if(vertices[rot[1]]->list_size > 1)
            {
              if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
              {
                re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
              }
              else
              {
                re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
              }
            }
            break;
          }
        }
        assert(i < vertices[rot[1]]->list_size);
      }
    }

    // encode v4,v5,v6,v7
    //we first need to tell the decoder whether there is a vertex or edge join
    re_conn_join->encode(rmHut_HasJoin,(num_v_visited > 4) ? 1 : 0 );

    if(num_v_visited > 4) //if there are joins
    {
#ifdef SVH_DUMB_JOIN_CODING
        bool v4Joined = vertices[rot[4]]->use_count > 0;
        bool v5Joined = vertices[rot[5]]->use_count > 0;
        bool v6Joined = vertices[rot[6]]->use_count > 0;
        bool v7Joined = vertices[rot[7]]->use_count > 0;

        unsigned int joinedString = (v7Joined ? 8 : 0) + (v6Joined ? 4 : 0) + (v5Joined ? 2 : 0) + (v4Joined ? 1 : 0);
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "joinedString is %i\n", joinedString);
#endif
        re_conn_join->encode (rmHut_JoinedVertices, joinedString - 1);

        if(v4Joined)
        {
          int dv_index = dv->getRelativeIndex(vertices[rot[4]]);
          re_conn_join->writeRange(dv->size(), dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 4 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
#ifdef PRINT_CONTROL_OUTPUT
          hut_join++;
#endif
        }
        else
        {
          addAndCompressV4(vertices, rot);
        }
        if(v7Joined)
        {
          int dv_index = dv->getRelativeIndex(vertices[rot[7]]);
          re_conn_join->writeRange(dv->size(), dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 7 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
#ifdef PRINT_CONTROL_OUTPUT
          hut_join++;
#endif
        }
        else
        {
          addAndCompressV7(vertices, rot);
        }
        if(v5Joined)
        {
          int dv_index = dv->getRelativeIndex(vertices[rot[5]]);
          re_conn_join->writeRange(dv->size(), dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 5 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
#ifdef PRINT_CONTROL_OUTPUT
          hut_join++;
#endif
        }
        else
        {
          addAndCompressV5(vertices, rot);
        }
        if(v6Joined)
        {
          int dv_index = dv->getRelativeIndex(vertices[rot[6]]);
          re_conn_join->writeRange(dv->size(), dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 6 is joined to index %i (dv->size() %i)\n", dv_index, dv->size());
#endif
#ifdef PRINT_CONTROL_OUTPUT
          hut_join++;
#endif
        }
        else
        {
          addAndCompressV6(vertices, rot);
        }
#else
        fprintf(stderr, "NOT YET IMPLEMENTED: intelligent join for SVH_STEP\n");
        exit(1);
#endif
      
    }
    else
    {
      addAndCompressV4(vertices, rot);
      addAndCompressV7(vertices, rot);
      addAndCompressV5(vertices, rot);
      addAndCompressV6(vertices, rot);
    }
  
#ifdef SVH_COMPRESS_CELL_DATA
    //now compress the cell data.
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      compressCellDataHut(f_visited[0]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
      compressCellDataHut(f_visited[0]->c_prop_f, cellDataTmp);
#endif
    }
#endif  
  }
  else if(num_f_visited==2) //SVH_STEP and SVH_ROOF
  {
    //rotate the hexahedron such that the first visited face is f0:
    if(f_visited[0])
      ;//nothing to do
    else if(f_visited[1])
      rotateHexXIQ(rot, f_visited);
    else if(f_visited[3])
      rotateHexXQ(rot, f_visited);
    else if(f_visited[2])
      rotateHexYQ(rot, f_visited);
    else if(f_visited[4])
      rotateHexYIQ(rot, f_visited);
    else
    {
      fprintf(stderr, "ERROR: num_f_visited==2 faces 0 to 4 are unvisited. \n");
      exit(1);
    }
    //and check
    assert(f_visited[0] != 0);

    if(f_visited[5]) //SVH_ROOF
    {
      op = SVH_ROOF;
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "SVH_ROOF\n");
#endif
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
      re_conn_op->encode(rmOpX[last_fin][last_op], SVH_ROOF);
#else
      re_conn_op->encode(rmOp[last_op], SVH_ROOF);
#endif
#ifdef PRINT_CONTROL_OUTPUT
      op_roof++;
#endif
      // check whether one of the known faces is in the face cache
      // if it is face1, then rotate the hexa such that f1 becomes f0 and f0 becomes f1.
      // that way if there is a face in the cache, it must be f0.
      if(num_v_cached>=4)
      {
        if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
        {
          // great, the first edge of f0 is in the face cache, nothing to do!!!
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5]->next)) != -1)
        {
          // exchange faces
          rotateHexXH(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
        {
          // exchange faces.
          rotateHexZQ(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5])) != -1)
        {
          // exchange faces. TODO: do it in 1 step
          rotateHexZQ(rot, f_visited);
          rotateHexXH(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
        {
          // exchange faces.
          rotateHexZH(rot, f_visited);
          // great, the first edge of f0 is in the face cache,
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5]->next->next->next)) != -1)
        {
          // exchange faces TODO: do it in 1 step
          rotateHexZH(rot, f_visited);
          rotateHexXH(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
        {
          // exchange faces
          rotateHexZIQ(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5]->next->next)) != -1)
        {
          // exchange faces TODO: do it in 1 step
          rotateHexZIQ(rot, f_visited);
          rotateHexXH(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 6);
#endif
      if(lc_pos_f != -1)
      {
        // yes ... we can use a face from the cache
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmRoof_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
        re_conn_cache->encode(rmRoof_FaceCache[last_op],lc_pos_f);
#endif
#ifdef PRINT_CONTROL_OUTPUT
        roof_hit_face++;
#endif
        assert(lc_f->pos(f_visited[0]) == lc_pos_f);
      }
      else
      {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmRoof_FaceCacheX[last_fin][last_op],6);
#else
        re_conn_cache->encode(rmRoof_FaceCache[last_op],6);
#endif
        // check whether two vertices of one of the known
        // faces are in vertex cache. if yes, then rotate the hexa so that this face is f0
        if(num_v_cached>=2)
        {
          if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[5]]->in_cache != -1))
          {
            // exchange faces
            rotateHexXH(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            // exchange faces
            rotateHexZQ(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[7]]->in_cache != -1) &&(vertices[rot[4]]->in_cache != -1))
          {
            // exchange faces
            rotateHexZQ(rot, f_visited);
            rotateHexXH(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
          {
            // exchange faces
            rotateHexZH(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[6]]->in_cache != -1) &&(vertices[rot[7]]->in_cache != -1))
          {
            // exchange faces
            rotateHexZH(rot, f_visited);
            rotateHexXH(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
          {
            // exchange faces
            rotateHexZIQ(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[5]]->in_cache != -1) &&(vertices[rot[6]]->in_cache != -1))
          {
            // exchange faces
            rotateHexZIQ(rot, f_visited);
            rotateHexXH(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 24);
#endif
        if(lc_pos_e != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          roof_hit_edge++;
#endif
          re_conn_cache->encode(rmRoof_EdgeCache[last_op],lc_pos_e);
          //then we have to tell the decoder which of the faces around this edge f0 is.
          if(lc_pos_e < 24)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v1 and v0 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else
            assert(false);
        }
        else
        {
          re_conn_cache->encode(rmRoof_EdgeCache[last_op], 24);
          // check whether 1 vertex of one of the known
          // faces is in vertex cache. if yes, then rotate the hexa so that this vertex is v1
          if(num_v_cached)
          {
            if(vertices[rot[1]]->in_cache != -1)
            {
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[4]]->in_cache != -1)
            {
              // exchange faces
              rotateHexXH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[0]]->in_cache != -1)
            {
              // exchange faces
              rotateHexZQ(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[7]]->in_cache != -1)
            {
              // exchange faces
              rotateHexZQ(rot, f_visited);
              rotateHexXH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[3]]->in_cache != -1)
            {
              // exchange faces
              rotateHexZH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[6]]->in_cache != -1)
            {
              // exchange faces
              rotateHexZH(rot, f_visited);
              rotateHexXH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[5]]->in_cache != -1)
            {
              // exchange faces
              rotateHexZIQ(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[2]]->in_cache != -1)
            {
              // exchange faces
              rotateHexZIQ(rot, f_visited);
              rotateHexXH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
          }
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 8);
#endif
          if(lc_pos_v != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            roof_hit_vertex++;
#endif
            assert(vertices[rot[1]]->in_cache == lc_pos_v);

            re_conn_cache->encode(rmRoof_VertexCache[last_op],lc_pos_v);
          }
          else
          {
#ifdef PRINT_CONTROL_OUTPUT
            roof_miss++;
#endif
            // no ... we cannot use a vertex from the cache
            re_conn_cache->encode(rmRoof_VertexCache[last_op], 8);
            int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
            re_conn_index->writeRange(dv->size(),dv_index);
          }
          //then we have to tell the decoder which of the faces around this vertex f0 is.
          // encode which of the faces incident to v1 is f0
          // how many potential candidates are around that vertex?
          for(i = 0; i < vertices[rot[1]]->list_size; i++)
          {
            if(vertices[rot[1]]->list[i] == f_visited[0])
            {
              if(vertices[rot[1]]->list_size > 1)
              {
                if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                {
                  re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                }
                else
                {
                  re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                }
              }
              break;
            }
          }
          assert(i < vertices[rot[1]]->list_size);
        }
      }

      //dont bother with cache for face 5: there is no reason why it should be in cache. 
      //so just give vertex 7 and which neighbouring face f5 is.
      assert(f_visited[5] != 0);
      int dv_index = dv->getRelativeIndex(vertices[rot[7]]);
      re_conn_index->writeRange(dv->size(),dv_index);
      // how many potential candidates are around that vertex?
      for(i = 0; i < vertices[rot[7]]->list_size; i++)
      {
        if(vertices[rot[7]]->list[i] == f_visited[5])
        {
          if(vertices[rot[7]]->list_size > 1)
          {
            if(vertices[rot[7]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
            {
              re_conn_candi->encode(rm_VertexCandi[vertices[rot[7]]->list_size],i);
            }
            else
            {
              re_conn_candi->writeRange(vertices[rot[7]]->list_size,i);
            }
          }
          break;
        }
      }
      assert(i < vertices[rot[7]]->list_size);

#ifdef SVH_COMPRESS_CELL_DATA
      //now compress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        compressCellDataRoof(f_visited[0]->c_prop_f, f_visited[5]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[5]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        compressCellDataRoof(f_visited[0]->c_prop_f, f_visited[5]->c_prop_f, cellDataTmp);
#endif
      }
#endif  
    }
    else //SVH_STEP
    {
      op = SVH_STEP;
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "SVH_STEP\n");
#endif
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
      re_conn_op->encode(rmOpX[last_fin][last_op], SVH_STEP);
#else
      re_conn_op->encode(rmOp[last_op], SVH_STEP);
#endif
#ifdef PRINT_CONTROL_OUTPUT
      op_step++;
#endif
      //rotate the hexahedron to put it in the canonical STEP state
      if(f_visited[1])
        ;//nothing to do
      else if(f_visited[2])
        rotateHexZIQ(rot, f_visited);
      else if(f_visited[3])
        rotateHexZH(rot, f_visited);
      else if(f_visited[4])
        rotateHexZQ(rot, f_visited);
      else
      {
        fprintf(stderr, "ERROR: SVH_STEP operation, but faces 1 to 4 are unvisited. \n");
      }
      //and check
      assert(f_visited[1] != 0);

      // check whether one of the known faces is in the face cache
      // if it is face1, then rotate the hexa such that f1 becomes f0 and f0 becomes f1.
      // that way if there is a face in the cache, it must be f0.
      // if one of the faces is indeed in in the face cache, then tell which edge it is.
      //   lc_pos_f = 0 means that f_visited[0] is 0th in the face cache
      //   lc_pos_f = 1 means that f_visited[0] is 1st in the face cache
      //   lc_pos_f = 2 means that f_visited[0] is 2nd in the face cache
      //   lc_pos_f = 3 means that f_visited[0] is 3rd in the face cache
      //   lc_pos_f = 4 means that f_visited[0] is 4th in the face cache
      //   lc_pos_f = 5 means that f_visited[0] is 5th in the face cache
      //   lc_pos_f = 6 means that f_visited[0]->next is 0th in the face cache
      //   lc_pos_f = 7 means that f_visited[0]->next is 1st in the face cache
      //   ...
      //   lc_pos_f = 23 means that f_visited[0]->next->next->next is 5th in the face cache
      //   lc_pos_f = 24 means that neither f0 not f1 are in the face cache
      if(num_v_cached>=4)
      {
        if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
        {
          // great, the first edge of f0 is in the face cache, nothing to do!!!
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next->next)) != -1)
        {
          // exchange faces
          rotateHexExchangeF0F1(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
        {
          lc_pos_f += 6;
          // great, the second edge of f0 is in the face cache, nothing to do!!!
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1])) != -1)
        {
          lc_pos_f += 6;
          // exchange faces
          rotateHexExchangeF0F1(rot, f_visited);
          // great, the second edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
        {
          lc_pos_f += 12;
          // great, the third edge of f0 is in the face cache, nothing to do!!!
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next)) != -1)
        {
          lc_pos_f += 12;
          // exchange faces
          rotateHexExchangeF0F1(rot, f_visited);
          // great, the third edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
        {
          lc_pos_f += 18;
          // great, the fourth edge of f0 is in the face cache, nothing to do!!!
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next)) != -1)
        {
          lc_pos_f += 18;
          // exchange faces
          rotateHexExchangeF0F1(rot, f_visited);
          // great, the fourth edge of f0 is in the face cache.
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 24);
#endif
      if(lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        step_hit_face++;
#endif
        if(lc_pos_f < 6)
          assert(lc_f->pos(f_visited[0]) == lc_pos_f);
        else if(lc_pos_f < 12)
          assert(lc_f->pos(f_visited[0]->next) == lc_pos_f-6);
        else if(lc_pos_f < 18)
          assert(lc_f->pos(f_visited[0]->next->next) == lc_pos_f-12);
        else
          assert(lc_f->pos(f_visited[0]->next->next->next) == lc_pos_f-18);

#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
        re_conn_cache->encode(rmSCGPD_FaceCache[last_op],lc_pos_f);
#endif

      }
      else
      {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],24);
#else
        re_conn_cache->encode(rmSCGPD_FaceCache[last_op],24);
#endif
        // check whether two vertices of one of the known
        // faces are in vertex cache. if yes, then rotate the hexa so that this face is f0
        if(num_v_cached>=2)
        {
          if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[7]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            // exchange faces
            rotateHexExchangeF0F1(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            //ok, second edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) +24;
          }
          //only one case here for symmetry reasons
          else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
          {
            //ok, third edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[3]]->in_cache, vertices[rot[2]]->in_cache) +48;
          }
          else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[4]]->in_cache != -1))
          {
            // exchange faces
            rotateHexExchangeF0F1(rot, f_visited);
            //ok, third edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[3]]->in_cache, vertices[rot[2]]->in_cache) +48;
          }
          else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
          {
            //ok, fourth edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[2]]->in_cache, vertices[rot[1]]->in_cache) +72;
          }
          else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[7]]->in_cache != -1))
          {
            // exchange faces
            rotateHexExchangeF0F1(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[2]]->in_cache, vertices[rot[1]]->in_cache) +72;
          }
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 96);
#endif
        if(lc_pos_e != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          step_hit_edge++;
#endif
          re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],lc_pos_e);
          //then we have to tell the decoder which of the faces around this edge f0 is.
          if(lc_pos_e < 24)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v1 and v0 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 48)
          {
            assert(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1));
            // encode which of the faces incident to v0 and v3 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[0]]->list_size; i++)
            {
              if(vertices[rot[0]]->list[i]->target == vertices[rot[3]])
              {
                if(vertices[rot[0]]->list[i]->next->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 72)
          {
            assert(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1));
            // encode which of the faces incident to v3 and v2 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[3]]->list_size; i++)
            {
              if(vertices[rot[3]]->list[i]->target == vertices[rot[2]])
              {
                if(vertices[rot[3]]->list[i]->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 96)
          {
            assert(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1));
            // encode which of the faces incident to v2 and v1 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[2]]->list_size; i++)
            {
              if(vertices[rot[2]]->list[i]->target == vertices[rot[1]])
              {
                if(vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else
            assert(false);


        }
        else
        {
          re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],96);
          // check whether 1 vertex of one of the known
          // faces is in vertex cache. if yes, then rotate the hexa so that this face is f0
          //   lc_pos_v = 0 means that vertices[rot[1]] is 0th in the face cache
          //   lc_pos_v = 1 means that vertices[rot[1]] is 1st in the face cache
          //   ...
          //   lc_pos_v = 8 means that vertices[rot[0]] is 0th in the face cache
          //   ...
          //   lc_pos_v = 16 means that vertices[rot[2]] is 0th in the face cache
          //   ...
          //   lc_pos_v = 23 means that vertices[rot[2]] is 7th in the face cache
          //   lc_pos_v = 24 means that no vertices of f0 nor f1 are in the face cache
          if(num_v_cached)
          {
            if(vertices[rot[1]]->in_cache != -1)
            {
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[7]]->in_cache != -1)
            {
              // exchange faces
              rotateHexExchangeF0F1(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[0]]->in_cache != -1)
            {
              //ok, the 2nd vertex of f0 is in the cache
              lc_pos_v = 8+vertices[rot[0]]->in_cache;
            }
            else if(vertices[rot[3]]->in_cache != -1)
            {
              // exchange faces
              rotateHexExchangeF0F1(rot, f_visited);
              //ok, the 2nd vertex of f0 is in the cache
              lc_pos_v = 8+vertices[rot[0]]->in_cache;
            }
            //3 and 0 (3rd vertices): already done
            else if(vertices[rot[2]]->in_cache != -1)
            {
              //ok, the 4th vertex of f0 is in the cache
              lc_pos_v = 24+vertices[rot[2]]->in_cache;
            }
            else if(vertices[rot[4]]->in_cache != -1)
            {
              // exchange faces
              rotateHexExchangeF0F1(rot, f_visited);
              //ok, the 4th vertex of f0 is in the cache
              lc_pos_v = 24+vertices[rot[2]]->in_cache;
            }
          }

#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 32);
#endif
          if(lc_pos_v != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            step_hit_vertex++;
#endif
            re_conn_cache->encode(rmSCGPD_VertexCache[last_op],lc_pos_v);
            if(lc_pos_v < 8)
            {
              assert(vertices[rot[1]]->in_cache == lc_pos_v);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[1]]->list_size; i++)
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  if(vertices[rot[1]]->list_size > 1)
                  {
                    if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[1]]->list_size);
            }
            else if(lc_pos_v < 16)
            {
              assert(vertices[rot[0]]->in_cache == lc_pos_v-8);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[0]]->list_size; i++)
              {
                if(vertices[rot[0]]->list[i]->next->next->next == f_visited[0])
                {
                  if(vertices[rot[0]]->list_size > 1)
                  {
                    if(vertices[rot[0]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[0]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[0]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[0]]->list_size);
            }
            else if(lc_pos_v < 24)
            {
              assert(false);
            }
            else if(lc_pos_v < 32)
            {
              assert(vertices[rot[2]]->in_cache == lc_pos_v-24);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[2]]->list_size; i++)
              {
                if(vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  if(vertices[rot[2]]->list_size > 1)
                  {
                    if(vertices[rot[2]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[2]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[2]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[2]]->list_size);
            }
            else
              assert(false);
          }
          else
          {
#ifdef PRINT_CONTROL_OUTPUT
            step_miss++;
#endif
            // no ... we cannot use a vertex from the cache
            re_conn_cache->encode(rmSCGPD_VertexCache[last_op],32);
            int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
            re_conn_index->writeRange(dv->size(),dv_index);
            //then we have to tell the decoder which of the faces around this vertex f0 is.
            // encode which of the faces incident to <vertex> is f0
            // how many potential candidates are around that vertex?
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i] == f_visited[0])
              {
                if(vertices[rot[1]]->list_size > 1)
                {
                  if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                  {
                    re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                  }
                  else
                  {
                    re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                  }
                }
                break;
              }
            }
            assert(i < vertices[rot[1]]->list_size);
          }
        }
      }

      // encode which of the faces incident to v3 and v0 is f1
      int candidate=-1, candidate_count=0;
      for(i = 0; i < vertices[rot[3]]->list_size; i++)
      {
        if(vertices[rot[3]]->list[i]->target == vertices[rot[0]])
        {
          if(vertices[rot[3]]->list[i] == f_visited[1])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
      assert(candidate != -1);
      if(candidate_count > 1)
      {
        if(candidate_count == 2)
          re_conn_candi->encode(rm_FaceCandi,candidate);
        else
          re_conn_candi->writeRange(candidate_count,candidate);
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "actual operation is SVH_STEP\n");
#endif

      //if needed, tell the decoder that this is not a SVH_CORNER/SVH_GAP/SVH_PIT/SVH_DEN
      // (see backtobackhexas.svg)
      //look if pseudo-f2 (v4->v0->v1) exists
      for(i = 0; i < vertices[rot[4]]->list_size; i++)
      {
        if(( vertices[rot[4]]->list[i]->target == vertices[rot[0]]) &&(vertices[rot[4]]->list[i]->next->target == vertices[rot[1]]))
        {
          re_conn->encode(rm_ConfirmFaceMerge, 0);
#ifdef PRINT_CONTROL_OUTPUT
          confim_no++;
#endif
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "Back-to-back case (vertices[rot[1]]->list_size==3).\n");
#endif
          break;
        }
      }

      // encode v5,v6
      //we first need to tell the decoder whether there is a vertex or edge join
      re_conn_join->encode(rmStep_HasJoin,(num_v_visited > 6) ? 1 : 0 );

      if(num_v_visited > 6) //if there are joins
      {
#ifdef SVH_DUMB_JOIN_CODING
        bool v5Joined = vertices[rot[5]]->use_count > 0;
        bool v6Joined = vertices[rot[6]]->use_count > 0;

        re_conn_join->encode (rmStep_JoinedVertices, (v6Joined ? 2 : 0) + (v5Joined ? 1 : 0) - 1);

#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "joinedString is %i\n", (v6Joined ? 2 : 0) + (v5Joined ? 1 : 0));
#endif

        if(v5Joined)
        {
          int dv_index = dv->getRelativeIndex(vertices[rot[5]]);
          re_conn_join->writeRange(dv->size(), dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 5 is joined to index %i\n", dv_index);
#endif
#ifdef PRINT_CONTROL_OUTPUT
          step_join++;
#endif
        }
        else
        {
          addAndCompressV5(vertices, rot);
        }
        if(v6Joined)
        {
          int dv_index = dv->getRelativeIndex(vertices[rot[6]]);
          re_conn_join->writeRange(dv->size(), dv_index);
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr, "vertex 6 is joined to index %i\n", dv_index);
#endif
#ifdef PRINT_CONTROL_OUTPUT
          step_join++;
#endif
        }
        else
        {
          addAndCompressV6(vertices, rot);
        }
#else
        fprintf(stderr, "NOT YET IMPLEMENTED: intelligent join for SVH_STEP\n");
        exit(1);
#endif
      }
      else
      {
        addAndCompressV5(vertices, rot);
        addAndCompressV6(vertices, rot);
      }

#ifdef SVH_COMPRESS_CELL_DATA
      //now compress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        compressCellDataStep(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        compressCellDataStep(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, cellDataTmp);
#endif
      }
#endif  
    }
  }
  else if(num_f_visited==3) //SVH_CORNER and SVH_BRIDGE
  {
    //rotate the hexahedron such that the first visited face is f0
    if(f_visited[0])
      ;//nothing to do
    else if(f_visited[1])
      rotateHexXIQ(rot, f_visited);
    else if(f_visited[3])
      rotateHexXQ(rot, f_visited);
    else if(f_visited[2])
      rotateHexYQ(rot, f_visited);
    else
    {
      fprintf(stderr, "ERROR: num_f_visited==3 faces 0 to 3 are unvisited. \n");
      exit(1);
    }
    //and check
    assert(f_visited[0] != 0);
    //rotate the hexahedron such that the second visited face is f1
    if(f_visited[1])
      ;//nothing to do
    else if(f_visited[2])
      rotateHexZIQ(rot, f_visited);
    else if(f_visited[3])
      rotateHexZH(rot, f_visited);
    else if(f_visited[4])
      rotateHexZQ(rot, f_visited);
    else
    {
      fprintf(stderr, "ERROR: num_f_visited==3 but only two faces are visited.\n");
      exit(1);
    }
    //and check
    assert(f_visited[1] != 0);


    if(f_visited[5] || f_visited[3]) //SVH_BRIDGE
    {
      op = SVH_BRIDGE;
#ifdef PRINT_CONTROL_OUTPUT
      op_bridge++;
#endif
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
      re_conn_op->encode(rmOpX[last_fin][last_op], SVH_BRIDGE);
#else
      re_conn_op->encode(rmOp[last_op], SVH_BRIDGE);
#endif
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "SVH_BRIDGE\n");
#endif

      //put it into canonical orientation:
      if(f_visited[5])
        rotateHexXIQ(rot, f_visited);
      assert(f_visited[0] != 0);
      assert(f_visited[1] != 0);
      assert(f_visited[3] != 0);

      if(num_v_cached>=4)
      {
        if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
        {
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
        {
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
        {
          rotateHexZH(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
        {
          rotateHexZH(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 12);
#endif
      if(lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        bridge_hit_face++;
#endif
        if(lc_pos_f < 6)
          assert(lc_f->pos(f_visited[0]) == lc_pos_f);
        else if(lc_pos_f < 12)
          assert(lc_f->pos(f_visited[0]->next) == lc_pos_f-6);
        else
          assert(false);

#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmBT_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
        re_conn_cache->encode(rmBT_FaceCache[last_op],lc_pos_f);
#endif
      }
      else
      {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmBT_FaceCacheX[last_fin][last_op],12);
#else
        re_conn_cache->encode(rmBT_FaceCache[last_op],12);
#endif
        // check whether two vertices of one of the known
        // face is in the edge cache
        if(num_v_cached>=2)
        {
          if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            //ok, second edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
          else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
          {
            rotateHexZH(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
          {
            rotateHexZH(rot, f_visited);
            //ok, second edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
          
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 48);
#endif
        if(lc_pos_e != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          bridge_hit_edge++;
#endif
          re_conn_cache->encode(rmBT_EdgeCache[last_op],lc_pos_e);
          //then we have to tell the decoder which of the faces around this edge f0 is.
          if(lc_pos_e < 24)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v1 and v0 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 48)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1));
            // encode which of the faces incident to v0 and v3 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[0]]->list_size; i++)
            {
              if(vertices[rot[0]]->list[i]->target == vertices[rot[3]])
              {
                if(vertices[rot[0]]->list[i]->next->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else
            assert(false);
        }
        else
        {
          re_conn_cache->encode(rmBT_EdgeCache[last_op], 48);
          // check whether 1 vertex of one of f0 is in the vertex cache
          if(num_v_cached)
          {
            if(vertices[rot[1]]->in_cache != -1)
            {
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[0]]->in_cache != -1)
            {
              //ok, the 2nd vertex of f0 is in the cache
              lc_pos_v = vertices[rot[0]]->in_cache + 8;
            }
            else if(vertices[rot[3]]->in_cache != -1)
            {
              rotateHexZH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[2]]->in_cache != -1)
            {
              rotateHexZH(rot, f_visited);
              //ok, the 2nd vertex of f0 is in the cache
              lc_pos_v = vertices[rot[0]]->in_cache + 8;
            }
          }

#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 16);
#endif
          if(lc_pos_v != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            bridge_hit_vertex++;
#endif
            re_conn_cache->encode(rmBT_VertexCache[last_op],lc_pos_v);
            if(lc_pos_v < 8)
            {
              assert(vertices[rot[1]]->in_cache == lc_pos_v);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[1]]->list_size; i++)
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  if(vertices[rot[1]]->list_size > 1)
                  {
                    if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[1]]->list_size);
            }
            else if(lc_pos_v < 16)
            {
              assert(vertices[rot[0]]->in_cache == lc_pos_v-8);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[0]]->list_size; i++)
              {
                if(vertices[rot[0]]->list[i]->next->next->next == f_visited[0])
                {
                  if(vertices[rot[0]]->list_size > 1)
                  {
                    if(vertices[rot[0]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[0]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[0]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[0]]->list_size);
            }
            else
              assert(false);
          }
          else
          {
#ifdef PRINT_CONTROL_OUTPUT
            bridge_miss++;
#endif
            // no ... we cannot use a vertex from the cache
            re_conn_cache->encode(rmBT_VertexCache[last_op], 16);
            int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
            re_conn_index->writeRange(dv->size(),dv_index);
            //then we have to tell the decoder which of the faces around this vertex f0 is.
            // encode which of the faces incident to <vertex> is f0
            // how many potential candidates are around that vertex?
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i] == f_visited[0])
              {
                if(vertices[rot[1]]->list_size > 1)
                {
                  if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                  {
                    re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                  }
                  else
                  {
                    re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                  }
                }
                break;
              }
            }
            assert(i < vertices[rot[1]]->list_size);
          }
        }
      }

      // encode which of the faces incident to v3 and v0 is f1
      int candidate=-1, candidate_count=0;
      for(i = 0; i < vertices[rot[3]]->list_size; i++)
      {
        if(vertices[rot[3]]->list[i]->target == vertices[rot[0]])
        {
          if(vertices[rot[3]]->list[i] == f_visited[1])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
      assert(candidate != -1);
      if(candidate_count > 1)
      {
        if(candidate_count == 2)
          re_conn_candi->encode(rm_FaceCandi,candidate);
        else
          re_conn_candi->writeRange(candidate_count,candidate);
      }

      // encode which of the faces incident to v1 and v2 is f3
      candidate=-1; candidate_count=0;
      for(i = 0; i < vertices[rot[1]]->list_size; i++)
      {
        if(vertices[rot[1]]->list[i]->target == vertices[rot[2]])
        {
          if(vertices[rot[1]]->list[i]->next->next->next == f_visited[3])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
      assert(candidate != -1);
      if(candidate_count > 1)
      {
        if(candidate_count == 2)
          re_conn_candi->encode(rm_FaceCandi,candidate);
        else
          re_conn_candi->writeRange(candidate_count,candidate);
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "actual operation is SVH_BRIDGE\n");
#endif

#ifdef SVH_COMPRESS_CELL_DATA
      //now compress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        compressCellDataBridge(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        compressCellDataBridge(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, cellDataTmp);
#endif
      }
#endif  
    }
    else //SVH_CORNER
    {
      op = SVH_CORNER; 
#ifdef PRINT_CONTROL_OUTPUT
      op_corner++;
#endif
      //tell the decoder that this is a step since it can figure out that it is indeed an SVH_CORNER
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
      re_conn_op->encode(rmOpX[last_fin][last_op], SVH_STEP);
#else
      re_conn_op->encode(rmOp[last_op], SVH_STEP);
#endif
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "SVH_STEP\n");
#endif
      //put it into canonical orientation:
      if(f_visited[4])
        rotateHexZQ(rot, f_visited);

      if(num_v_cached>=4)
      {
        if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
        {
          // great, the first edge of f0 is in the face cache, nothing to do!!!
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1])) != -1)
        {
          // exchange faces
          rotateHex06IT(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2])) != -1)
        {
          // exchange faces
          rotateHex06T(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
        {
          // great, the second edge of f0 is in the face cache!!!
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next)) != -1)
        {
          // exchange faces
          rotateHex06IT(rot, f_visited);
          // great, the second edge of f0 is in the face cache.
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2]->next)) != -1)
        {
          // exchange faces
          rotateHex06T(rot, f_visited);
          // great, the second edge of f0 is in the face cache.
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
        {
          // great, the third edge of f0 is in the face cache!!!
          lc_pos_f += 12;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next)) != -1)
        {
          // exchange faces
          rotateHex06IT(rot, f_visited);
          // great, the third edge of f0 is in the face cache.
          lc_pos_f += 12;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2]->next->next)) != -1)
        {
          // exchange faces
          rotateHex06T(rot, f_visited);
          // great, the third edge of f0 is in the face cache.
          lc_pos_f += 12;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
        {
          // great, the fourth edge of f0 is in the face cache!!!
          lc_pos_f += 18;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next->next)) != -1)
        {
          // exchange faces
          rotateHex06IT(rot, f_visited);
          // great, the fourth edge of f0 is in the face cache.
          lc_pos_f += 18;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2]->next->next->next)) != -1)
        {
          // exchange faces
          rotateHex06T(rot, f_visited);
          // great, the fourth edge of f0 is in the face cache.
          lc_pos_f += 18;
        }
        assert(f_visited[0] != 0);
        assert(f_visited[1] != 0);
        assert(f_visited[2] != 0);
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 24);
#endif
      if(lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        corner_hit_face++;
#endif
        if(lc_pos_f < 6)
          assert(lc_f->pos(f_visited[0]) == lc_pos_f);
        else if(lc_pos_f < 12)
          assert(lc_f->pos(f_visited[0]->next) == lc_pos_f-6);
        else if(lc_pos_f < 18)
          assert(lc_f->pos(f_visited[0]->next->next) == lc_pos_f-12);
        else
          assert(lc_f->pos(f_visited[0]->next->next->next) == lc_pos_f-18);

#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
        re_conn_cache->encode(rmSCGPD_FaceCache[last_op],lc_pos_f);
#endif
      }
      else
      {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],24);
#else
        re_conn_cache->encode(rmSCGPD_FaceCache[last_op],24);
#endif
        // check whether two vertices of one of the known
        // faces are in vertex cache. if yes, then rotate the hexa so that this face is f0
        if(num_v_cached>=2)
        {
          if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            // exchange faces
            rotateHex06IT(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            // exchange faces
            rotateHex06T(rot, f_visited);
            //ok, first edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          //second edges: already done
          else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
          {
            //ok, third edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[3]]->in_cache, vertices[rot[2]]->in_cache) +48;
          }
          else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[7]]->in_cache != -1))
          {
            // exchange faces
            rotateHex06IT(rot, f_visited);
            //ok, third edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[3]]->in_cache, vertices[rot[2]]->in_cache) +48;
          }
          else if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[5]]->in_cache != -1))
          {
            // exchange faces
            rotateHex06T(rot, f_visited);
            //ok, third edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[3]]->in_cache, vertices[rot[2]]->in_cache) +48;
          }
          else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
          {
            //ok, fourth edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[2]]->in_cache, vertices[rot[1]]->in_cache) +72;
          }
          else if(( vertices[rot[7]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            // exchange faces
            rotateHex06IT(rot, f_visited);
            //ok, fourth edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[2]]->in_cache, vertices[rot[1]]->in_cache) +72;
          }
          else if(( vertices[rot[5]]->in_cache != -1) &&(vertices[rot[4]]->in_cache != -1))
          {
            // exchange faces
            rotateHex06T(rot, f_visited);
            //ok, fourth edge of f0 is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[2]]->in_cache, vertices[rot[1]]->in_cache) +72;
          }
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 96);
#endif
        if(lc_pos_e != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          corner_hit_edge++;
#endif
          re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],lc_pos_e);
          //then we have to tell the decoder which of the faces around this edge f0 is.
          if(lc_pos_e < 24)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v1 and v0 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 48)
          {
            assert(false);
          }
          else if(lc_pos_e < 72)
          {
            assert(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1));
            // encode which of the faces incident to v3 and v2 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[3]]->list_size; i++)
            {
              if(vertices[rot[3]]->list[i]->target == vertices[rot[2]])
              {
                if(vertices[rot[3]]->list[i]->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 96)
          {
            assert(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v2 and v1 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[2]]->list_size; i++)
            {
              if(vertices[rot[2]]->list[i]->target == vertices[rot[1]])
              {
                if(vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else
            assert(false);
        }
        else
        {
          re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],96);
          // check whether 1 vertex of one of the known
          // faces is in vertex cache. if yes, then rotate the hexa so that this face is f0
          //   lc_pos_v = 0 means that vertices[rot[1]] is 0th in the face cache
          //   lc_pos_v = 1 means that vertices[rot[1]] is 1st in the face cache
          //   ...
          //   lc_pos_v = 8 means that vertices[rot[0]] is 0th in the face cache
          //   ...
          if(num_v_cached)
          {
            if(vertices[rot[1]]->in_cache != -1)
            {
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[3]]->in_cache != -1)
            {
              // exchange faces
              rotateHex06IT(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[4]]->in_cache != -1)
            {
              // exchange faces
              rotateHex06T(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[0]]->in_cache != -1)
            {
              //ok, the 2nd vertex of f0 is in the cache
              lc_pos_v = vertices[rot[0]]->in_cache+8;
            }
            else if(vertices[rot[2]]->in_cache != -1)
            {
              //ok, the 4th vertex of f0 is in the cache
              lc_pos_v = vertices[rot[2]]->in_cache+24;
            }
            else if(vertices[rot[7]]->in_cache != -1)
            {
              // exchange faces
              rotateHex06IT(rot, f_visited);
              //ok, the 4th vertex of f0 is in the cache
              lc_pos_v = vertices[rot[2]]->in_cache+24;
            }
            else if(vertices[rot[5]]->in_cache != -1)
            {
              // exchange faces
              rotateHex06T(rot, f_visited);
              //ok, the 4th vertex of f0 is in the cache
              lc_pos_v = vertices[rot[2]]->in_cache+24;
            }
          }

#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 32);
#endif
          if(lc_pos_v != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            corner_hit_vertex++;
#endif
            re_conn_cache->encode(rmSCGPD_VertexCache[last_op],lc_pos_v);
            if(lc_pos_v < 8)
            {
              assert(vertices[rot[1]]->in_cache == lc_pos_v);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[1]]->list_size; i++)
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  if(vertices[rot[1]]->list_size > 1)
                  {
                    if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[1]]->list_size);
            }
            else if(lc_pos_v < 16)
            {
              assert(vertices[rot[0]]->in_cache == lc_pos_v-8);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[0]]->list_size; i++)
              {
                if(vertices[rot[0]]->list[i]->next->next->next == f_visited[0])
                {
                  if(vertices[rot[0]]->list_size > 1)
                  {
                    if(vertices[rot[0]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[0]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[0]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[0]]->list_size);
            }
            else if(lc_pos_v < 24)
            {
              assert(false);
            }
            else if(lc_pos_v < 32)
            {
              assert(vertices[rot[2]]->in_cache == lc_pos_v-24);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[2]]->list_size; i++)
              {
                if(vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  if(vertices[rot[2]]->list_size > 1)
                  {
                    if(vertices[rot[2]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[2]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[2]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[2]]->list_size);
            }
            else
              assert(false);
          }
          else
          {
#ifdef PRINT_CONTROL_OUTPUT
            corner_miss++;
#endif
            // no ... we cannot use a vertex from the cache
            re_conn_cache->encode(rmSCGPD_VertexCache[last_op],32);
            int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
            re_conn_index->writeRange(dv->size(),dv_index);
            //then we have to tell the decoder which of the faces around this vertex f0 is.
            // encode which of the faces incident to <vertex> is f0
            // how many potential candidates are around that vertex?
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i] == f_visited[0])
              {
                if(vertices[rot[1]]->list_size > 1)
                {
                  if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                  {
                    re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                  }
                  else
                  {
                    re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                  }
                }
                break;
              }
            }
            assert(i < vertices[rot[1]]->list_size);
          }
        }
      }

      // encode which of the faces incident to v3 and v0 is f1
      int candidate=-1, candidate_count=0;
      for(i = 0; i < vertices[rot[3]]->list_size; i++)
      {
        if(vertices[rot[3]]->list[i]->target == vertices[rot[0]])
        {
          if(vertices[rot[3]]->list[i] == f_visited[1])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
      assert(candidate != -1);
      if(candidate_count > 1)
      {
        if(candidate_count == 2)
          re_conn_candi->encode(rm_FaceCandi,candidate);
        else
          re_conn_candi->writeRange(candidate_count,candidate);
      }

      // no need to encode which of the faces incident to v4->v0 and v0->v1 is f2
      // the decoder can figure it out itself
      
      //confirm the SVH_CORNER face merge
      re_conn->encode(rm_ConfirmFaceMerge, 1);
#ifdef PRINT_CONTROL_OUTPUT
      confim_yes++;
#endif

      //if needed, tell the decoder that this is not a SVH_GAP/SVH_PIT/SVH_DEN
      //look if pseudo-f3 (v5->v1->v2) exists
      for(i = 0; i < vertices[rot[5]]->list_size; i++)
      {
        if(( vertices[rot[5]]->list[i]->target == vertices[rot[1]]) &&(vertices[rot[5]]->list[i]->next->target == vertices[rot[2]]))
        {
          re_conn->encode(rm_ConfirmFaceMerge, 0);
#ifdef PRINT_CONTROL_OUTPUT
          confim_no++;
#endif
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "Back-to-back case (vertices[rot[1]]->list_size==3).\n");
#endif
          break;
        }
      }
      
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "actual operation is SVH_CORNER\n");
#endif

      // encode v6
      //we first need to tell the decoder whether there is a vertex or edge join
      re_conn_join->encode(rmCorner_HasJoin,(num_v_visited ==8) ? 1 : 0);

      if(num_v_visited == 8) //if there are joins
      {
#ifdef PRINT_CONTROL_OUTPUT
        corner_join++;
#endif

#ifdef SVH_DUMB_JOIN_CODING
        int dv_index = dv->getRelativeIndex(vertices[rot[6]]);
  #ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr, "vertex 6 is joined to index %i\n", dv_index);
  #endif
        re_conn_join->writeRange(dv->size(), dv_index);
#else
        //can v6 be reached via an existing halfedge out of v2, v5 or v7 ?
        //try v2 first
        int candidate=-1, candidate_count=0;
        for(i = 0; i < vertices[rot[2]]->list_size; i++)
        {
          if((vertices[rot[2]]->list[i]->target != vertices[rot[1]]) && (vertices[rot[2]]->list[i]->target != vertices[rot[3]]))
          {
            if(vertices[rot[2]]->list[i]->target == vertices[rot[6]])
            {
              assert(candidate==-1);
              candidate = candidate_count;
            }
            candidate_count++;
          }
        }
        if(candidate_count)
        {
          //use edge from v2
          re_conn_candi->encode(rmCorner_EdgeJoinFrom, 0);
          //code which edge that is
          if(candidate_count > 1)
          {
            re_conn_candi->writeRange(candidate_count, candidate); //TODO: improve that with a range model
          }
        }
        else
        {
          //then try v5 if not succeeded
          for(i = 0; i < vertices[rot[5]]->list_size; i++)
          {
            if((vertices[rot[5]]->list[i]->target != vertices[rot[4]]) && (vertices[rot[5]]->list[i]->target != vertices[rot[1]]))
            {
              if(vertices[rot[5]]->list[i]->target == vertices[rot[6]])
              {
                assert(candidate==-1);
                candidate = candidate_count;
              }
              candidate_count++;
            }
          }
          if(candidate_count)
          {
            //use edge from v2
            re_conn_candi->encode(rmCorner_EdgeJoinFrom, 1);
            //code which edge that is
            if(candidate_count > 1)
            {
              re_conn_candi->writeRange(candidate_count, candidate); //TODO: improve that with a range model
            }
          }
          else
          {
            //then try v7 if not succeeded
            for(i = 0; i < vertices[rot[7]]->list_size; i++)
            {
              if((vertices[rot[7]]->list[i]->target != vertices[rot[4]]) && (vertices[rot[7]]->list[i]->target != vertices[rot[3]]))
              {
                if(vertices[rot[7]]->list[i]->target == vertices[rot[6]])
                {
                  assert(candidate==-1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            if(candidate_count)
            {
              //use edge from v2
              re_conn_candi->encode(rmCorner_EdgeJoinFrom, 2);
              //code which edge that is
              if(candidate_count > 1)
              {
                re_conn_candi->writeRange(candidate_count, candidate); //TODO: improve that with a range model
              }
            }
            else
            {
              //no edge, resort to direct coding.
              re_conn_candi->encode(rmCorner_EdgeJoinFrom, 3);
              int dv_index = dv->getRelativeIndex(vertices[rot[6]]);
              re_conn_index->writeRange(dv->size(), dv_index);
            }
          }
        }
#endif  
      }
      else
      {
        // check that the last vertex is not joined.
        assert(vertices[rot[6]]->use_count == 0);
        addAndCompressV6(vertices, rot);
      }

#ifdef SVH_COMPRESS_CELL_DATA
      //now compress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        compressCellDataCorner(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        compressCellDataCorner(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, cellDataTmp);
#endif
      }
#endif  
    }
  }
  else if(num_f_visited==4) //SVH_GAP and SVH_TUNNEL
  {
    //rotate the hexahedron such that the first visited face is f0
    if(f_visited[0])
      ;//nothing to do
    else if(f_visited[1])
      rotateHexXIQ(rot, f_visited);
    else if(f_visited[2])
      rotateHexYQ(rot, f_visited);
    else
    {
      fprintf(stderr, "ERROR: num_f_visited==4 but faces 0 to 2 are unvisited. \n");
      exit(1);
    }
    //and check
    assert(f_visited[0] != 0);
    //rotate the hexahedron such that the second visited face is f1
    if(f_visited[1])
      ;//nothing to do
    else if(f_visited[2])
      rotateHexZIQ(rot, f_visited);
    else if(f_visited[3])
      rotateHexZH(rot, f_visited);
    else
    {
      fprintf(stderr, "ERROR: num_f_visited==4 but only three faces are visited.\n");
      exit(1);
    }
    //and check
    assert(f_visited[1] != 0);


    if(!(f_visited[2] || f_visited[4])) //SVH_TUNNEL
    {
      op = SVH_TUNNEL;
#ifdef PRINT_CONTROL_OUTPUT
      op_tunnel++;
#endif
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
      re_conn_op->encode(rmOpX[last_fin][last_op], SVH_BRIDGE); //tell the decoder that this is a bridge, it will figure out alone that it is indeed a tunnel
#else
      re_conn_op->encode(rmOp[last_op], SVH_BRIDGE);
#endif
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "SVH_BRIDGE\n");
#endif

      //it is already in canonical orientation:
      assert(f_visited[0] != 0);
      assert(f_visited[1] != 0);
      assert(f_visited[5] != 0);
      assert(f_visited[3] != 0);

      if(num_v_cached>=4)
      {
        if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
        {
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next)) != -1)
        {
          rotateHexXIQ(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5]->next)) != -1)
        {
          rotateHexXH(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[3])) != -1)
        {
          rotateHexXQ(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
        {
          rotateHexZH(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next->next)) != -1)
        {
          rotateHexXIQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5]->next->next->next)) != -1)
        {
          rotateHexXH(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[3]->next->next)) != -1)
        {
          rotateHexXQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the first edge of f0 is in the face cache
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
        {
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next)) != -1)
        {
          rotateHexXIQ(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5]->next->next)) != -1)
        {
          rotateHexXH(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[3]->next)) != -1)
        {
          rotateHexXQ(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
        {
          rotateHexZH(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[1])) != -1)
        {
          rotateHexXIQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[5])) != -1)
        {
          rotateHexXH(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[3]->next->next->next)) != -1)
        {
          rotateHexXQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 12);
#endif
      if(lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        tunnel_hit_face++;
#endif
        if(lc_pos_f < 6)
          assert(lc_f->pos(f_visited[0]) == lc_pos_f);
        else if(lc_pos_f < 12)
          assert(lc_f->pos(f_visited[0]->next) == lc_pos_f-6);
        else
          assert(false);

#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmBT_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
        re_conn_cache->encode(rmBT_FaceCache[last_op],lc_pos_f);
#endif
      }
      else
      {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmBT_FaceCacheX[last_fin][last_op],12);
#else
        re_conn_cache->encode(rmBT_FaceCache[last_op],12);
#endif
        // check whether two vertices of one of the known
        // face is in the edge cache
        if(num_v_cached>=2)
        {
          if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[4]]->in_cache != -1))
          {
            rotateHexXIQ(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[5]]->in_cache != -1))
          {
            rotateHexXH(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[5]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
          {
            rotateHexXQ(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
          {
            rotateHexZH(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[7]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            rotateHexXIQ(rot, f_visited);
            rotateHexZH(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[6]]->in_cache != -1) &&(vertices[rot[7]]->in_cache != -1))
          {
            rotateHexXH(rot, f_visited);
            rotateHexZH(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[6]]->in_cache != -1))
          {
            rotateHexXQ(rot, f_visited);
            rotateHexZH(rot, f_visited);
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            //ok, second edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
          else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[7]]->in_cache != -1))
          {
            rotateHexXIQ(rot, f_visited);
            //ok, second edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
          else if(( vertices[rot[5]]->in_cache != -1) &&(vertices[rot[6]]->in_cache != -1))
          {
            rotateHexXH(rot, f_visited);
            //ok, second edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
          else if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
          {
            rotateHexXQ(rot, f_visited);
            //ok, second edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 48);
#endif
        if(lc_pos_e != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          tunnel_hit_edge++;
#endif
          re_conn_cache->encode(rmBT_EdgeCache[last_op],lc_pos_e);
          //then we have to tell the decoder which of the faces around this edge f0 is.
          if(lc_pos_e < 24)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v1 and v0 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 48)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1));
            // encode which of the faces incident to v0 and v3 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[0]]->list_size; i++)
            {
              if(vertices[rot[0]]->list[i]->target == vertices[rot[3]])
              {
                if(vertices[rot[0]]->list[i]->next->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else
            assert(false);
        }
        else
        {
          re_conn_cache->encode(rmBT_EdgeCache[last_op], 48);
          // check whether 1 vertex of f0 is in the vertex cache
          if(num_v_cached)
          {
            if(vertices[rot[1]]->in_cache != -1)
            {
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[0]]->in_cache != -1)
            {
              rotateHexXIQ(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[0]]->in_cache;
            }
            else if(vertices[rot[4]]->in_cache != -1)
            {
              rotateHexXH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[5]]->in_cache != -1)
            {
              rotateHexXQ(rot, f_visited); 
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[0]]->in_cache;
            }
            else if(vertices[rot[3]]->in_cache != -1)
            {
              rotateHexZH(rot, f_visited); 
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[7]]->in_cache != -1)
            {
              rotateHexXIQ(rot, f_visited);
              rotateHexZH(rot, f_visited); 
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[0]]->in_cache;
            }
            else if(vertices[rot[6]]->in_cache != -1)
            {
              rotateHexXH(rot, f_visited);
              rotateHexZH(rot, f_visited); 
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[2]]->in_cache != -1)
            {
              rotateHexXQ(rot, f_visited); 
              rotateHexZH(rot, f_visited); 
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[0]]->in_cache;
            }
          }
#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 16);
#endif
          if(lc_pos_v != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            tunnel_hit_vertex++;
#endif
            assert(vertices[rot[1]]->in_cache == lc_pos_v);

            re_conn_cache->encode(rmBT_VertexCache[last_op],lc_pos_v);
          }
          else
          {
#ifdef PRINT_CONTROL_OUTPUT
            tunnel_miss++;
#endif
            // no ... we cannot use a vertex from the cache
            re_conn_cache->encode(rmBT_VertexCache[last_op], 16);
            int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
            re_conn_index->writeRange(dv->size(),dv_index);
          }
          //then we have to tell the decoder which of the faces around this vertex f0 is.
          // encode which of the faces incident to <vertex> is f0
          // how many potential candidates are around that vertex?
          for(i = 0; i < vertices[rot[1]]->list_size; i++)
          {
            if(vertices[rot[1]]->list[i] == f_visited[0])
            {
              if(vertices[rot[1]]->list_size > 1)
              {
                if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                {
                  re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                }
                else
                {
                  re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                }
              }
              break;
            }
          }
          assert(i < vertices[rot[1]]->list_size);
        }
      }

      // encode which of the faces incident to v3 and v0 is f1
      int candidate=-1, candidate_count=0;
      for(i = 0; i < vertices[rot[3]]->list_size; i++)
      {
        if(vertices[rot[3]]->list[i]->target == vertices[rot[0]])
        {
          if(vertices[rot[3]]->list[i] == f_visited[1])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
      assert(candidate != -1);
      if(candidate_count > 1)
      {
        if(candidate_count == 2)
          re_conn_candi->encode(rm_FaceCandi,candidate);
        else
          re_conn_candi->writeRange(candidate_count,candidate);
      }

      // encode which of the faces incident to v1 and v2 is f3
      candidate=-1; candidate_count=0;
      for(i = 0; i < vertices[rot[1]]->list_size; i++)
      {
        if(vertices[rot[1]]->list[i]->target == vertices[rot[2]])
        {
          if(vertices[rot[1]]->list[i]->next->next->next == f_visited[3])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
      assert(candidate != -1);
      if(candidate_count > 1)
      {
        if(candidate_count == 2)
          re_conn_candi->encode(rm_FaceCandi,candidate);
        else
          re_conn_candi->writeRange(candidate_count,candidate);
      }

      // no need to encode the last face, it can be deducted by the decoder.
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "actual operation is SVH_TUNNEL\n");
#endif

#ifdef SVH_COMPRESS_CELL_DATA
      //now compress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        compressCellDataTunnel(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, f_visited[5]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, f_visited[5]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        compressCellDataTunnel(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[3]->c_prop_f, f_visited[5]->c_prop_f, cellDataTmp);
#endif
      }
#endif  
    }
    else //SVH_GAP
    {
      op = SVH_GAP; 
#ifdef PRINT_CONTROL_OUTPUT
      op_gap++;
#endif
      //tell the decoder that this is a step since it can figure out that it is indeed an SVH_GAP
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
      re_conn_op->encode(rmOpX[last_fin][last_op], SVH_STEP);
#else
      re_conn_op->encode(rmOp[last_op], SVH_STEP);
#endif
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "SVH_STEP\n");
#endif
      //put it into canonical orientation:
      if(f_visited[5])
        rotateHexXIQ(rot, f_visited);
      else if(!f_visited[3])
        rotateHexZQ(rot, f_visited);
      if(!f_visited[2])
        rotateHexZH(rot, f_visited);
      assert(f_visited[0] != 0);
      assert(f_visited[1] != 0);
      assert(f_visited[2] != 0);
      assert(f_visited[3] != 0);

      if(num_v_cached>=4)
      {
        if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
        {
          // great, the first edge of f0 is in the face cache, nothing to do!!!
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2]->next)) != -1)
        {
          // exchange faces
          rotateHexYQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the first edge of f0 is in the face cache.
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
        {
          // great, the second edge of f0 is in the face cache
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2]->next->next)) != -1)
        {
          // exchange faces
          rotateHexYQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the second edge of f0 is in the face cache.
          lc_pos_f += 6;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
        {
          // great, the third edge of f0 is in the face cache
          lc_pos_f += 12;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2]->next->next->next)) != -1)
        {
          // exchange faces
          rotateHexYQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the third edge of f0 is in the face cache.
          lc_pos_f += 12;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
        {
          // great, the third edge of f0 is in the face cache
          lc_pos_f += 18;
        }
        else if(( lc_pos_f = lc_f->pos(f_visited[2])) != -1)
        {
          // exchange faces
          rotateHexYQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          // great, the third edge of f0 is in the face cache.
          lc_pos_f += 18;
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 24);
#endif
      if(lc_pos_f != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        gap_hit_face++;
#endif
        if(lc_pos_f < 6)
          assert(lc_f->pos(f_visited[0]) == lc_pos_f);
        else if(lc_pos_f < 12)
          assert(lc_f->pos(f_visited[0]->next) == lc_pos_f-6);
        else if(lc_pos_f < 18)
          assert(lc_f->pos(f_visited[0]->next->next) == lc_pos_f-12);
        else
          assert(lc_f->pos(f_visited[0]->next->next->next) == lc_pos_f-18);

#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
        re_conn_cache->encode(rmSCGPD_FaceCache[last_op],lc_pos_f);
#endif
      }
      else
      {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
        re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],24);
#else
        re_conn_cache->encode(rmSCGPD_FaceCache[last_op],24);
#endif
        // check whether two vertices of one of the known
        // faces are in vertex cache. if yes, then rotate the hexa so that this face is f0
        if(num_v_cached>=2)
        {
          if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            //ok, first edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
          }
          //one case only for first edge
          else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
          {
            //ok, second edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
          else if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[5]]->in_cache != -1))
          {
            rotateHexYQ(rot, f_visited);
            rotateHexZH(rot, f_visited);
            //ok, second edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[0]]->in_cache, vertices[rot[3]]->in_cache) + 24;
          }
          else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
          {
            //ok, third edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[3]]->in_cache, vertices[rot[2]]->in_cache) + 48;
          }
          else if(( vertices[rot[5]]->in_cache != -1) &&(vertices[rot[4]]->in_cache != -1))
          {
            rotateHexYQ(rot, f_visited);
            rotateHexZH(rot, f_visited);
            //ok, third edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[3]]->in_cache, vertices[rot[2]]->in_cache) + 48;
          }
          else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
          {
            //ok, third edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[2]]->in_cache, vertices[rot[1]]->in_cache) + 72;
          }
          else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
          {
            rotateHexYQ(rot, f_visited);
            rotateHexZH(rot, f_visited);
            //ok, third edge is in the cache.
            lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[2]]->in_cache, vertices[rot[1]]->in_cache) + 72;
          }
          
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 96);
#endif
        if(lc_pos_e != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          gap_hit_edge++;
#endif
          re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],lc_pos_e);
          //then we have to tell the decoder which of the faces around this edge f0 is.
          if(lc_pos_e < 24)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v1 and v0 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 48)
          {
            assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1));
            // encode which of the faces incident to v0 and v3 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[0]]->list_size; i++)
            {
              if(vertices[rot[0]]->list[i]->target == vertices[rot[3]])
              {
                if(vertices[rot[0]]->list[i]->next->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 72)
          {
            assert(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1));
            // encode which of the faces incident to v3 and v2 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[3]]->list_size; i++)
            {
              if(vertices[rot[3]]->list[i]->target == vertices[rot[2]])
              {
                if(vertices[rot[3]]->list[i]->next->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else if(lc_pos_e < 96)
          {
            assert(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
            // encode which of the faces incident to v2 and v1 is f0
            int candidate=-1, candidate_count=0;
            for(i = 0; i < vertices[rot[2]]->list_size; i++)
            {
              if(vertices[rot[2]]->list[i]->target == vertices[rot[1]])
              {
                if(vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  assert(candidate == -1);
                  candidate = candidate_count;
                }
                candidate_count++;
              }
            }
            assert(candidate != -1);
            if(candidate_count > 1)
            {
              if(candidate_count == 2)
                re_conn_candi->encode(rm_FaceCandi,candidate);
              else
                re_conn_candi->writeRange(candidate_count,candidate);
            }
          }
          else
            assert(false);
        }
        else
        {
          re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],96);
          // check whether 1 vertex of one of the known
          // faces is in vertex cache. if yes, then rotate the hexa so that this face is f0
          //   lc_pos_v = 0 means that vertices[rot[1]] is 0th in the face cache
          //   lc_pos_v = 1 means that vertices[rot[1]] is 1st in the face cache
          //   ...
          //   lc_pos_v = 8 means that vertices[rot[0]] is 0th in the face cache
          //   ...
          //   lc_pos_v = 16 means that vertices[rot[2]] is 0th in the face cache
          //   ...
          //   lc_pos_v = 23 means that vertices[rot[2]] is 7th in the face cache
          //   lc_pos_v = 24 means that no vertices of f0 nor f1 are in the face cache
          if(num_v_cached)
          {
            if(vertices[rot[1]]->in_cache != -1)
            {
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            else if(vertices[rot[0]]->in_cache != -1)
            {
              rotateHexYQ(rot, f_visited);
              rotateHexZH(rot, f_visited);
              //ok, the 1st vertex of f0 is in the cache
              lc_pos_v = vertices[rot[1]]->in_cache;
            }
            //second vertex: already done.
            else if(vertices[rot[3]]->in_cache != -1)
            {
              //ok, the 3rd vertex of f0 is in the cache
              lc_pos_v = vertices[rot[3]]->in_cache + 16;
            }
            else if(vertices[rot[5]]->in_cache != -1)
            {
              rotateHexYQ(rot, f_visited);
              rotateHexZH(rot, f_visited);
              //ok, the 3rd vertex of f0 is in the cache
              lc_pos_v = vertices[rot[3]]->in_cache+16;
            }
            else if(vertices[rot[2]]->in_cache != -1)
            {
              //ok, the 4th vertex of f0 is in the cache
              lc_pos_v = vertices[rot[2]]->in_cache+24;
            }
            else if(vertices[rot[4]]->in_cache != -1)
            {
              rotateHexYQ(rot, f_visited);
              rotateHexZH(rot, f_visited);
              //ok, the 4th vertex of f0 is in the cache
              lc_pos_v = vertices[rot[2]]->in_cache+24;
            }
          }

#ifdef PRINT_DEBUG_OUTPUT
          fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 32);
#endif
          if(lc_pos_v != -1)
          {
#ifdef PRINT_CONTROL_OUTPUT
            gap_hit_vertex++;
#endif
            re_conn_cache->encode(rmSCGPD_VertexCache[last_op],lc_pos_v);
            if(lc_pos_v < 8)
            {
              assert(vertices[rot[1]]->in_cache == lc_pos_v);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[1]]->list_size; i++)
              {
                if(vertices[rot[1]]->list[i] == f_visited[0])
                {
                  if(vertices[rot[1]]->list_size > 1)
                  {
                    if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[1]]->list_size);
            }
            else if(lc_pos_v < 16)
            {
              assert(false);
            }
            else if(lc_pos_v < 24)
            {
              assert(vertices[rot[3]]->in_cache == lc_pos_v-16);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[3]]->list_size; i++)
              {
                if(vertices[rot[3]]->list[i]->next->next == f_visited[0])
                {
                  if(vertices[rot[3]]->list_size > 1)
                  {
                    if(vertices[rot[3]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[3]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[3]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[3]]->list_size);
            }
            else if(lc_pos_v < 32)
            {
              assert(vertices[rot[2]]->in_cache == lc_pos_v-24);
              //then we have to tell the decoder which of the faces around this vertex f0 is.
              // encode which of the faces incident to <vertex> is f0
              // how many potential candidates are around that vertex?
              for(i = 0; i < vertices[rot[2]]->list_size; i++)
              {
                if(vertices[rot[2]]->list[i]->next == f_visited[0])
                {
                  if(vertices[rot[2]]->list_size > 1)
                  {
                    if(vertices[rot[2]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                    {
                      re_conn_candi->encode(rm_VertexCandi[vertices[rot[2]]->list_size],i);
                    }
                    else
                    {
                      re_conn_candi->writeRange(vertices[rot[2]]->list_size,i);
                    }
                  }
                  break;
                }
              }
              assert(i < vertices[rot[2]]->list_size);
            }
            else
              assert(false);
          }
          else
          {
#ifdef PRINT_CONTROL_OUTPUT
            gap_miss++;
#endif
            // no ... we cannot use a vertex from the cache
            re_conn_cache->encode(rmSCGPD_VertexCache[last_op],32);
            int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
            re_conn_index->writeRange(dv->size(),dv_index);
            //then we have to tell the decoder which of the faces around this vertex f0 is.
            // encode which of the faces incident to <vertex> is f0
            // how many potential candidates are around that vertex?
            for(i = 0; i < vertices[rot[1]]->list_size; i++)
            {
              if(vertices[rot[1]]->list[i] == f_visited[0])
              {
                if(vertices[rot[1]]->list_size > 1)
                {
                  if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
                  {
                    re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
                  }
                  else
                  {
                    re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
                  }
                }
                break;
              }
            }
            assert(i < vertices[rot[1]]->list_size);
          }
          
        }
      }

      // encode which of the faces incident to v3 and v0 is f1
      int candidate=-1, candidate_count=0;
      for(i = 0; i < vertices[rot[3]]->list_size; i++)
      {
        if(vertices[rot[3]]->list[i]->target == vertices[rot[0]])
        {
          if(vertices[rot[3]]->list[i] == f_visited[1])
          {
            assert(candidate == -1);
            candidate = candidate_count;
          }
          candidate_count++;
        }
      }
      assert(candidate != -1);
      if(candidate_count > 1)
      {
        if(candidate_count == 2)
          re_conn_candi->encode(rm_FaceCandi,candidate);
        else
          re_conn_candi->writeRange(candidate_count,candidate);
      }

      // no need to encode the other faces, 
      // the decoder can figure it out itself
      
      //confirm the SVH_CORNER face merge
      re_conn->encode(rm_ConfirmFaceMerge, 1);
#ifdef PRINT_CONTROL_OUTPUT
      confim_yes++;
#endif
      //confirm the SVH_GAP face merge
      re_conn->encode(rm_ConfirmFaceMerge, 1);
#ifdef PRINT_CONTROL_OUTPUT
      confim_yes++;
#endif


#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr, "actual operation is SVH_GAP\n");
#endif

#ifdef SVH_COMPRESS_CELL_DATA
      //now compress the cell data.
      if(c_pnum_f)
      {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        compressCellDataGap(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
        compressCellDataGap(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, cellDataTmp);
#endif
      }
#endif  
    }
  }
  else if(num_f_visited==5) //SVH_PIT
  {
    op = SVH_PIT; 
#ifdef PRINT_CONTROL_OUTPUT
    op_pit++;
#endif
    //tell the decoder that this is a step since it can figure out that it is indeed an SVH_PIT
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
    re_conn_op->encode(rmOpX[last_fin][last_op], SVH_STEP);
#else
    re_conn_op->encode(rmOp[last_op], SVH_STEP);
#endif
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "SVH_STEP\n");
#endif

    //put the hex in canonical position
    if(!f_visited[5])
      ; //nothing to do
    else if(!f_visited[0])
      rotateHexXH(rot, f_visited);
    else if(!f_visited[1])
      rotateHexXQ(rot, f_visited);
    else if(!f_visited[3])
      rotateHexXIQ(rot, f_visited);
    else if(!f_visited[4])
      rotateHexYQ(rot, f_visited);
    else if(!f_visited[2])
      rotateHexYIQ(rot, f_visited);
    
    //and check
    assert(f_visited[0] != 0);
    assert(f_visited[1] != 0);
    assert(f_visited[2] != 0);
    assert(f_visited[3] != 0);
    assert(f_visited[4] != 0);
    
    if(num_v_cached>=4)
    {
      if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
      {
        // great, the first edge of f0 is in the face cache
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
      {
        rotateHexZQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
      {
        rotateHexZH(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
      {
        rotateHexZIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
    }
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 24);
#endif
    if(lc_pos_f != -1)
    {
#ifdef PRINT_CONTROL_OUTPUT
      pit_hit_face++;
#endif
      assert(lc_f->pos(f_visited[0]) == lc_pos_f);

#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
      re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
      re_conn_cache->encode(rmSCGPD_FaceCache[last_op],lc_pos_f);
#endif
    }
    else
    {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
      re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],24);
#else
      re_conn_cache->encode(rmSCGPD_FaceCache[last_op],24);
#endif
      // check whether two vertices of one of the known
      // faces are in vertex cache. if yes, then rotate the hexa so that this face is f0
      if(num_v_cached>=2)
      {
        if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
        {
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
        {
          rotateHexZQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
        {
          rotateHexZH(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
        {
          rotateHexZIQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 96);
#endif
      if(lc_pos_e != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        pit_hit_edge++;
#endif
        re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],lc_pos_e);
        //then we have to tell the decoder which of the faces around this edge f0 is.
        
        assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
        // encode which of the faces incident to v1 and v0 is f0
        int candidate=-1, candidate_count=0;
        for(i = 0; i < vertices[rot[1]]->list_size; i++)
        {
          if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
          {
            if(vertices[rot[1]]->list[i] == f_visited[0])
            {
              assert(candidate == -1);
              candidate = candidate_count;
            }
            candidate_count++;
          }
        }
        assert(candidate != -1);
        if(candidate_count > 1)
        {
          if(candidate_count == 2)
            re_conn_candi->encode(rm_FaceCandi,candidate);
          else
            re_conn_candi->writeRange(candidate_count,candidate);
        }
      }
      else
      {
        re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],96);
        // check whether 1 vertex of one of the known
        // faces is in vertex cache. if yes, then rotate the hexa so that this face is f0
        //   lc_pos_v = 0 means that vertices[rot[1]] is 0th in the face cache
        //   lc_pos_v = 1 means that vertices[rot[1]] is 1st in the face cache
        //   ...
        //   lc_pos_v = 8 means that vertices[rot[0]] is 0th in the face cache
        //   ...
        //   lc_pos_v = 16 means that vertices[rot[2]] is 0th in the face cache
        //   ...
        //   lc_pos_v = 23 means that vertices[rot[2]] is 7th in the face cache
        //   lc_pos_v = 24 means that no vertices of f0 nor f1 are in the face cache
        if(num_v_cached)
        {
          if(vertices[rot[1]]->in_cache != -1)
          {
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[0]]->in_cache != -1)
          {
            rotateHexZQ(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[3]]->in_cache != -1)
          {
            rotateHexZH(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[2]]->in_cache != -1)
          {
            rotateHexZIQ(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 32);
#endif
        if(lc_pos_v != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          pit_hit_vertex++;
#endif
          assert(vertices[rot[1]]->in_cache == lc_pos_v);

          re_conn_cache->encode(rmSCGPD_VertexCache[last_op],lc_pos_v);
        }
        else
        {
#ifdef PRINT_CONTROL_OUTPUT
          pit_miss++;
#endif
          // no ... we cannot use a vertex from the cache
          re_conn_cache->encode(rmSCGPD_VertexCache[last_op],32);
          int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
          re_conn_index->writeRange(dv->size(),dv_index);
        }
        //then we have to tell the decoder which of the faces around this vertex f0 is.
        // encode which of the faces incident to <vertex> is f0
        // how many potential candidates are around that vertex?
        for(i = 0; i < vertices[rot[1]]->list_size; i++)
        {
          if(vertices[rot[1]]->list[i] == f_visited[0])
          {
            if(vertices[rot[1]]->list_size > 1)
            {
              if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
              {
                re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
              }
              else
              {
                re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
              }
            }
            break;
          }
        }
        assert(i < vertices[rot[1]]->list_size);
      }
    }

    // encode which of the faces incident to v3 and v0 is f1
    int candidate=-1, candidate_count=0;
    for(i = 0; i < vertices[rot[3]]->list_size; i++)
    {
      if(vertices[rot[3]]->list[i]->target == vertices[rot[0]])
      {
        if(vertices[rot[3]]->list[i] == f_visited[1])
        {
          assert(candidate == -1);
          candidate = candidate_count;
        }
        candidate_count++;
      }
    }
    assert(candidate != -1);
    if(candidate_count > 1)
    {
      if(candidate_count == 2)
        re_conn_candi->encode(rm_FaceCandi,candidate);
      else
        re_conn_candi->writeRange(candidate_count,candidate);
    }

    // no need to encode the other faces, 
    // the decoder can figure it out itself
    
    //confirm the SVH_CORNER face merge
    re_conn->encode(rm_ConfirmFaceMerge, 1);
#ifdef PRINT_CONTROL_OUTPUT
    confim_yes++;
#endif
    //confirm the SVH_GAP face merge
    re_conn->encode(rm_ConfirmFaceMerge, 1);
#ifdef PRINT_CONTROL_OUTPUT
    confim_yes++;
#endif

#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "actual operation is SVH_PIT\n");
#endif

#ifdef SVH_COMPRESS_CELL_DATA
    //now compress the cell data.
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      compressCellDataPit(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, f_visited[4]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
      compressCellDataPit(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, cellDataTmp);
#endif
    }
#endif  
  }
  else if(num_f_visited==6) //SVH_DEN
  {
    op = SVH_DEN; 
#ifdef PRINT_CONTROL_OUTPUT
    op_den++;
#endif
    //tell the decoder that this is a step since it can figure out that it is indeed an SVH_DEN
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
    re_conn_op->encode(rmOpX[last_fin][last_op], SVH_STEP);
#else
    re_conn_op->encode(rmOp[last_op], SVH_STEP);
#endif
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "SVH_STEP\n");
#endif

    //put the hex is in canonical position
    
    if(num_v_cached>=4)
    {
      if(( lc_pos_f = lc_f->pos(f_visited[0])) != -1)
      {
        // great, the first edge of f0 is in the face cache
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next)) != -1)
      {
        rotateHexZQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next)) != -1)
      {
        rotateHexZH(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[0]->next->next->next)) != -1)
      {
        rotateHexZIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[1]->next)) != -1)
      {
        rotateHexXIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next)) != -1)
      {
        rotateHexXIQ(rot, f_visited);
        rotateHexZQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[1]->next->next->next)) != -1)
      {
        rotateHexXIQ(rot, f_visited);
        rotateHexZH(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[1])) != -1)
      {
        rotateHexXIQ(rot, f_visited);
        rotateHexZIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[5]->next)) != -1)
      {
        rotateHexXH(rot, f_visited);
        // great, the first edge of f0 is in the face cache
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[5]->next->next)) != -1)
      {
        rotateHexXH(rot, f_visited);
        rotateHexZQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[5]->next->next->next)) != -1)
      {
        rotateHexXH(rot, f_visited);
        rotateHexZH(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[5])) != -1)
      {
        rotateHexXH(rot, f_visited);
        rotateHexZIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[3])) != -1)
      {
        rotateHexXQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[3]->next)) != -1)
      {
        rotateHexXQ(rot, f_visited);
        rotateHexZQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[3]->next->next)) != -1)
      {
        rotateHexXQ(rot, f_visited);
        rotateHexZH(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[3]->next->next->next)) != -1)
      {
        rotateHexXQ(rot, f_visited);
        rotateHexZIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[2]->next->next->next)) != -1)
      {
        rotateHexYQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[2])) != -1)
      {
        rotateHexYQ(rot, f_visited);
        rotateHexZQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[2]->next)) != -1)
      {
        rotateHexYQ(rot, f_visited);
        rotateHexZH(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[2]->next->next)) != -1)
      {
        rotateHexYQ(rot, f_visited);
        rotateHexZIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[4]->next)) != -1)
      {
        rotateHexYIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[4]->next->next)) != -1)
      {
        rotateHexYIQ(rot, f_visited);
        rotateHexZQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[4]->next->next->next)) != -1)
      {
        rotateHexYIQ(rot, f_visited);
        rotateHexZH(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
      else if(( lc_pos_f = lc_f->pos(f_visited[4])) != -1)
      {
        rotateHexYIQ(rot, f_visited);
        rotateHexZIQ(rot, f_visited);
        // great, the first edge of f0 is in the face cache.
      }
    }
#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr,  "lc_pos_f is %i\n", lc_pos_f!=-1 ? lc_pos_f : 24);
#endif
    if(lc_pos_f != -1)
    {
#ifdef PRINT_CONTROL_OUTPUT
      den_hit_face++;
#endif
      assert(lc_f->pos(f_visited[0]) == lc_pos_f);

#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
      re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],lc_pos_f);
#else
      re_conn_cache->encode(rmSCGPD_FaceCache[last_op],lc_pos_f);
#endif
    }
    else
    {
#ifdef SVH_FACECACHE_DEPENDS_ON_FINALIZATION
      re_conn_cache->encode(rmSCGPD_FaceCacheX[last_fin][last_op],24);
#else
      re_conn_cache->encode(rmSCGPD_FaceCache[last_op],24);
#endif
      // check whether two vertices of one of the known
      // faces are in vertex cache. if yes, then rotate the hexa so that this face is f0
      if(num_v_cached>=2)
      {
        if(( vertices[rot[1]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
        {
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
        {
          rotateHexZQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[3]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
        {
          rotateHexZH(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[2]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
        {
          rotateHexZIQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[5]]->in_cache != -1))
        {
          rotateHexXH(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[5]]->in_cache != -1) &&(vertices[rot[6]]->in_cache != -1))
        {
          rotateHexXH(rot, f_visited);
          rotateHexZQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[6]]->in_cache != -1) &&(vertices[rot[7]]->in_cache != -1))
        {
          rotateHexXH(rot, f_visited);
          rotateHexZH(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[7]]->in_cache != -1) &&(vertices[rot[4]]->in_cache != -1))
        {
          rotateHexXH(rot, f_visited);
          rotateHexZIQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[4]]->in_cache != -1) &&(vertices[rot[0]]->in_cache != -1))
        {
          rotateHexXIQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[5]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1))
        {
          rotateHexXQ(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[6]]->in_cache != -1) &&(vertices[rot[2]]->in_cache != -1))
        {
          rotateHexXQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
        else if(( vertices[rot[7]]->in_cache != -1) &&(vertices[rot[3]]->in_cache != -1))
        {
          rotateHexXIQ(rot, f_visited);
          rotateHexZH(rot, f_visited);
          //ok, first edge is in the cache.
          lc_pos_e=getEdgeNumberFromLocalIndices(vertices[rot[1]]->in_cache, vertices[rot[0]]->in_cache);
        }
      }
#ifdef PRINT_DEBUG_OUTPUT
      fprintf(stderr,  "lc_pos_e is %i\n", lc_pos_e!=-1 ? lc_pos_e : 96);
#endif
      if(lc_pos_e != -1)
      {
#ifdef PRINT_CONTROL_OUTPUT
        den_hit_edge++;
#endif
        re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],lc_pos_e);
        //then we have to tell the decoder which of the faces around this edge f0 is.
        
        assert(( vertices[rot[0]]->in_cache != -1) &&(vertices[rot[1]]->in_cache != -1));
        // encode which of the faces incident to v1 and v0 is f0
        int candidate=-1, candidate_count=0;
        for(i = 0; i < vertices[rot[1]]->list_size; i++)
        {
          if(vertices[rot[1]]->list[i]->target == vertices[rot[0]])
          {
            if(vertices[rot[1]]->list[i] == f_visited[0])
            {
              assert(candidate == -1);
              candidate = candidate_count;
            }
            candidate_count++;
          }
        }
        assert(candidate != -1);
        if(candidate_count > 1)
        {
          if(candidate_count == 2)
            re_conn_candi->encode(rm_FaceCandi,candidate);
          else
            re_conn_candi->writeRange(candidate_count,candidate);
        }
      }
      else
      {
        re_conn_cache->encode(rmSCGPD_EdgeCache[last_op],96);
        // check whether 1 vertex of one of the known
        // faces is in vertex cache. if yes, then rotate the hexa so that this face is f0
        //   lc_pos_v = 0 means that vertices[rot[1]] is 0th in the face cache
        //   lc_pos_v = 1 means that vertices[rot[1]] is 1st in the face cache
        //   ...
        //   lc_pos_v = 8 means that vertices[rot[0]] is 0th in the face cache
        //   ...
        //   lc_pos_v = 16 means that vertices[rot[2]] is 0th in the face cache
        //   ...
        //   lc_pos_v = 23 means that vertices[rot[2]] is 7th in the face cache
        //   lc_pos_v = 24 means that no vertices of f0 nor f1 are in the face cache
        if(num_v_cached)
        {
          if(vertices[rot[1]]->in_cache != -1)
          {
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[0]]->in_cache != -1)
          {
            rotateHexZQ(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[3]]->in_cache != -1)
          {
            rotateHexZH(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[2]]->in_cache != -1)
          {
            rotateHexZIQ(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[4]]->in_cache != -1)
          {
            rotateHexXH(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[5]]->in_cache != -1)
          {
            rotateHexXH(rot, f_visited);
            rotateHexZQ(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[6]]->in_cache != -1)
          {
            rotateHexXH(rot, f_visited);
            rotateHexZH(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
          else if(vertices[rot[7]]->in_cache != -1)
          {
            rotateHexXH(rot, f_visited);
            rotateHexZIQ(rot, f_visited);
            //ok, the 1st vertex of f0 is in the cache
            lc_pos_v = vertices[rot[1]]->in_cache;
          }
        }
#ifdef PRINT_DEBUG_OUTPUT
        fprintf(stderr,  "lc_pos_v is %i\n", lc_pos_v!=-1 ? lc_pos_v : 32);
#endif
        if(lc_pos_v != -1)
        {
#ifdef PRINT_CONTROL_OUTPUT
          den_hit_vertex++;
#endif
          assert(vertices[rot[1]]->in_cache == lc_pos_v);

          re_conn_cache->encode(rmSCGPD_VertexCache[last_op],lc_pos_v);
        }
        else
        {
#ifdef PRINT_CONTROL_OUTPUT
          den_miss++;
#endif
          // no ... we cannot use a vertex from the cache
          re_conn_cache->encode(rmSCGPD_VertexCache[last_op],32);
          int dv_index = dv->getRelativeIndex(vertices[rot[1]]);
          re_conn_index->writeRange(dv->size(),dv_index);
        }
        //then we have to tell the decoder which of the faces around this vertex f0 is.
        // encode which of the faces incident to <vertex> is f0
        // how many potential candidates are around that vertex?
        for(i = 0; i < vertices[rot[1]]->list_size; i++)
        {
          if(vertices[rot[1]]->list[i] == f_visited[0])
          {
            if(vertices[rot[1]]->list_size > 1)
            {
              if(vertices[rot[1]]->list_size < SVH_MAGIC_NUMBER_F0_AROUND_V1_RM)
              {
                re_conn_candi->encode(rm_VertexCandi[vertices[rot[1]]->list_size],i);
              }
              else
              {
                re_conn_candi->writeRange(vertices[rot[1]]->list_size,i);
              }
            }
            break;
          }
        }
        assert(i < vertices[rot[1]]->list_size);
      }
    }

    // encode which of the faces incident to v3 and v0 is f1
    int candidate=-1, candidate_count=0;
    for(i = 0; i < vertices[rot[3]]->list_size; i++)
    {
      if(vertices[rot[3]]->list[i]->target == vertices[rot[0]])
      {
        if(vertices[rot[3]]->list[i] == f_visited[1])
        {
          assert(candidate == -1);
          candidate = candidate_count;
        }
        candidate_count++;
      }
    }
    assert(candidate != -1);
    if(candidate_count > 1)
    {
      if(candidate_count == 2)
        re_conn_candi->encode(rm_FaceCandi,candidate);
      else
        re_conn_candi->writeRange(candidate_count,candidate);
    }

    // no need to encode the other faces, 
    // the decoder can figure it out itself

    //confirm the SVH_CORNER face merge
    re_conn->encode(rm_ConfirmFaceMerge, 1);
#ifdef PRINT_CONTROL_OUTPUT
    confim_yes++;
#endif
      //confirm the SVH_GAP face merge
    re_conn->encode(rm_ConfirmFaceMerge, 1);
#ifdef PRINT_CONTROL_OUTPUT
    confim_yes++;
#endif

#ifdef PRINT_DEBUG_OUTPUT
    fprintf(stderr, "actual operation is SVH_DEN\n");
#endif

#ifdef SVH_COMPRESS_CELL_DATA
    //now compress the cell data.
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      compressCellDataDen(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, f_visited[5]->c_prop_f, cellDataTmp, f_visited[0]->c_prop_f+c_pnum_f, f_visited[1]->c_prop_f+c_pnum_f, f_visited[2]->c_prop_f+c_pnum_f, f_visited[3]->c_prop_f+c_pnum_f, f_visited[4]->c_prop_f+c_pnum_f, f_visited[5]->c_prop_f+c_pnum_f, tmpCellDataResidual);
#else
      compressCellDataDen(f_visited[0]->c_prop_f, f_visited[1]->c_prop_f, f_visited[2]->c_prop_f, f_visited[3]->c_prop_f, f_visited[4]->c_prop_f, f_visited[5]->c_prop_f, cellDataTmp);
#endif
    }
#endif  
  }

#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr,"op %d %d\n",op,c_count);
#endif

  // statistic
  //fprintf(stderr,"reordered %d %d %d %d\n",vertices[rot[0]]->index, vertices[rot[1]]->index, vertices[rot[2]]->index, vertices[rot[3]]->index);

  // update data structures
  if(f_visited[0])
  {
    removeEdgeFromVertex(f_visited[0]->next->next->next, vertices[rot[2]]);
    removeEdgeFromVertex(f_visited[0]->next->next, vertices[rot[3]]);
    removeEdgeFromVertex(f_visited[0]->next, vertices[rot[0]]);
    removeEdgeFromVertex(f_visited[0], vertices[rot[1]]);

    deallocEdge(f_visited[0]->next->next->next);
    deallocEdge(f_visited[0]->next->next);
    deallocEdge(f_visited[0]->next);
    deallocEdge(f_visited[0]);

    f_visited[0] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
#endif
    }
#endif

    e0->target = vertices[rot[1]];
    e1->target = vertices[rot[2]];
    e2->target = vertices[rot[3]];
    e3->target = vertices[rot[0]];

    addEdgeToVertex(e0, vertices[rot[0]]);
    addEdgeToVertex(e1, vertices[rot[1]]);
    addEdgeToVertex(e2, vertices[rot[2]]);
    addEdgeToVertex(e3, vertices[rot[3]]);

    f_visited[0] = e0;
  }

  if(f_visited[1])
  {
    removeEdgeFromVertex(f_visited[1]->next->next->next, vertices[rot[7]]);
    removeEdgeFromVertex(f_visited[1]->next->next, vertices[rot[4]]);
    removeEdgeFromVertex(f_visited[1]->next, vertices[rot[0]]);
    removeEdgeFromVertex(f_visited[1], vertices[rot[3]]);

    deallocEdge(f_visited[1]->next->next->next);
    deallocEdge(f_visited[1]->next->next);
    deallocEdge(f_visited[1]->next);
    deallocEdge(f_visited[1]);

    f_visited[1] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[rot[3]];
    e1->target = vertices[rot[7]];
    e2->target = vertices[rot[4]];
    e3->target = vertices[rot[0]];

    addEdgeToVertex(e0, vertices[rot[0]]);
    addEdgeToVertex(e1, vertices[rot[3]]);
    addEdgeToVertex(e2, vertices[rot[7]]);
    addEdgeToVertex(e3, vertices[rot[4]]);

    f_visited[1] = e0;
  }

  if(f_visited[2])
  {
    removeEdgeFromVertex(f_visited[2]->next->next->next, vertices[rot[5]]);
    removeEdgeFromVertex(f_visited[2]->next->next, vertices[rot[1]]);
    removeEdgeFromVertex(f_visited[2]->next, vertices[rot[0]]);
    removeEdgeFromVertex(f_visited[2], vertices[rot[4]]);

    deallocEdge(f_visited[2]->next->next->next);
    deallocEdge(f_visited[2]->next->next);
    deallocEdge(f_visited[2]->next);
    deallocEdge(f_visited[2]);

    f_visited[2] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[rot[4]];
    e1->target = vertices[rot[5]];
    e2->target = vertices[rot[1]];
    e3->target = vertices[rot[0]];

    addEdgeToVertex(e0, vertices[rot[0]]);
    addEdgeToVertex(e1, vertices[rot[4]]);
    addEdgeToVertex(e2, vertices[rot[5]]);
    addEdgeToVertex(e3, vertices[rot[1]]);

    f_visited[2] = e0;
  }

  if(f_visited[3])
  {
    removeEdgeFromVertex(f_visited[3]->next->next->next, vertices[rot[6]]);
    removeEdgeFromVertex(f_visited[3]->next->next, vertices[rot[2]]);
    removeEdgeFromVertex(f_visited[3]->next, vertices[rot[1]]);
    removeEdgeFromVertex(f_visited[3], vertices[rot[5]]);

    deallocEdge(f_visited[3]->next->next->next);
    deallocEdge(f_visited[3]->next->next);
    deallocEdge(f_visited[3]->next);
    deallocEdge(f_visited[3]);

    f_visited[3] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[rot[5]];
    e1->target = vertices[rot[6]];
    e2->target = vertices[rot[2]];
    e3->target = vertices[rot[1]];

    addEdgeToVertex(e0, vertices[rot[1]]);
    addEdgeToVertex(e1, vertices[rot[5]]);
    addEdgeToVertex(e2, vertices[rot[6]]);
    addEdgeToVertex(e3, vertices[rot[2]]);

    f_visited[3] = e0;
  }

  if(f_visited[4])
  {
    removeEdgeFromVertex(f_visited[4]->next->next->next, vertices[rot[7]]);
    removeEdgeFromVertex(f_visited[4]->next->next, vertices[rot[3]]);
    removeEdgeFromVertex(f_visited[4]->next, vertices[rot[2]]);
    removeEdgeFromVertex(f_visited[4], vertices[rot[6]]);

    deallocEdge(f_visited[4]->next->next->next);
    deallocEdge(f_visited[4]->next->next);
    deallocEdge(f_visited[4]->next);
    deallocEdge(f_visited[4]);

    f_visited[4] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[rot[6]];
    e1->target = vertices[rot[7]];
    e2->target = vertices[rot[3]];
    e3->target = vertices[rot[2]];

    addEdgeToVertex(e0, vertices[rot[2]]);
    addEdgeToVertex(e1, vertices[rot[6]]);
    addEdgeToVertex(e2, vertices[rot[7]]);
    addEdgeToVertex(e3, vertices[rot[3]]);

    f_visited[4] = e0;
  }

  if(f_visited[5])
  {
    removeEdgeFromVertex(f_visited[5]->next->next->next, vertices[rot[6]]);
    removeEdgeFromVertex(f_visited[5]->next->next, vertices[rot[5]]);
    removeEdgeFromVertex(f_visited[5]->next, vertices[rot[4]]);
    removeEdgeFromVertex(f_visited[5], vertices[rot[7]]);

    deallocEdge(f_visited[5]->next->next->next);
    deallocEdge(f_visited[5]->next->next);
    deallocEdge(f_visited[5]->next);
    deallocEdge(f_visited[5]);

    f_visited[5] = 0;
  }
  else
  {
    SVedge* e0 = allocEdge();
    SVedge* e1 = allocEdge();
    SVedge* e2 = allocEdge();
    SVedge* e3 = allocEdge();

    e0->next = e1;
    e1->next = e2;
    e2->next = e3;
    e3->next = e0;

#ifdef SVH_COMPRESS_CELL_DATA
    if(c_pnum_f)
    {
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e0->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f+c_pnum_f, tmpCellDataResidual, c_pnum_f*sizeof(float));
#else
      memcpy(e0->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e1->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e2->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
      memcpy(e3->c_prop_f, cellDataTmp, c_pnum_f*sizeof(float));
#endif
    }
#endif
    e0->target = vertices[rot[7]];
    e1->target = vertices[rot[6]];
    e2->target = vertices[rot[5]];
    e3->target = vertices[rot[4]];

    addEdgeToVertex(e0, vertices[rot[4]]);
    addEdgeToVertex(e1, vertices[rot[7]]);
    addEdgeToVertex(e2, vertices[rot[6]]);
    addEdgeToVertex(e3, vertices[rot[5]]);

    f_visited[5] = e0;
  }

  // update last_op
  last_op = op;

  // increment vertex use_counts, create edges, and update edge degrees

  for(i = 0; i < 8; i++)
  {
    vertices[i]->use_count++;
  }

#if defined(SVH_FACECACHE_DEPENDS_ON_FINALIZATION) || defined(SVH_RMOP_DEPENDS_ON_FINALIZATION)
  last_fin=0;
#endif
  // write finalization info and finalize vertices
  for(i = 0; i < 8; i++)
  {

#ifdef SVH_IMPLICIT_FINALIZATION
    //guess finalization
    if (vertices[rot[i]]->list_size == 0)
    {
      vertices[rot[i]]->finalize = 1;
  #if defined(SVH_FACECACHE_DEPENDS_ON_FINALIZATION) || defined(SVH_RMOP_DEPENDS_ON_FINALIZATION)
      last_fin = last_fin | (1<<i);
  #endif
      //delete finalized vertex
      SVvertex* vertex = vertices[rot[i]];
      vertices[rot[i]] = 0;
      vertex_hash->erase(hash_elements[rot[i]]);
      dv->removeElement(vertex);
      for (int j = 0; j < vertex->list_size; j++)
      {
        SVedge* edge = vertex->list[j];
        removeEdgeFromVertex(edge->next->next->next, edge->next->next->target);
        removeEdgeFromVertex(edge->next->next, edge->next->target);
        removeEdgeFromVertex(edge->next, edge->target);
        //removeEdgeFromVertex(edge, vertex);
        deallocEdge(edge->next->next->next);
        deallocEdge(edge->next->next);
        deallocEdge(edge->next);
        deallocEdge(edge);
      }
      deallocVertex(vertex);
    }
#else
    //transmit finalization info
    int degree_one = vertices[rot[i]]->list_size;
    if(degree_one >= SVH_MAX_DEGREE_ONE)
    {
      degree_one = SVH_MAX_DEGREE_ONE-1;
    }
    int use_count = vertices[rot[i]]->use_count;
    if(use_count >= SVH_MAX_USE_COUNT)
    {
      use_count = SVH_MAX_USE_COUNT-1;
    }
    re_conn_final->encode(rmFinalized[degree_one][use_count], vertices[rot[i]]->finalize);
    //delete finalized vertex
    if(vertices[rot[i]]->finalize)
    {
  #if defined(SVH_FACECACHE_DEPENDS_ON_FINALIZATION) || defined(SVH_RMOP_DEPENDS_ON_FINALIZATION)
      last_fin = last_fin | (1<<i);
  #endif

      SVvertex* vertex = vertices[rot[i]];
      vertices[rot[i]] = 0;
      vertex_hash->erase(hash_elements[rot[i]]);
      dv->removeElement(vertex);
      for (int j = 0; j < vertex->list_size; j++)
      {
        SVedge* edge = vertex->list[j];
        removeEdgeFromVertex(edge->next->next->next, edge->next->next->target);
        removeEdgeFromVertex(edge->next->next, edge->next->target);
        removeEdgeFromVertex(edge->next, edge->target);
        //removeEdgeFromVertex(edge, edge->next->next->next->target);
        deallocEdge(edge->next->next->next);
        deallocEdge(edge->next->next);
        deallocEdge(edge->next);
        deallocEdge(edge);
      }
      deallocVertex(vertex);
    }
#endif
  }

  // clean cache flag of current elements in cache
  for(i = 0; i < 8; i++)
  {
    SVvertex* vertex =(SVvertex*) lc->get(i);
    if(vertex) vertex->in_cache = -1;
  }

  // update cache elements
  lc->put(vertices[rot[0]], vertices[rot[1]], vertices[rot[2]], vertices[rot[3]], vertices[rot[4]], vertices[rot[5]], vertices[rot[6]], vertices[rot[7]]);
  lc_f->put(f_visited[0], f_visited[1], f_visited[2], f_visited[3], f_visited[4], f_visited[5]);

  // set cache flag of current elements in cache
  for(i = 0; i < 8; i++)
  {
    if(vertices[rot[i]]) vertices[rot[i]]->in_cache = i;
  }

#ifdef PRINT_DEBUG_OUTPUT
  fprintf(stderr, "END (con/de)compress_hexahedron()\n");
#endif

}


int SVwriter_svh::test_write_hexahedron(const I32 *c_idx)
{
  int i;

  my_vertex_hash::iterator hash_elements[8];
  SVvertex* vertices[8];

  // get vertices from hash
  for(i = 0; i < 8; i++)
  {
    hash_elements[i] = vertex_hash->find(c_idx[i]);
    if(hash_elements[i] == vertex_hash->end())
      vertices[i] = NULL;
    else
      vertices[i] =(*hash_elements[i]).second;
  }

  // check which vertices are already visited
  SVvertex* v_visited[] = {0,0,0,0,0,0,0,0};
  int num_v_visited = 0;

  for(i = 0; i < 8; i++)
  {
    if (vertices[i] && vertices[i]->use_count)
    {
      num_v_visited++;
      v_visited[i] = vertices[i];
    }
  }

  // check which faces are already visited.
  // note: since the already visited faces point outwards from current advancing front, the face vertices should be given in reverse order (i.e. pointing inside hexahedron).
  SVedge* f_visited[] = {NULL,NULL,NULL,NULL,NULL,NULL};
  if(num_v_visited)
  {
    // check face 0 (0,1,2,3) halfedge 1->0
    f_visited[0] = isFaceVisited(v_visited[0], v_visited[1], v_visited[2], v_visited[3]);
    // check face 1 (0,3,7,4) halfedge 3->0
    f_visited[1] = isFaceVisited(v_visited[0], v_visited[3], v_visited[7], v_visited[4]);
    // check face 2 (0,4,5,1) halfedge 4->0
    f_visited[2] = isFaceVisited(v_visited[0], v_visited[4], v_visited[5], v_visited[1]);
    // check face 3 (1,5,6,2) halfedge 5->1
    f_visited[3] = isFaceVisited(v_visited[1], v_visited[5], v_visited[6], v_visited[2]);
    // check face 4 (2,6,7,3) halfedge 6->2
    f_visited[4] = isFaceVisited(v_visited[2], v_visited[6], v_visited[7], v_visited[3]);
    // check face 5 (4,7,6,5) halfedge 7->4
    f_visited[5] = isFaceVisited(v_visited[4], v_visited[7], v_visited[6], v_visited[5]);
  }

  //SVH_CORNER case (test the eight corners):
  if(f_visited[0])
  {
    if(f_visited[1])
    {
      if(f_visited[2])
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_DEN;
            }
            else
            {
              return SVH_PIT;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_PIT;
            }
            else
            {
              return SVH_GAP;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_PIT;
            }
            else
            {
              return SVH_GAP;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_CORNER;
            }
          }
        }
      }
      else
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_PIT;
            }
            else
            {
              return SVH_GAP;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_TUNNEL;
            }
            else
            {
              return SVH_BRIDGE;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_CORNER;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_BRIDGE;
            }
            else
            {
              return SVH_STEP;
            }
          }
        }
      }
    }
    else
    {
      if(f_visited[2])
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_PIT;
            }
            else
            {
              return SVH_GAP;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_CORNER;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_TUNNEL;
            }
            else
            {
              return SVH_BRIDGE;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_BRIDGE;
            }
            else
            {
              return SVH_STEP;
            }
          }
        }
      }
      else
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_CORNER;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_BRIDGE;
            }
            else
            {
              return SVH_STEP;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_BRIDGE;
            }
            else
            {
              return SVH_STEP;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_ROOF;
            }
            else
            {
              return SVH_HUT;
            }
          }
        }
      }
    }
  }
  else
  {
    if(f_visited[1])
    {
      if(f_visited[2])
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_PIT;
            }
            else
            {
              return SVH_TUNNEL;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_BRIDGE;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_BRIDGE;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_CORNER;
            }
            else
            {
              return SVH_STEP;
            }
          }
        }
      }
      else
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_BRIDGE;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_BRIDGE;
            }
            else
            {
              return SVH_ROOF;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_CORNER;
            }
            else
            {
              return SVH_STEP;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_STEP;
            }
            else
            {
              return SVH_HUT;
            }
          }
        }
      }
    }
    else
    {
      if(f_visited[2])
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_GAP;
            }
            else
            {
              return SVH_BRIDGE;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_CORNER;
            }
            else
            {
              return SVH_STEP;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_BRIDGE;
            }
            else
            {
              return SVH_ROOF;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_STEP;
            }
            else
            {
              return SVH_HUT;
            }
          }
        }
      }
      else
      {
        if(f_visited[3])
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_CORNER;
            }
            else
            {
              return SVH_STEP;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_STEP;
            }
            else
            {
              return SVH_HUT;
            }
          }
        }
        else
        {
          if(f_visited[4])
          {
            if(f_visited[5])
            {
              return SVH_STEP;
            }
            else
            {
              return SVH_HUT;
            }
          }
          else
          {
            if(f_visited[5])
            {
              return SVH_HUT;
            }
            else
            {
              return SVH_START;
            }
          }
        }
      }
    }
  }
}

void SVwriter_svh::write_hexahedron(const I32* c_idx, const bool* c_final)
{
  if(!c_final)
  {
    fprintf(stderr, "ERROR: delayed finalization not supported by SVwriter_svh\n");
    exit(0);
  }

  compress_hexahedron(c_idx, 0, c_final);

  c_count++;
}

void SVwriter_svh::write_hexahedron(const I32* c_idx, const float* c_prop_f, const bool* c_final)
{
  if(!c_final)
  {
    fprintf(stderr, "ERROR: delayed finalization not supported by SVwriter_svh\n");
    exit(0);
  }

  compress_hexahedron(c_idx, c_prop_f, c_final);

  c_count++;
}

void SVwriter_svh::write_finalized(I32 final_idx)
{
  fprintf(stderr, "ERROR: delayed finalization not supported by SVwriter_svh\n");
  exit(0);
}

void SVwriter_svh::close(bool close_file, bool update_header)
{
#ifdef SVH_COMPRESS_CELL_DATA
  if(cellDataTmp) 
  {
    delete[] cellDataTmp;
    cellDataTmp = 0;
  }

#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
  if(tmpCellDataResidual) 
  {
    delete[] tmpCellDataResidual;
    tmpCellDataResidual = 0;
  }
#endif
#endif

  // close of SVwriter_svh
  if(edge_buffer_size)
  {
#ifdef SVH_RMOP_DEPENDS_ON_FINALIZATION
    re_conn_op->encode(rmOpX[last_fin][last_op], SVH_START);
#else
    re_conn_op->encode(rmOp[last_op], SVH_START);
#endif
    re_conn_op->encode(rmStart_VisitedVertices, 256);
  }
  else
  {
    assert(dv->size() == 0);
    re_conn_op->encode(rmDone, 1); // done
  }

#ifdef CALCULATE_PREDICTION_ERROR
  if(total_pred_err_normal)
  {
    fprintf(stdout,"err %.1f %.1f %.1f ",(float)(total_pred_err/(predict_inside+predict_across)),(float)(total_pred_err_normal/(predict_inside+predict_across)),(float)(total_pred_err_tangential/(predict_inside+predict_across)));
  }
  else
  {
    fprintf(stdout,"err %.1f ",(float)(total_pred_err/(predict_inside+predict_across)));
  }
#endif

#ifdef CALCULATE_QUANTIZATION_ERROR
  fprintf(stdout,"qnt %g ",max_quant_err);
#endif

#ifdef PRINT_CONTROL_OUTPUT
  fprintf(stderr,"edge_buffer_size %d edge_buffer_maxsize %d\n",edge_buffer_size,edge_buffer_maxsize);
  fprintf(stderr,"vertex_buffer_size %d vertex_buffer_maxsize %d\n",vertex_buffer_size,vertex_buffer_maxsize);
#endif

  finishEncoder(v_count, c_count);
  if(v_pq)
  {
    v_ic[0]->FinishCompressor();
    v_ic[1]->FinishCompressor();
    v_ic[2]->FinishCompressor();
    delete v_ic[0];
    delete v_ic[1];
    delete v_ic[2];
    delete v_pq;
  }
  else
  {
    v_fc[0]->FinishCompressor();
    v_fc[1]->FinishCompressor();
    v_fc[2]->FinishCompressor();
    delete v_fc[0];
    delete v_fc[1];
    delete v_fc[2];
  }
#ifdef SVH_COMPRESS_CELL_DATA
  if(c_pnum_f)
  {
    if (c_quantizer)
    {
      for(int i=0; i<c_pnum_f; ++i)
      {
        c_ic[i]->FinishCompressor();
        delete c_ic[i];
      }
      free(c_ic);
      delete c_quantizer;
    }
    else
    {
      for(int i=0; i<c_pnum_f; ++i)
      {
        c_fc[i]->FinishCompressor();
        delete c_fc[i];
      }
      free(c_fc);
    }
  }
#endif // SVH_COMPRESS_CELL_DATA
  finishModels();

  destroyVertexBuffer();
  destroyEdgeBuffer();

  if(dv->size()) fprintf(stderr,"WARNING: there are %d unfinalized vertices in dynamic vector\n",dv->size());
  if(vertex_hash->size()) fprintf(stderr,"WARNING: there are %d unfinalized vertices in vertex hash\n",(int)vertex_hash->size());

  delete dv;
  delete vertex_hash;
  delete lc_f;
  delete lc;

#ifdef STORE_QUANTIZED_COORDINATES
  fclose(quantized_coordinates_file);
#endif

#if !defined(SVH_COMPRESS_GEOMETRY)
  fclose(geofile);
#endif

  // close of SVwriter interface

  if(nverts != -1) if(nverts != v_count) fprintf(stderr,"WARNING: set nverts %d but v_count %d\n",nverts,v_count);
  if(ncells != -1) if(ncells != c_count) fprintf(stderr,"WARNING: set ncells %d but c_count %d\n",ncells,c_count);
  nverts = v_count;
  ncells = c_count;
  v_count = -1;
  c_count = -1;

  if(close_file && file && file != stdout) fclose(file);
  file = 0;
}

void SVwriter_svh::write_header()
{
  int i;

  // initialize internal memory management
  initVertexBuffer(1024, v_pnum_f);
  //initEdgeBuffer(1024, c_pnum_f);
#ifdef SVH_COMPRESS_CELL_DATA
  initEdgeBuffer(1024, c_pnum_f);
#else
  initEdgeBuffer(1024);
#endif

  // write nverts
  if(nverts == -1)
  {
    re_conn->writeRange(2,0);
  }
  else
  {
    re_conn->writeRange(2,1);
    re_conn->writeInt(nverts);
  }
  // write ncells
  if(ncells == -1)
  {
    re_conn->writeRange(2,0);
  }
  else
  {
    re_conn->writeRange(2,1);
    re_conn->writeInt(ncells);
  }
  // write which properties are there
  re_geom->writeByte(( !!v_pnum_f) |(( !!c_pnum_f) << 2));
#ifdef PRINT_CONTROL_OUTPUT
  printf("property mapping is %d\n",(( !!v_pnum_f) |(( !!c_pnum_f) << 2)));
#endif

  if(v_pnum_f)
  {
    re_geom->writeByte(v_pnum_f);
#ifdef PRINT_CONTROL_OUTPUT
    printf("have %d float vertex attributes\n",v_pnum_f);
#endif
    if(v_pmin_f && v_pmax_f)
    {
#ifdef PRINT_CONTROL_OUTPUT
      fprintf(stderr,"bbox is %f %f %f, %f %f %f\n",v_pmin_f[0], v_pmin_f[1], v_pmin_f[2], v_pmax_f[0], v_pmax_f[1], v_pmax_f[2]);
#endif
      re_geom->writeRange(2,1);
#ifdef PRINT_CONTROL_OUTPUT
      printf("have bounding box\n");
#endif
      for(i=0; i < v_pnum_f; i++)
      {
        re_geom->writeFloat(v_pmin_f[i]);
        re_geom->writeFloat(v_pmax_f[i]);
      }
    }
    else
    {
      re_geom->writeRange(2,0);
#ifdef PRINT_CONTROL_OUTPUT
      printf("have no bounding box\n");
#endif
    }
  }
  if(0) // v_pnum_i
  {
    re_geom->writeByte(0);
    if(0)
    {
    }
    else
    {
      re_geom->writeRange(2,0);
    }
  }
  if(c_pnum_f)
  {
    re_geom->writeByte(c_pnum_f);
    if(c_pmin_f && c_pmax_f)
    {
      re_geom->writeRange(2,1);
      for(i=0; i < c_pnum_f; i++)
      {
        re_geom->writeFloat(c_pmin_f[i]);
        re_geom->writeFloat(c_pmax_f[i]);
      }
    }
    else
    {
      re_geom->writeRange(2,0);
#ifdef PRINT_CONTROL_OUTPUT
      printf("have no bounding box\n");
#endif
    }
  }
  if(0)
  {
    re_geom->writeByte(0);
    if(0)
    {
      re_geom->writeRange(2,1);
    }
    else
    {
      re_geom->writeRange(2,0);
    }
  }

  if(true) // if have no comments
  {
    re_conn->writeRange(2,0);
#ifdef PRINT_CONTROL_OUTPUT
    printf("have no comments\n");
#endif
  }
  else
  {
    /* yet to be implemented */
  }

  // setup float quantizers and integer compressors

  if(v_pnum_f < 3)
  {
    fprintf(stderr,"WARNING: v_pnum_f is %d ... not implemented yet\n", v_pnum_f);
    exit(1);
  }

  re_geom->writeRange(17,high_bits);
#ifdef PRINT_CONTROL_OUTPUT
  printf("high_bits %d\n", high_bits);
#endif

  if(v_pmin_f && v_pmax_f && nbits)
  {
#ifdef PRINT_CONTROL_OUTPUT
    fprintf(stderr,"quantizing v_prop_f with %d bits\n",nbits);
#endif
    v_pq = new PositionQuantizerNew();
    v_pq->SetMinMax(v_pmin_f, v_pmax_f);
    v_pq->SetPrecision(nbits);
    v_pq->SetupQuantizer();

#ifdef PRINT_CONTROL_OUTPUT
    printf("aiQuantRange %d %d %d\n", v_pq->m_aiQuantRange[0], v_pq->m_aiQuantRange[1], v_pq->m_aiQuantRange[2]);
#endif

    v_ic[0] = new RangeIntegerCompressor();
    v_ic[0]->SetRange(v_pq->m_aiQuantRange[0]);
    v_ic[0]->SetupCompressor(re_geom, SVH_NUMBER_OF_PREDICTIONS, high_bits);

    v_ic[1] = new RangeIntegerCompressor();
    v_ic[1]->SetRange(v_pq->m_aiQuantRange[1]);
    v_ic[1]->SetupCompressor(re_geom, SVH_NUMBER_OF_PREDICTIONS, high_bits);

    v_ic[2] = new RangeIntegerCompressor();
    v_ic[2]->SetRange(v_pq->m_aiQuantRange[2]);
    v_ic[2]->SetupCompressor(re_geom, SVH_NUMBER_OF_PREDICTIONS, high_bits);
  }
  else
  {
#ifdef PRINT_CONTROL_OUTPUT
    if(nbits)
    {
      fprintf(stderr,"no bounding box. quantizing v_prop_f adaptively to %d bits\n",nbits);
    }
    else
    {
      fprintf(stderr,"no quantizing. lossless v_prop_f floating-point compression.\n");
    }
#endif
    v_pq = 0;

    v_fc[0] = new RangeFloatCompressor();
    v_fc[0]->SetPrecision(nbits);
    v_fc[0]->SetupCompressor(re_geom, SVH_NUMBER_OF_PREDICTIONS);

    v_fc[1] = new RangeFloatCompressor();
    v_fc[1]->SetPrecision(nbits);
    v_fc[1]->SetupCompressor(re_geom, SVH_NUMBER_OF_PREDICTIONS);

    v_fc[2] = new RangeFloatCompressor();
    v_fc[2]->SetPrecision(nbits);
    v_fc[2]->SetupCompressor(re_geom, SVH_NUMBER_OF_PREDICTIONS);
  }

#ifdef SVH_COMPRESS_CELL_DATA
  if(c_pnum_f)
  {
    re_conn->writeRange(25, cellDataBits);
    if(this->cellDataBits)
    {
      re_celldata->writeRange(17,cellDataHighBits);
#ifdef PRINT_CONTROL_OUTPUT
      printf("cellDataHighBits %d\n", cellDataHighBits);
#endif
      if(!(c_pmin_f && c_pmax_f))
      {
        fprintf(stderr,"FATAL ERROR: quantizeCellData set to TRUE but no BBox.\n");
        exit(1);
      }
#ifdef PRINT_CONTROL_OUTPUT
      fprintf(stderr,"quantizing c_prop_f with %d bits\n", cellDataBits);
#endif
      c_quantizer = new PositionQuantizerNew();
      c_quantizer->SetMinMax(c_pmin_f, c_pmax_f);
      c_quantizer->SetPrecision(cellDataBits);
      c_quantizer->SetupQuantizer();
      c_ic = (RangeIntegerCompressor**)malloc(sizeof(RangeIntegerCompressor*)*c_pnum_f);
      for(int i=0; i<c_pnum_f; ++i)
      {
        c_ic[i] = new RangeIntegerCompressor();
#ifdef SVH_CELL_DATA_SECONDORDER_PREDICTION
        c_ic[i]->SetRange(2*c_quantizer->m_aiQuantRange[i]);
        c_ic[i]->SetupCompressor(re_celldata, SVH_NUMBER_OF_OPERATIONS, cellDataHighBits+1);
#else
        c_ic[i]->SetRange(c_quantizer->m_aiQuantRange[i]);
        c_ic[i]->SetupCompressor(re_celldata, SVH_NUMBER_OF_OPERATIONS, cellDataHighBits);
#endif
      }
    }
    else
    {
#ifdef PRINT_CONTROL_OUTPUT
      fprintf(stderr,"lossless c_prop_f floating-point compression\n");
#endif
      c_quantizer= NULL;
      c_fc = (RangeFloatCompressor**)malloc(sizeof(RangeFloatCompressor*)*c_pnum_f);
      for(int i=0; i<c_pnum_f; ++i)
      {
        c_fc[i] = new RangeFloatCompressor();
        c_fc[i]->SetPrecision(0);
        c_fc[i]->SetupCompressor(re_celldata, SVH_NUMBER_OF_OPERATIONS);
      }
    }
  }
#endif

}

SVwriter_svh::SVwriter_svh()
{
  // init of SVwriter_svh
}

SVwriter_svh::~SVwriter_svh()
{
  // clean-up for SVwriter_svh
}
